/* obj_bpt.c - breakpoint object handling */

/*  Copyright 1991 Mark Russell, University of Kent at Canterbury.
 *
 *  You can do what you like with this source code as long as
 *  you don't try to make money out of it and you include an
 *  unaltered copy of this message (including the copyright).
 */

char ups_obj_bpt_c_sccsid[] = "@(#)obj_bpt.c	1.31 15/9/92 (UKC)";

#include <mtrprog/ifdefs.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <local/wn.h>
#include <local/ukcprog.h>
#include <local/obj/obj.h>

#include <local/obj/fed.h>
#include <local/obj/obj.h>

#include "objtypes.h"
#include "ups.h"
#include "symtab.h"
#include "breakpoint.h"
#include "ci.h"
#include "src.h"
#include "ui.h"
#include "text.h"
#include "exec.h"
#include "data.h"
#include "obj_bpt.h"
#include "obj_stack.h"
#include "menudata.h"
#include "state.h"
#include "debug.h"
#include "tdr.h"
#include "proc.h"
#include "expr.h"
#include "va_priv.h"

/*  Needed for temporary gross hack.
 */
extern proc_t Current_proc;

typedef struct bpdescst {
	breakpoint_t bd_breakpoint;
	editblock_id_t bd_editblock_id;
	fil_t *bd_fil;
	func_t *bd_func;
	int bd_lnum;
	bool bd_code_bad;
	bool bd_data_needs_initialising;
	bool bd_code_has_func_calls;
	bool want_save;
	bool bd_disabled;
	parse_id_t bd_parse_id;
	code_id_t bd_code_id;
} bpdesc_t;

typedef struct {
	FILE *ba_fp;
	char *ba_path;
} bptargs_t;

#define HAS_CODE(bd)	((bd)->bd_parse_id != NULL || (bd)->bd_code_bad)

static void bpfunc_draw PROTO((struct drawst *dets));
static void bpstate_draw PROTO((struct drawst *dets));
static int move_bpt PROTO((objid_t obj, const char *newstring, int edit_lnum));
static void show_bpt_source PROTO((bpdesc_t *bd, bool beep_on_failure));
static void set_bpt_state PROTO((objid_t obj, bool disable));
static int bpfunc_quitfunc PROTO((struct edescst *edesc, int n_tries));
static void bpfunc_edit PROTO((struct drawst fdets));
static void bplnum_edit PROTO((struct drawst fdets));
static void build_source_lines PROTO((fil_t *fil, int lnum, char *text,
				      const char ***p_lines, int *p_nlines));
static void load_breakpoints PROTO((void));
static void select_bd PROTO((editblock_id_t editblock_id,
			     editblock_action_t action));
static void update_editblock PROTO((bpdesc_t *bd, bool is_bpt, char *text));
static bool change_bd PROTO((editblock_id_t editblock_id,
			     const char **lines, int nlines,
			     callback_arg_t callback_arg));
static int set_initialise_data_flag PROTO((objid_t obj, fval_t unused_arg));
static bool execute_bp_code_internal PROTO((bpdesc_t *bd,
					    taddr_t fp, taddr_t ap,
					    int button));
static void save_interpreted_code_to_file PROTO((const char *path, FILE *fp, 
                                                 bpdesc_t *bd));
static void save_breakpoint_to_file PROTO((const char *path, FILE *fp, 
                                           bpdesc_t *bd));
static bool fil_and_lnum_to_bdobj PROTO((fil_t *fil, int lnum, objid_t *p_obj));
static int get_bp_lnum PROTO((func_t *f, const char *lnumstr,
			      const char *offsetstr));
static bool get_code_text PROTO((FILE *fp, int *p_lnum, ebuf_t *eb));
static int skip_to_close_curly PROTO(( FILE *fp, int *p_lnum));
static bool handle_editblock_change PROTO((editblock_id_t eb, const char *text,
					   bool force, size_t *p_error_point));
static bpdesc_t *make_bd PROTO((fil_t *fil, int lnum, func_t *func));

#ifdef OS_SUNOS_4
static taddr_t get_bpt_addr PROTO((objid_t obj));
#endif

/*  Maximum length of a function name. If you change this you must also
 *  change the Bpt_format string to match.
 */
#define LEN_BP_FNAME 	61

const char Bpt_format[] = "function:%[-]61cb line:%[B]6cn%[-]8cq\n";

/*  Field numbers for breakpoints
 */
/*#define FN_BPT_TYPESTR	0*/
#define FN_BPT_FNAME	0
#define FN_BPT_LNUM	1
#define FN_BPT_STATE	2
#define FN_BPT_LAST	3

/*  Flist terminator for breakpoint fields for set_all_fields().
 *  Can't use NULL as one of the fields is an int which can legitimately
 *  have the value 0.
 */
#define BPT_FLIST_TERMINATOR	((fval_t)-1)

fnamemap_t Bpt_fnamemap[] = {
	"function-name",	FN_BPT_FNAME,
	"lnum",			FN_BPT_LNUM,
	"state",		FN_BPT_STATE,
	NULL,			0,
};

const char Bphead_format[] = "Breakpoints\n";

#define BPHEAD_OBJCODE	((objid_t)Bphead_format)

static const char Stop_keyword[] = "#stop";

fdef_t Bpt_fdefs[] = {
	'b', bpfunc_draw, bpfunc_edit, NULL,
	'n', n_draw, bplnum_edit, NULL,
	'q', bpstate_draw, no_edit, NULL,
	0, NULL, NULL, NULL,
};


/************** Begin section from version 3.14 *************************/
static const char *Statefile_path = NULL;

void
set_breakpoint_statefile_path(path)
const char *path;
{
	Statefile_path = path;
}

static void
load_breakpoints()
{
	static char *lastpath = NULL;
	char *path;
	bool junk;

	if (lastpath == NULL)
		lastpath = strsave("");
	
	if (prompt_for_string("filename", "Load breakpoints from file: ",
			      lastpath, &path) != 0)
		return;

	if (strcmp(path, lastpath) != 0) {
		free(lastpath);
		lastpath = strsave(path);
	}

	read_config_file(path, FALSE, TRUE, TRUE, FALSE, &junk);
}

int
save_all_breakpoints_to_file(path, fp)
const char *path;
FILE *fp;
{
	objid_t obj;

	for (obj = get_code(BPHEAD_OBJCODE, OBJ_CHILD);
	     obj != NULL;
	     obj = get_code(obj, OBJ_NEXT)) {
		bpdesc_t *bd;

		bd = (bpdesc_t *)obj;
		
		if (bd->want_save && save_object_to_file(path, fp, bd) != 0) {
			return -1;
		}
	}

	return 0;
}

static int
save_object_to_file(path, fp, bd)
const char *path;
FILE *fp;
bpdesc_t *bd;
{
	if (bd->bd_breakpoint != NULL)
		save_breakpoint_to_file(path, fp, bd);
	else
		save_interpreted_code_to_file(path, fp, bd);

	fflush(fp);

	if (ferror(fp)) {
		errf("Error saving object to", "file", path);
		return -1;
	}

	return 0;
}

static void
save_interpreted_code_to_file(path, fp, bd)
const char *path;
FILE *fp;
bpdesc_t *bd;
{
  int nlines, i;
  char **text;

  fprintf(fp, "code %s %d {\n", bd->bd_fil->fi_name, bd->bd_lnum);
  text = get_editblock_text(bd->bd_editblock_id, &nlines);
  for (i = 0; i < nlines; i++, text++)
    dump_text(*text, fp);
  fputs("}\n", fp);
}

static void
save_breakpoint_to_file(path, fp, bd)
const char *path;
FILE *fp;
bpdesc_t *bd;
{
	func_t *f;

	f = bd->bd_func;
	
	fprintf(fp, "breakpoint %s", f->fu_name);
	
	if (f->fu_fil != NULL) {
		lno_t *ln;
		int n_code_lines;

		n_code_lines = 0;
		for (ln = FU_LNOS(f); ln != NULL; ln = ln->ln_next) {
			if (ln->ln_num == bd->bd_lnum)
				break;
			++n_code_lines;
		}
		
		fprintf(fp, " %s %d [+%d]",
			f->fu_fil->fi_name, bd->bd_lnum, n_code_lines);
	}
	
	if (HAS_CODE(bd)) {
	  int nlines, i;
	  char **text;

	  fputs(" {\n", fp);
	  text = get_editblock_text(bd->bd_editblock_id, &nlines);
	  for (i = 0; i < nlines; i++, text++)
	    dump_text(*text, fp);
	  fputc('}', fp);
	}

	fputc('\n', fp);
}

int
handle_add_breakpoint_command(cmd, args, nargs, from_statefile, ignore,
			      fp, p_lnum)
const char *cmd;
char **args;
int nargs;
bool from_statefile, ignore;
FILE *fp;
int *p_lnum;
{
	func_t *f, *f1;
	char *funcname;
	const char *filename, *lnumstr, *offsetstr;
	bool ok, have_code;
	fil_t *fil;
	int lnum;
	objid_t obj;

	have_code = nargs > 1 && strcmp(args[nargs - 1], "{") == 0;

	if (have_code)
		--nargs;

	if (nargs < 1 || nargs > 4) {
		errf("Usage: breakpoint funcname [filename [lnum [offset]]] [{]");
		return -1;
	}

	funcname = args[0];
	filename = (nargs > 1) ? args[1] : NULL;
	lnumstr = (nargs > 2) ? args[2] : NULL;
	offsetstr = (nargs > 3) ? args[3] : NULL;
	
	if (filename == NULL) {
		if (find_func_by_name(funcname, &f, &f1, 0) != 0) {
			skip_to_close_curly(fp, p_lnum);
			return -1;
		}

		fil = f->fu_fil;
		lnum = -1;
	}
	else {
		if ((fil = name_to_fil(filename)) == NULL ||
		    (f = name_and_fil_to_func(funcname, fil)) == NULL) {
			/*  Function may have moved to another file.
			 */
			if (find_func_by_name(funcname, &f, &f1, 0) != 0) {
				skip_to_close_curly(fp, p_lnum);
				return -1;
			}
			fil = f->fu_fil;
		}

		if (lnumstr == NULL)
			lnum = -1;
		else
			lnum = get_bp_lnum(f, lnumstr, offsetstr);
	}

	if (ignore) {
		if (have_code && skip_to_close_curly(fp, p_lnum) != 0)
			return -1;
		return 0;
	}

	if (from_statefile && fil_and_lnum_to_bdobj(f->fu_fil, lnum, &obj))
		remove_object(obj, OBJ_SELF);
	
	if ((obj = add_breakpoint_object(f, fil, lnum)) == NULL) {
		if (have_code)
			skip_to_close_curly(fp, p_lnum);
		return -1;
	}

	((bpdesc_t *)obj)->want_save = from_statefile;

	if (have_code) {
		ebuf_t *eb;
		
		eb = ebuf_create(TRUE);

		ok = get_code_text(fp, p_lnum, eb) &&
		     change_bd_text(obj, ebuf_get(eb, (int *)NULL));

		if (!ok)
			remove_object(obj, OBJ_SELF);

		ebuf_free(eb);
	}
	else {
		ok = TRUE;
	}

	return ok ? 0 : -1;
}

int
handle_add_code_command(cmd, args, nargs, from_statefile, ignore, fp, p_lnum)
const char *cmd;
char **args;
int nargs;
bool from_statefile, ignore;
FILE *fp;
int *p_lnum;
{
	fil_t *fil;
	const char *filename, *lnumstr;
	int lnum;
	objid_t obj;
	ebuf_t *eb;
	char *text, *ends;
	
	if (nargs != 3 || strcmp(args[2], "{") != 0) {
		errf("Usage: code filename lnum {");
		return -1;
	}

	filename = args[0];
	lnumstr = args[1];
	
	if ((fil = name_to_fil(filename)) == NULL) {
		skip_to_close_curly(fp, p_lnum);
		return -1;
	}
	
	lnum = strtol(lnumstr, &ends, 10);

	if (ends == lnumstr || *ends != '\0') {
		errf("Line number `%s' is not a decimal integer", lnumstr);
		skip_to_close_curly(fp, p_lnum);
		return -1;
	}

	if (ignore)
		return skip_to_close_curly(fp, p_lnum);

	eb = ebuf_create(TRUE);
	
	if (!get_code_text(fp, p_lnum, eb)) {
		ebuf_free(eb);
		skip_to_close_curly(fp, p_lnum);
		return -1;
	}
	text = ebuf_get(eb, (int *)NULL);
	
	if (from_statefile && fil_and_lnum_to_bdobj(fil, lnum, &obj))
		remove_object(obj, OBJ_SELF);

	obj = add_interpreted_code(fil, lnum, strsave(text)); 
	ebuf_free(eb);
	
	if (obj != NULL)
		((bpdesc_t *)obj)->want_save = from_statefile;

	return (obj != NULL) ? 0 : -1;
}

static int
skip_to_close_curly(fp, p_lnum)
FILE *fp;
int *p_lnum;
{
	char *line;

	while ((line = fpgetline(fp)) != NULL) {
		if (strcmp(line, "}") == 0)
			return 0;
		++*p_lnum;
	}

	errf("Missing closing `}' in breakpoint code");
	return -1;
}

static bool
get_code_text(fp, p_lnum, eb)
FILE *fp;
int *p_lnum;
ebuf_t *eb;
{
	int code_lnum;
	char *line;

	code_lnum = *p_lnum;

	while ((line = fpgetline(fp)) != NULL) {

		++*p_lnum;

		if (strcmp(line, "}") == 0) {
			ebuf_add(eb, "", 1);
			return TRUE;
		}
		
		if (*line == '#')
			continue;
		if (*line == '\t')
			++line;

		ebuf_addstr(eb, line);
		ebuf_add(eb, "\n", 1);
	}

	*p_lnum = code_lnum;
	errf("Missing `}'");
	return FALSE;
}

static int
get_bp_lnum(f, lnumstr, offsetstr)
func_t *f;
const char *lnumstr, *offsetstr;
{
	int lnum;
	char *ends;
			
	lnum = strtol(lnumstr, &ends, 10);
	
	if (lnum < 0 || ends == lnumstr || *ends != '\0') {
		errf("Ignored illegal line number `%s'", lnumstr);
		lnum = -1;
	}

	if (offsetstr != NULL) {
		lno_t *ln;
		int i, count;
		char junkc;

		if (sscanf(offsetstr, "[+%d]%c", &count, &junkc) != 1)
			errf("Ignored malformed offset `%s'", offsetstr);
				
		ln = FU_LNOS(f);
		for (i = 0; i < count && ln != NULL; ++i)
			ln = ln->ln_next;

		if (ln != NULL)
			lnum = ln->ln_num;
		else
			errf("Ignored out of range offset `%s'", offsetstr);
	}

	return lnum;
}

bool
change_bd_text(obj, text)
objid_t obj;
const char *text;
{
	bpdesc_t *bd;
	size_t junk;

	bd = (bpdesc_t *)obj;
	
	update_editblock(bd, bd->bd_breakpoint != NULL, strsave(text));
	return handle_editblock_change(bd->bd_editblock_id, text, FALSE, &junk);
}

static bool
handle_editblock_change(eb, text, force, p_error_pos)
editblock_id_t eb;
const char *text;
bool force;
size_t *p_error_pos;
{
	bpdesc_t *bd;
	compile_res_t *cr;
	block_t *block;
	char **lines;
	const char *start_text, *end_text;
	int nlines;
	lexinfo_t lxbuf;
			      
	bd = (bpdesc_t *)get_editblock_data(eb);
	
/*	td_record_bpt_code_edit((objid_t)bd, text);*/

	if (bd->bd_func == NULL) {
		block = bd->bd_fil->fi_block;
		start_text = "";
		end_text = "";
	}
	else {
		block = FU_BLOCKS(bd->bd_func);
		while (block != NULL) {
			block_t *child;

			for (child = block->bl_blocks;
			     child != NULL;
			     child = child->bl_next) {
				if (child->bl_start_lnum <= bd->bd_lnum &&
				    bd->bd_lnum <= child->bl_end_lnum)
					break;
			}
		
			if (child == NULL)
				break;
			block = child;
		}

		if (block == NULL)
			block = bd->bd_fil->fi_block;

		start_text = "void $start(void) {";
		end_text = "}";
	}
	
	/*  Make sure block vars and type information of fil are loaded.
	 */
	FI_VARS(bd->bd_fil);
	
	if (bd->bd_parse_id != NULL)
		ci_free_parse_id(bd->bd_parse_id);
	
	if (bd->bd_code_id != NULL) {
		ci_free_code_id(bd->bd_code_id);
		bd->bd_code_id = NULL;
	}
/*	if (bd->machine != NULL) {
		ci_free_machine(bd->machine);
		bd->machine = NULL;
	} RGA*/

	lines = ssplit(text, "\0\n");
	for (nlines = 0; lines[nlines] != NULL; ++nlines)
		;
	
	cr = compile_code((const char **)lines, nlines, block, (char *)NULL,
					&lxbuf, start_text, end_text,
					Stop_keyword, Stop_funcname);

	bd->bd_parse_id = cr->cr_parse_id;
	bd->bd_code_id = cr->cr_code_id;
/*	bd->machine = cr->cr_machine;*/
/*	bd->bd_code_bad = cr->cr_machine == NULL || cr->cr_parse_id == NULL;*/
	bd->bd_code_bad = cr->cr_code_id == NULL || cr->cr_parse_id == NULL;
	bd->bd_code_has_func_calls = cr->cr_code_has_func_calls;
	bd->want_save = TRUE;

	if (bd->bd_code_bad) {
		int lnum;

		*p_error_pos = lxbuf.lx_cnum;
		
		for (lnum = 0; lnum < nlines && lnum < lxbuf.lx_lnum; ++lnum)
			*p_error_pos += strlen(lines[lnum]) + 1;

		/*  BUG: the following stuff should not really be necessary,
		 *       but I don't have time right now to investigate.
		 */
		if (*p_error_pos >= strlen(text)) {
			*p_error_pos = strlen(text);

			if (*p_error_pos != 0)
				--*p_error_pos;
		}

		return force;
	}
	
	return TRUE;
}

static bool
fil_and_lnum_to_bdobj(fil, lnum, p_obj)
fil_t *fil;
int lnum;
objid_t *p_obj;
{
	objid_t obj;

	for (obj = get_code(BPHEAD_OBJCODE, OBJ_CHILD);
	     obj != NULL;
	     obj = get_code(obj, OBJ_NEXT)) {
		bpdesc_t *bd;

		bd = (bpdesc_t *)obj;
		
		if (bd->bd_fil == fil && bd->bd_lnum == lnum) {
			*p_obj = obj;
			return TRUE;
		}
	}

	return FALSE;
}

/************** End section from version 3.14 *************************/

const char *
bpt_getobjname(obj)
objid_t obj;
{
	static char *last = NULL;
	bpdesc_t *bd;

	bd = (bpdesc_t *)obj;

	if (last != NULL)
		free(last);

	last = strf("%s:%d", bd->bd_func->fu_name, bd->bd_lnum);

	return last;
}

static void
bpfunc_draw(dets)
register struct drawst *dets;
{
	wn_ttext(dets->dr_wn, (char *)dets->dr_fval, dets->dr_x, dets->dr_y,
						    dets->dr_fg, dets->dr_bg);
}

static void
bpstate_draw(dets)
register struct drawst *dets;
{
	wn_ttext(dets->dr_wn, (char *)dets->dr_fval, dets->dr_x, dets->dr_y,
						    dets->dr_fg, dets->dr_bg);
}

#ifdef OS_SUNOS_4
static taddr_t
get_bpt_addr(obj)
objid_t obj;
{
	bpdesc_t *bd;

	bd = (bpdesc_t *)obj;
	return (bd->bd_lnum != 0) ? lnum_to_addr(bd->bd_func, bd->bd_lnum)
				  : min_bpt_addr(bd->bd_func);
}
#endif

/*  Build a source line for a breakpoint.  We duplicate any whitespace
 *  at the beginning of the source line following the breakpoint.
 */
static void
build_source_lines(fil, lnum, text, p_lines, p_nlines)
fil_t *fil;
int lnum;
char *text;
const char ***p_lines;
int *p_nlines;
{
	static const char *linestab[32];
	char *c, *start = text, *d;
	int i = 1;

	if (lnum < 1 || open_source_file(fil, FALSE, FALSE, FALSE) != 0 ||
					   lnum > so_get_nlines(fil->fi_so))
		linestab[0] = text;
	else {
		static char buf[512];
		const char *line, *cptr;
		int len;
		
		/*  We don't check for tabs in the loop below because
		 *  so_getline turns them into spaces for us.
		 */
		line = so_getline(fil->fi_so, lnum - 1);
		for (cptr = line; *cptr == ' '; ++cptr)
			;
		len = cptr - line;

		if (td_have_window()) {
			font_t *srcfont, *editfont;

			src_get_fonts(get_current_srcwin(), &srcfont,&editfont);
			if (editfont->ft_width != 0)
				len = (len * srcfont->ft_width) /
							editfont->ft_width;
		}

		if (len > sizeof(buf) - strlen(text))
			len = sizeof(buf) - strlen(text);
		memset(buf, ' ', len);
		for (i = 0, c = text, d = buf; *c; c++)
		  if (*c == '\n')
		  {
		    *c = 0;
		    d += len;
		    strcpy(d, start);
		    linestab[i] = d;
		    start = c + 1;
		    d += strlen(linestab[i]) + 1;
		    i++;
		  }
		if (start == text)
		{
		  strcpy(buf + len, text);
		  linestab[i] = buf;
		  i++;
		}
	      }

	*p_lines = linestab;
	*p_nlines = i;
}

#ifdef OS_SUNOS_4
void
recalculate_bpt_addrs()
{
	objid_t obj;
	bpdesc_t *bd;

	obj = get_code(BPHEAD_OBJCODE, OBJ_CHILD);
	for (; obj != NULL; obj = get_code(obj, OBJ_NEXT)) {
		bd = (bpdesc_t *)obj;
		if (bd->bd_breakpoint != NULL) {
			remove_breakpoint(bd->bd_breakpoint);
			bd->bd_breakpoint = add_breakpoint(get_bpt_addr(obj));
			set_breakpoint_data(bd->bd_breakpoint, (long)obj);
		}
		if (bd->bd_code_has_func_calls) {
			bd->bd_code_id = recompile_code(bd->bd_parse_id,
							bd->bd_code_id,
							(char *)NULL);
		}
	}
}
#endif /* OS_SUNOS_4 */

void
bpt_select(wn, obj, x, y, width, height, flags)
int wn;
objid_t obj;
int x, y, width, height, flags;
{
	bpdesc_t *bd;

	bd = (bpdesc_t *)obj;

	if (flags & SEL_CHANGING && bd->bd_editblock_id != NULL) {
		editblock_action_t action;

		action = (flags & SEL_ON) ? EDL_SELECT : EDL_DESELECT;
		select_editblock(bd->bd_editblock_id, action,
							DONT_CALL_CALLBACK);
	}
	gen_select(wn, obj, x, y, width, height, flags);
}

int
change_bpt_lines(obj, lines, nlines)
objid_t obj;
const char **lines;
int nlines;
{
	bpdesc_t *bd;
	lexinfo_t lxbuf;

	bd = (bpdesc_t *)obj;

	return change_editblock(bd->bd_editblock_id, lines, nlines,
			 	DO_CALL_CALLBACK, (callback_arg_t)&lxbuf);
	
}

/*  Handle a change to the text of an editblock block for a breakpoint.
 *
 *  This function is registered as a callback for breakpoint editblock objects.
 *
 *  We never return FALSE (which indicates an error) because we always let
 *  edits complete.  We set a flag in the breakpoint if the code is bad,
 *  and complain whenever execution hits the breakpoint.
 */
static bool
change_bd(editblock_id, lines, nlines, callback_arg)
editblock_id_t editblock_id;
const char **lines;
int nlines;
callback_arg_t callback_arg;
{
	compile_res_t *cr;
	bpdesc_t *bd;
	block_t *block;

	bd = (bpdesc_t *)get_editblock_data(editblock_id);

	td_record_bpt_code_edit((objid_t)bd, lines, nlines);

	block = FU_BLOCKS(bd->bd_func);
	while (block != NULL) {
		block_t *child;

		for (child = block->bl_blocks; child != NULL; child = child->bl_next)
			if (child->bl_start_lnum <= bd->bd_lnum &&
						bd->bd_lnum <= child->bl_end_lnum)
				break;
		if (child == NULL)
			break;
		block = child;
	}

	if (block == NULL)
		block = bd->bd_func->fu_fil->fi_block;
	
	/*  Make sure block vars and type information of fil are loaded.
	 */
	FI_VARS(bd->bd_func->fu_fil);
	
	if (bd->bd_parse_id != NULL)
		ci_free_parse_id(bd->bd_parse_id);
	if (bd->bd_code_id != NULL) {
		ci_free_code_id(bd->bd_code_id);
		bd->bd_code_id = NULL;
	}

	cr = compile_code(lines, nlines, block, (char *)NULL,
					(lexinfo_t *)callback_arg,
					"void $start(void) {", "}",
					Stop_keyword, Stop_funcname);

	bd->bd_parse_id = cr->cr_parse_id;
	bd->bd_code_id = cr->cr_code_id;
	bd->bd_code_bad = cr->cr_code_id == NULL || cr->cr_parse_id == NULL;
	bd->bd_code_has_func_calls = cr->cr_code_has_func_calls;

	return TRUE;
}

/*  Select or deselect a breakpoint object.  This function is registered
 *  as a callback for breakpoint editblock objects.
 */
static void
select_bd(editblock_id, action)
editblock_id_t editblock_id;
editblock_action_t action;
{
	bpdesc_t *bd;

	bd = (bpdesc_t *)get_editblock_data(editblock_id);
	switch (action) {
	case EDL_SELECT:
		select_object((objid_t)bd, TRUE, OBJ_SELF);
		break;
	case EDL_DESELECT:
		select_object((objid_t)bd, FALSE, OBJ_SELF);
		break;
	case EDL_REMOVE:
		bd->bd_editblock_id = NULL;
		break;
	default:
		panic("unknown action in sb");
	}
}

/*  Make a new breakpoint object.
 */
objid_t
add_breakpoint_object(f, fil, lnum)
func_t *f;
fil_t *fil;
int lnum;
{
	fval_t fields[FN_BPT_LAST + 1];
	bpdesc_t *bd;
	const char *fname;
	taddr_t addr;

	if (f == NULL)
	{
		fname = "";
		fil = NULL;
	}
	else {
		if (lnum == -1)
			addr = min_bpt_addr(f);
		else {
			if (map_lnum_to_addr(f, lnum, &addr) != 0)
				return NULL;
		}

		if (addr_to_breakpoint(addr) != 0) {
			errf("There is already a breakpoint at line %d of %s",
							      lnum, f->fu_name);
			return NULL;
		}

		lnum = addr_to_lnum(f, addr);
		fname = f->fu_name;
	}

	fields[FN_BPT_FNAME] = (fval_t)fname;
	fields[FN_BPT_LNUM] = (fval_t)lnum;
	fields[FN_BPT_STATE] = (fval_t)"active";
	fields[FN_BPT_LAST] = BPT_FLIST_TERMINATOR;

	/*  Add an object for the breakpoint.
	 */
	td_set_obj_updating(OBJ_UPDATING_OFF);
	bd = (bpdesc_t *) e_malloc(sizeof(bpdesc_t));
	bd->bd_breakpoint = NULL;
	bd->bd_func = f;
	bd->bd_lnum = lnum;
	bd->bd_editblock_id = NULL;
	bd->bd_parse_id = NULL;
	bd->bd_code_id = NULL;
	bd->bd_code_bad = FALSE;
	bd->bd_data_needs_initialising = FALSE;
	bd->bd_code_has_func_calls = FALSE;
	bd->bd_disabled = FALSE;
	bd->bd_fil = fil;
	bd->want_save = TRUE;

	new_object((objid_t)bd, OT_BPT, BPHEAD_OBJCODE, OBJ_LAST_CHILD);
	set_all_fields((objid_t)bd, fields, BPT_FLIST_TERMINATOR);
	td_set_obj_updating(OBJ_UPDATING_ON);
	ensure_visible((objid_t)bd);

	if (bd->bd_func != NULL) {
		bd->bd_breakpoint = add_breakpoint(addr);
		set_breakpoint_data(bd->bd_breakpoint, (long)bd);
		update_editblock(bd, TRUE, (char *)NULL);
	}
	else {
		td_obj_edit_field((objid_t)bd, FN_BPT_FNAME, 0, 0);
		if (bd->bd_func == NULL) {
			remove_object((objid_t)bd, OBJ_SELF);
			bd = NULL;
		}
	}

	return (objid_t)bd;
}

static bpdesc_t *
make_bd(fil, lnum, func)
fil_t *fil;
int lnum;
func_t *func;
{
	bpdesc_t *bd;
	
	bd = (bpdesc_t *)e_malloc(sizeof(bpdesc_t));
	bd->bd_breakpoint = NULL;
	bd->bd_fil = fil;
	bd->bd_func = func;
	bd->bd_lnum = lnum;
	bd->bd_editblock_id = NULL;
	bd->bd_parse_id = NULL;
	bd->bd_code_bad = FALSE;
	bd->bd_data_needs_initialising = FALSE;
	bd->bd_code_has_func_calls = FALSE;
	bd->want_save = TRUE;
	bd->bd_disabled = FALSE;

	return bd;
}
		     
objid_t
add_interpreted_code(fil, lnum, text)
fil_t *fil;
int lnum;
char *text;
{
	objid_t obj;
	fval_t fields[FN_BPT_LAST + 1];
	bpdesc_t *bd;
	func_t *f;
	taddr_t addr;

	if ((f = lnum_to_func(fil, lnum)) != NULL &&
	    (map_lnum_to_addr(f, lnum, &addr) != 0 && addr != 0)) {
		errf("There is executable code at line %d of %s",
		     lnum, fil->fi_name);
		return NULL;
	}
	
	if (fil_and_lnum_to_bdobj(fil, lnum, &obj)) {
		bd = (bpdesc_t *)obj;
		
		errf("There is already %s at line %d of %s",
		     (bd->bd_breakpoint != NULL) ? "a breakpoint"
		     			      : "interpreted code",
		     lnum, fil->fi_name);
		return NULL;
	}

	
/*	fields[FN_BPT_TYPESTR] = "file";*/
	fields[FN_BPT_FNAME] = (fval_t)fil->fi_name;
	fields[FN_BPT_LNUM] = (fval_t)lnum;
	fields[FN_BPT_LAST] = BPT_FLIST_TERMINATOR;

	bd = make_bd(fil, lnum, (func_t *)NULL);
	
	td_set_obj_updating(OBJ_UPDATING_OFF);
	new_object((objid_t)bd, OT_BPT, BPHEAD_OBJCODE, OBJ_LAST_CHILD);
	set_all_fields((objid_t)bd, fields, BPT_FLIST_TERMINATOR);
	td_set_obj_updating(OBJ_UPDATING_ON);
	ensure_visible((objid_t)bd);

	update_editblock(bd, FALSE, text);

	return (objid_t)bd;
}

/*  Update the editblock for a breakpoint after a change of status
 *  or position.  If necessary, create the editblock block.
 */
static void
update_editblock(bd, is_bpt, text)
bpdesc_t *bd;
bool is_bpt;
char *text;
{
	const char **lines;
	int nlines;
	editblock_id_t editblock_id;
	char textbuf[sizeof(Stop_keyword) + 1];	/* space for added semicolon */

	editblock_id = bd->bd_editblock_id;

	if (bd->bd_func->fu_fil == NULL) {
		if (editblock_id != NULL) {
			remove_editblock(editblock_id, DONT_CALL_CALLBACK);
			bd->bd_editblock_id = NULL;
		}
		return;
	}

	strcpy(textbuf, Stop_keyword);
	strcat(textbuf, ";");
	build_source_lines(bd->bd_func->fu_fil, bd->bd_lnum,
			   text ? text : textbuf, &lines, &nlines);

	if (editblock_id != NULL) {
		editblockinfo_t eibuf;

		get_editblock_info(editblock_id, &eibuf);
		if (bd->bd_func->fu_fil != eibuf.ei_fil ||
					   bd->bd_lnum != eibuf.ei_fil_lnum) {
			remove_editblock(editblock_id, DONT_CALL_CALLBACK);
			editblock_id = NULL;
		}
	}

	if (editblock_id != NULL)
		change_editblock(editblock_id, lines, nlines,
				 DONT_CALL_CALLBACK, (callback_arg_t)NULL);
	else {
		editblock_id = add_editblock(bd->bd_func->fu_fil, bd->bd_lnum,
								lines, nlines);
		set_editblock_data(editblock_id, (char *)bd);
		register_editblock_callback(editblock_id, select_bd);
		register_editblock_edit_callback(editblock_id, change_bd);
	}

	bd->bd_editblock_id = editblock_id;
}

/*  Attempt to move a breakpoint after editing the function name
 *  or line number.
 */
static int
move_bpt(obj, newstring, edit_lnum)
objid_t obj;
const char *newstring;
int edit_lnum;
{
	fval_t fields[FN_BPT_LAST + 1], fields1[FN_BPT_LAST + 1];
	func_t *f, *f1 = NULL;
	bpdesc_t *bd, *bd1 = NULL;
	char *fname;
	int lnum, lnum1, ret, method_call = 0, multi_func = 0, lnum_n;
	taddr_t addr, addr1;
	objid_t obj1 = NULL, obj_n;


	bd = (bpdesc_t *)obj;
	f = bd->bd_func;

	if (edit_lnum > 0) {
		lnum = lnum1 = atoi(newstring);
		fname = (char *)f->fu_name;
	}
	else {
		lnum = lnum1 = bd->bd_lnum;
		fname = (char *)newstring;
	}

	get_all_fields(obj, fields);
	if (strstr(fname, "::"))
	  method_call = 1;
	if (f == NULL || strcmp(fname, f->fu_name) != 0) {
	  if (method_call)
	  {
	    func_bp_list_t *fl, *funchead;
	    taddr_t call_bpt_addr;

	    funchead = 0;

	    get_matching_func_bp_list(fname, &funchead);

	    if (funchead)
	    {
	      multi_func = 1;
	      if (funchead->fl_next)
		errf("\bPlacing multiple breakpoints in overloaded/multiply linked methods");
	      for (fl = funchead; fl != NULL; fl = fl->fl_next)
	      {		 
		call_bpt_addr = min_bpt_addr(fl->fl_func);
		if (addr_to_breakpoint(call_bpt_addr) != 0)
		  continue;	/* duplicate */
		if (f == NULL)
		{		/* skip  - use code at bottom for this */
		  f = fl->fl_func;
		  addr = min_bpt_addr(f);
		  lnum = lnum1 = addr_to_lnum(f, addr);
		  continue;
		}
		lnum_n = addr_to_lnum(fl->fl_func, call_bpt_addr);
		if (!lnum_n)
		  lnum_n = -1;	/* allow multiple breakpoints in libraries */
		obj_n = add_breakpoint_object
		  (fl->fl_func, fl->fl_func->fu_fil, lnum_n);
	      }
	      do_free_func_bp_list_list(funchead);
	    }
	    else
	    {
	      errf("No function `%s'", fname);
	      return -1;
	    }
	  }
	  else
	  {
	    ret = find_func_by_name(fname, &f, &f1, 0);
	    if (ret != 0)
	      return ret;
	    
	    addr = addr1 = min_bpt_addr(f);
	    lnum = lnum1 = addr_to_lnum(f, addr);
	    if (bd->bd_func && (!f1 || edit_lnum < 0)) /* func name change */
	    {
	      fil_t *fil1;
	    
	      if (find_duplicate_file(bd->bd_func->fu_fil, &fil1))
	      {
		obj1 = (objid_t)get_editblock_in_tracking_file
		  (bd->bd_editblock_id, fil1);
		if (obj1)
		{
		  bd1 = (bpdesc_t *)obj1;
		  if (bd->bd_lnum != bd1->bd_lnum)
		    obj1 = NULL;
		  else
		    get_all_fields(obj1, fields1);
		}
	      }
	    }
	  }
	  if (f1)
	  {
	    func_bp_list_t *fl, *funchead;
	    taddr_t call_bpt_addr;

	    funchead = 0;

	    get_matching_func_bp_list(fname, &funchead);

	    if (funchead)
	    {
	      multi_func = 1;
	      if (method_call)
		errf("\bPlacing multiple breakpoints in overloaded/multiply linked methods");
	      else
		errf("\bAdding multiple breakpoints because function has been multiply linked");
	      for (fl = funchead->fl_next; fl != NULL; fl = fl->fl_next)
	      {		 /* skip first one - use code at bottom for this */
/*		if (!method_call && strcmp(fname, fl->fl_func->fu_name))
		  continue;*/
		call_bpt_addr = min_bpt_addr(fl->fl_func);
		if (addr_to_breakpoint(call_bpt_addr) != 0)
		  continue;	/* duplicate */
		lnum_n = addr_to_lnum(fl->fl_func, call_bpt_addr);
		if (!lnum_n)
		  lnum_n = -1;	/* allow multiple breakpoints in libraries */
		obj_n = add_breakpoint_object
		  (fl->fl_func, fl->fl_func->fu_fil, lnum_n);
	      }
	      do_free_func_bp_list_list(funchead);
	    }
/*	    return 0;*/
	  }
	}
	else {
	  fil_t *fil1;
	    
	  if (map_lnum_to_addr(f, lnum, &addr) != 0)
	    return -1;
	  lnum = lnum1 = addr_to_lnum(f, addr);
	  if (find_duplicate_file(bd->bd_func->fu_fil, &fil1))
	  {
	    obj1 = (objid_t)get_editblock_in_tracking_file
	      (bd->bd_editblock_id, fil1);
	    if (obj1)
	    {
	      bd1 = (bpdesc_t *)obj1;
	      f1 = bd1->bd_func;
	      get_all_fields(obj1, fields1);
	      if (map_lnum_to_addr(f1, lnum1, &addr1) != 0)
		return -1;
	      lnum1 = addr_to_lnum(f1, addr1);
	    }
	  }
	}
	if (f == NULL)
	{
	  errf("No match for `%s'", fname);
	  return -1;
	}
	
	/*  We don't allow more than one breakpoint at a given address.
	 */
	if (!multi_func && addr_to_breakpoint(addr) != 0) {
	  errf("There is already a breakpoint at line %d of %s",
	       lnum, f->fu_name);
	  return -1;
	}
	
	fields[FN_BPT_FNAME] = f->fu_name;
	fields[FN_BPT_LNUM] = (fval_t)lnum;
	fields[FN_BPT_LAST] = BPT_FLIST_TERMINATOR;
	set_all_fields(obj, fields, BPT_FLIST_TERMINATOR);
	
	bd->bd_func = f;
	bd->bd_lnum = lnum;
	
	/*  Move the actual breakpoint in the code to the new position.
	 */
	if (bd->bd_breakpoint != NULL)
	  remove_breakpoint(bd->bd_breakpoint);
	bd->bd_breakpoint = add_breakpoint(addr);
	set_breakpoint_data(bd->bd_breakpoint, (long)bd);
	update_editblock(bd, TRUE, (char *)NULL);
	if (obj1)
	{
	  if (!f1 && bd1->bd_breakpoint != NULL)
	    remove_object(obj1, OBJ_SELF);
	  else
	  {
	    fields1[FN_BPT_FNAME] = f1 ? f1->fu_name : f->fu_name;
	    fields1[FN_BPT_LNUM] = (fval_t)lnum1;
	    fields1[FN_BPT_LAST] = BPT_FLIST_TERMINATOR;
	    set_all_fields(obj1, fields1, BPT_FLIST_TERMINATOR);
	    
	    bd1->bd_func = f1 ? f1 : f;
	    bd1->bd_lnum = lnum1;
	    
	    /*  Move the actual breakpoint in the code to the new position.
	     */
	    if (bd1->bd_breakpoint != NULL)
	      remove_breakpoint(bd1->bd_breakpoint);
	    bd1->bd_breakpoint = add_breakpoint(addr1);
	    set_breakpoint_data(bd1->bd_breakpoint, (long)bd1);
	    update_editblock(bd1, TRUE, (char *)NULL);
	  }
	}

	return 0;
}


int
global_bp_enabled(set, reset)
     int set;
     int reset;
{
  static int enabled = 1;

  if (set)
    enabled = 1;
  if (reset)
    enabled = 0;
  return(enabled);
}

static bool
execute_bp_code_internal(bd, fp, ap, button)
bpdesc_t *bd;
taddr_t fp, ap;
int button;
{
	ci_exec_result_t res;

	if (!button && bd->bd_disabled)
		return FALSE;

	if (bd->bd_code_bad) {
		errf("Bad breakpoint code");
		return TRUE;
	}
	if (bd->bd_code_id == NULL)
	{
	  if (button)
	    errf("No breakpoint code to execute");
	  return TRUE;
	}
	
	if (bd->bd_data_needs_initialising) {
		ci_initialise_code(bd->bd_code_id, TRUE);
		bd->bd_data_needs_initialising = FALSE;
	}
	else
		ci_initialise_code(bd->bd_code_id, FALSE);

	res = ci_execute_code(bd->bd_code_id, fp, ap, dread, dwrite,
							call_target_function);

	if (res != CI_ER_TRAP && res != STOP)
		errf("%s in breakpoint code", ci_exec_result_to_string(res));

	return res != CI_ER_TRAP;
}
bool
execute_bp_code(bp, fp, ap)
breakpoint_t bp;
taddr_t fp, ap;
{
	objid_t obj;
	bpdesc_t *bd;

 	if (!global_bp_enabled(0, 0))
 		return FALSE;
 
	if ((obj = (objid_t)get_breakpoint_data(bp)) == NULL)
		return TRUE;

	bd = (bpdesc_t *)obj;
	return (execute_bp_code_internal(bd, fp, ap, 0));
}

static int
set_initialise_data_flag(obj, unused_arg)
objid_t obj;
fval_t unused_arg;
{
	((bpdesc_t *)obj)->bd_data_needs_initialising = TRUE;
	return 0;
}

void
reinitialise_bpt_code_data()
{
	visit_objects(BPHEAD_OBJCODE, OBJ_CHILDREN, set_initialise_data_flag,
						(fval_t)0, (objid_t *)NULL);
}

void
remove_breakpoint_object(obj)
objid_t obj;
{
	bpdesc_t *bd;

	bd = (bpdesc_t *)obj;

	if (bd->bd_editblock_id != NULL)
		remove_editblock(bd->bd_editblock_id, DONT_CALL_CALLBACK);

	if (bd->bd_breakpoint != NULL)
		remove_breakpoint(bd->bd_breakpoint);

	if (bd->bd_parse_id != NULL)
		ci_free_parse_id(bd->bd_parse_id);
	if (bd->bd_code_id != NULL)
		ci_free_code_id(bd->bd_code_id);
	
	free((char *)bd);
}

static void
show_bpt_source(bd, beep_on_failure)
bpdesc_t *bd;
bool beep_on_failure;
{
	if (bd->bd_lnum != 0 && bd->bd_func->fu_fil != NULL)
		show_source(bd->bd_func->fu_fil, bd->bd_lnum, TRUE, FALSE);
	else
	if (bd->bd_func)
	{
	  errf("%s`%s' was not compiled with the -g flag",
	       beep_on_failure ? "" : "\bNote: ",
	       bd->bd_func->fu_name);
	}
}

static void
set_bpt_state(obj, disable)
objid_t obj;
bool disable;
{
	fval_t fields[FN_BPT_LAST + 1];
	bpdesc_t *bd;

	if (disable)
	  errf("\bSelected breakpoint code will be ignored");
	else
	  if (global_bp_enabled(0, 0))
	    errf("\bSelected breakpoint code will be executed");
	  else
	    errf("\bSelected breakpoint code will be executed when globally enabled");
	bd = (bpdesc_t *)obj;
	bd->bd_disabled = disable;
/*	if (disable)
	  uninstall_breakpoint(bd->bd_breakpoint);*/
	get_all_fields(obj, fields);
	fields[FN_BPT_STATE] = disable ? (fval_t)"inactive" : (fval_t)"active";
	fields[FN_BPT_LAST] = BPT_FLIST_TERMINATOR;
	set_all_fields((objid_t)bd, fields, BPT_FLIST_TERMINATOR);
}

char *
bpt_format_obj(code)
objid_t code;
{
	const char *fname;
	int lnum;

	fname = get_field_value(code, FN_BPT_FNAME);
	lnum = (int)get_field_value(code, FN_BPT_LNUM);

	return strf("%s %d", fname, lnum);
}

int
pre_do_bpt(command, p_arg)
int command;
char **p_arg;
{
	bptargs_t *ba;
	const char *mode, *prompt;
	FILE *fp;
	char *path;
	
	if (command != MR_BPTMEN_SAVE) {
		*p_arg = NULL;
		return 0;
	}

	if (get_num_selected() == 1)
		prompt = "Save breakpoint to file: ";
	else
		prompt = "Save breakpoints to file: ";
	
	if (prompt_for_output_file(prompt, &path, &mode) != 0)
		return -1;

	if ((fp = fopen(path, mode)) == NULL) {
		errf("Can't create", "saved breakpoints file", path);
		return -1;
	}

	ba = (bptargs_t *)e_malloc(sizeof(bptargs_t));
	ba->ba_fp = fp;
	ba->ba_path = path;
	*p_arg = (char *)ba;
	return 0;
}

void
post_do_bpt(command, arg)
int command;
char *arg;
{
	bptargs_t *ba;
	
	if (command != MR_BPTMEN_SAVE)
		return;

	ba = (bptargs_t *)arg;

	if (fflush(ba->ba_fp) == EOF || ferror(ba->ba_fp)) {
		errf("Warning: Error writing to", "saved breakpoints file",
			 ba->ba_path);
	}

	fclose(ba->ba_fp);

	free(ba->ba_path);
	free((char *)ba);
}

int
do_bpt2(obj, command, arg)
objid_t obj;
int command;
char *arg;
{
	bptargs_t *ba;
	int oldstate, ret = 0;
	objid_t obj1 = NULL;
	bpdesc_t *bd;
	
	ba = (bptargs_t *)arg;

	oldstate = td_set_obj_updating(OBJ_UPDATING_OFF);
	bd = (bpdesc_t *)obj;
	if (command != MR_BPTMEN_SOURCE && bd->bd_func->fu_fil &&
	    (bd->bd_func->fu_fil->fi_flags & FI_DUPLICATE))
	{
	  fil_t *fil1;
				    
	  if (find_duplicate_file(bd->bd_func->fu_fil, &fil1))
	    obj1 = (objid_t)get_editblock_in_tracking_file
	      (bd->bd_editblock_id, fil1);
	}
	switch (command) {
	case MR_BPTMEN_SOURCE:
		show_bpt_source((bpdesc_t *)obj, TRUE);
		break;
	case MR_BPTMEN_REMOVE:
		remove_object(obj, OBJ_SELF);
		if (obj1)
		{
		  remove_object(obj1, OBJ_SELF);
		  ret = 1;	/* RGA redo selection list! */
		}
		break;
	case MR_BPTMEN_SAVE:
		save_object_to_file(ba->ba_path, ba->ba_fp, (bpdesc_t *)obj);
		break;
	case MR_BPTMEN_ENABLE:
		set_bpt_state(obj, FALSE);
		if (obj1)
		  set_bpt_state(obj1, FALSE);
		break;
	case MR_BPTMEN_DISABLE:
		set_bpt_state(obj, TRUE);
		if (obj1)
		  set_bpt_state(obj1, TRUE);
		break;
	case MR_BPTMEN_EXECUTE:
		if (Current_proc)
		{
		  execute_bp_code_internal((bpdesc_t *)obj,
					   proc_getreg(Current_proc, REG_FP),
					   proc_getreg(Current_proc, REG_AP), 1);
		  refresh_target_display(Current_proc, SR_SSTEP, FALSE);
		}
		else
		  errf("Target must be running to execute breakpoint code");
		break;
	default:
		panic("bad rv in db");
	}
	td_set_obj_updating(oldstate);
	return ret;
}

/*  Process the return from the breakpoint menu enabling and disabling
 *  ACTIVE, RESTART etc. and displaying source.
 */
void
do_bpt(obj, command, arg)
objid_t obj;
int command;
char *arg;
{
  do_bpt2(obj, command, arg);
}

void
remove_symtab_breakpoints(symtab_id)
symtab_id_t symtab_id;
{
	objid_t obj, next;
	int oldstate;

	oldstate = td_set_obj_updating(OBJ_UPDATING_OFF);
	for (obj = get_code(BPHEAD_OBJCODE, OBJ_CHILD); obj != NULL; obj = next) {
		next = get_code(obj, OBJ_NEXT);
		if ((((bpdesc_t *)obj)->bd_func)->fu_symtab_id == symtab_id)
			remove_object(obj, OBJ_SELF);
	}
	td_set_obj_updating(oldstate);
}

/*  Add the breakpoint header to the display.  Called on startup from main().
 */
void
add_breakpoint_header(par)
objid_t par;
{
	new_object(BPHEAD_OBJCODE, OT_BPHEAD, par, OBJ_CHILD);
}

/*  BUG - Main_func is used directly in stack.c
 */
func_t *Main_func;

void
set_main_func(f)
func_t *f;
{
	Main_func = f;
}

/*  Process the return from the breakpoint header menu, creating new
 *  breakpoints and such.
 */
/* ARGSUSED */
void
do_bps(unused_obj, command, arg)
objid_t unused_obj;
int command;
char *arg;
{
	bool junk;
	
	switch(command) {
	case MR_ADD_BREAKPOINT:
		add_breakpoint_object((func_t *)NULL, (fil_t *)NULL, 0);
		break;
	case MR_REMOVE_ALL_BREAKPOINTS:
		remove_object(BPHEAD_OBJCODE, OBJ_CHILDREN);
		break;
 	case MR_RESTORE_BREAKPOINTS:
		if (Statefile_path == NULL)
			errf("No saved state file to load breakpoints from");
		else
			read_config_file(Statefile_path, TRUE, TRUE, TRUE,
					 FALSE, &junk);
		break;
	case MR_LOAD_BREAKPOINTS:
		load_breakpoints();
		break;
       case MR_ENABLE_ALL_BREAKPOINTS:
		global_bp_enabled(1, 0); /* set */
		set_bphead_menu_state(1);
		errf("\bGlobally enable all active breakpoints");
		break;
	case MR_DISABLE_ALL_BREAKPOINTS:
		global_bp_enabled(0, 1); /* reset */
		set_bphead_menu_state(0);
		errf("\bGlobally disable all breakpoints");
		break;
	default:
		panic("bad cmd in db");
	}
}

#define LNUM_EDIT	EDF_USER1

static int
bpfunc_quitfunc(edesc, n_tries)
struct edescst *edesc;
int n_tries;
{
	objid_t obj;
	bool edit_lnum;
	int ret;
	
	edit_lnum = (edesc->ed_flags & LNUM_EDIT) != 0;

	if (message_on_null_name(edesc, edit_lnum ? "line number"
						  : "function name") == -1)
		return force_quit(n_tries);

	if (strcmp(edesc->ed_copy, edesc->ed_orig) == 0)
		return EDR_CONFIRM_NO_CHANGE;
	obj = (objid_t) edesc->ed_user;
	if (edesc->ed_orig[0] && !edit_lnum)
	  edit_lnum = -1;
	ret = move_bpt(obj, edesc->ed_copy, edit_lnum);
	if (ret == -1)
	  return force_quit(n_tries);
	else
	  if (ret == -2)
	    return EDR_CANT_QUIT_EXP;
	  else
	  return EDR_CONFIRM_CHANGE;
      }

static void
bplnum_edit(fdets)
struct drawst fdets;
{
	char buf[20];
	struct edescst edescbuf;
	int res;

	bpdesc_t *bd;

	bd = (bpdesc_t *)fdets.dr_code;

	if (HAS_CODE(bd)) {
		errf("Can't move a breakpoint that has interpreted code");
		wn_wait_for_release_of(fdets.dr_wn, B_ANY);
		return;
	}

	clear_selection();
	(void) sprintf(buf, "%d", (int)fdets.dr_fval);

	make_edesc(&edescbuf, fdets.dr_wn, buf, 5, fdets.dr_fg, fdets.dr_bg);
	edescbuf.ed_user = (int)bd;
	edescbuf.ed_keyfunc = (ed_keyfunc_t)num_filter;
	edescbuf.ed_quitfunc = (ed_quitfunc_t)bpfunc_quitfunc;
	edescbuf.ed_flags |= LNUM_EDIT;
	edescbuf.ed_char = ' ';
	res = suppress_ta_cursor_then_edit_field(&edescbuf, "lnum", 0, 0);

	if (res == EDR_CONFIRM_CHANGE)
		show_bpt_source(bd, FALSE);
}

static void
bpfunc_edit(fdets)
struct drawst fdets;
{
	func_t *f, *f1;
	struct edescst edescbuf;
	bpdesc_t *bd;
	int res;

	bd = (bpdesc_t *)fdets.dr_code;

	if (HAS_CODE(bd)) {
		errf("Can't move a breakpoint that has interpreted code");
		wn_wait_for_release_of(fdets.dr_wn, B_ANY);
		return;
	}

	clear_selection();
	make_edesc(&edescbuf, fdets.dr_wn, fdets.dr_fval,
					LEN_BP_FNAME, fdets.dr_fg, fdets.dr_bg);

	edescbuf.ed_user = (int)bd;
	edescbuf.ed_quitfunc = (ed_quitfunc_t)bpfunc_quitfunc;
	edescbuf.ed_char = ' ';

	/*  BUG: must think of a better way to do this.
	 *  BUG: must document what is going on here.
	 */
	if (td_have_window() && fdets.dr_fval[0] == '\0') {
		edescbuf.ed_meaning = EDM_SETCURSOR;
		edescbuf.ed_newpos = 0;
		do_edit(&edescbuf, EDM_SETCURSOR);
	}

	res = suppress_ta_cursor_then_edit_field(&edescbuf, "function-name", 1, 1);
	if (res == EDR_CONFIRM_CHANGE)
	  if (!bd->bd_func)	/* multi line paste perhaps ? */
	  {
	    if (find_func_by_name(edescbuf.ed_copy, &f, &f1, 0) != 0)
	    {
	      free(edescbuf.ed_copy);
	      return;
	    }
	  }
	free(edescbuf.ed_copy);
	show_bpt_source(bd, FALSE);
}

int breakpoint_is_disabled(bd)
	bpdesc_t *bd;
{
  if (!global_bp_enabled(0, 0) || bd->bd_disabled)
    return 1;
  else
    return 0;
}
