/* st_lookup.c - look up things by name or address in the symbol tables */

/*  Copyright 1992 Mark Russell, University of Kent at Canterbury.
 *
 *  You can do what you like with this source code as long as
 *  you don't try to make money out of it and you include an
 *  unaltered copy of this message (including the copyright).
 */

char ups_st_lookup_c_sccsid[] = "@(#)st_lookup.c	1.2 15/9/92 (UKC)";

#include <mtrprog/ifdefs.h>

#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <string.h>
#include <a.out.h>		/* needed for st_priv.h */
#include <stdlib.h>
#include <ctype.h>

#include <local/wn.h>
#include <local/ukcprog.h>
#include <local/arg.h>
#include <local/obj/obj.h>
#include <mtrprog/strcache.h>
#include <mtrprog/alloc.h>
#include <mtrprog/genmergesort.h>
#include <mtrprog/so.h>

#include "ups.h"
#include "symtab.h"
#include "ci.h"
#ifdef ARCH_MIPS
#include "mips_frame.h"
#endif
#include "st_priv.h"
#include "printf.h"

#define CMP_STRCMP 0
#define CMP_STRNCMP 1
#define CMP_STRSTR 2

/*  List of globals added to by find_matching_funcs() and
 *  find_matching_globals().
 */
typedef struct global_list_s {
	const char *gl_name;
	fil_t *gl_fil;
	func_t *gl_func;
	var_t *gl_var;
	taddr_t gl_addr;
	const char *st_name;
	struct global_list_s *gl_next;
} global_list_t;

static bool filename_matches PROTO((const char *st_name,
				    fil_t *fil, fil_t *matchfil,
				    char *name, int namelen,
				    int cmp_style));
static void find_matching_funcs PROTO((const char *st_name,
				       func_t *funclist,
				       fil_t *matchfil, const char *filename,
				       int flen, const char *funcname,
				       global_list_t **p_glhead,
				       bool *p_found_file));
static void find_matching_funcs_only PROTO((const char *st_name,
					    func_t *funclist,
					    fil_t *matchfil, const char *filename,
					    int flen, const char *funcname,
					    func_bp_list_t **funchead,
					    int cmp_style));
static void find_matching_globals PROTO((symtab_t *st, fil_t *matchfil, const char *filename, int filename_len, const char *pat, global_list_t **p_glhead));
static void scan_symtabs PROTO((fil_t *fil, const char *filename, int filename_len, const char *funcpat, bool want_vars, global_list_t **p_glhead, bool *p_found_file, bool all_syms));
static void add_var_to_global_list PROTO((global_list_t **p_glhead,
					  var_t *v, fil_t *fil,
					  const char *st_name));
static int glcmp PROTO((global_list_t *f1, global_list_t *f2));
static void add_common_block_globals PROTO((const char *name, func_t *f,
					    common_block_id_t *p_cblock,
					    global_list_t **p_glhead));
static void dupf_mesg PROTO((const char *name,
			     global_list_t *gl1, global_list_t *gl2));

GENERIC_MERGE_SORT(static,sort_global_list,global_list_t,gl_next)

ALLOC_NEW_FREELIST(static,global_list_t,global_list,gl_next)

GENERIC_MERGE_SORT(static,sort_func_bp_list,func_bp_list_t,fl_next)

ALLOC_NEW_FREELIST(static,func_bp_list_t,func_bp_list,fl_next)

FILE *popen PROTO((const char *name, const char *mode));

static bool
filename_matches(st_name, fil, matchfil, name, namelen, cmp_style)
const char *st_name;
fil_t *fil, *matchfil;
char *name;
int namelen;
int cmp_style;
{
	int len;
	bool match = FALSE;
	const char *file_name;
	char tmp;

	if (matchfil != NULL)
		return fil == matchfil;

	if (name == NULL)
		return TRUE;
	if (fil == NULL)
	{
	  if (namelen == 0)
	    return FALSE;
	  else
	    file_name = st_name;
	}
	else
	  file_name = fil->fi_name;
	if (namelen == 0)
		return TRUE;

	len = strlen(file_name);

	if (len < namelen)
	  return FALSE;
	else
	{
	  switch (cmp_style)
	  {
	  case CMP_STRCMP:
	    if (memcmp(&file_name[len - namelen], name, namelen) != 0)
	      return FALSE;
	    break;
	  case CMP_STRNCMP:
	    if (memcmp(file_name, name, namelen - 1) != 0)
	      return FALSE;
	    else
	      return TRUE;
	  case CMP_STRSTR:
	    tmp = name[namelen - 1];
	    name[namelen - 1] = 0;
	    match = !!strstr(file_name, name + 1);
	    name[namelen - 1] = tmp;
	    return match;
	  }
	}
	return len == namelen || file_name[len - namelen - 1] == '/';
}

static void
find_matching_funcs(st_name, funclist, matchfil, filename, flen, funcpat,
		    p_glhead, p_found_file)
const char *st_name;
func_t *funclist;
fil_t *matchfil;
const char *filename;
int flen;
const char *funcpat;
global_list_t **p_glhead;
bool *p_found_file;
{
	func_t *f;

	for (f = funclist; f != NULL; f = f->fu_next) {

		if (!filename_matches(st_name, f->fu_fil, matchfil,
				      (char *)filename, flen, CMP_STRCMP))
			continue;

		*p_found_file = TRUE;

		if (arg_match(f->fu_name, funcpat)) {
			global_list_t *gl;

			gl = new_global_list();
			gl->gl_name = f->fu_name;
			gl->gl_fil = f->fu_fil;
			gl->gl_func = f;
			gl->gl_var = NULL;
			gl->gl_addr = f->fu_addr;
			gl->gl_next = *p_glhead;
			gl->st_name = st_name;
			*p_glhead = gl;
		}
	}
}

fil_t *
name_to_fil(name)
const char *name;
{
	symtab_t *st;

	for (st = get_main_st(); st != NULL; st = st->st_next) {
		fil_t *fil;

		for (fil = st->st_sfiles; fil != NULL; fil = fil->fi_next) {
			if (strcmp(fil->fi_name, name) == 0)
				return fil;
		}
	}

	return NULL;
}

func_t *
name_and_fil_to_func(name, fil)
const char *name;
fil_t *fil;
{
	funclist_t *fl;

	for (fl = fil->fi_funclist_head; fl != NULL; fl = fl->fl_next) {
		if (strcmp(fl->fl_func->fu_name, name) == 0)
			return fl->fl_func;
	}

	return NULL;
}

func_t *
name_to_func(name)
const char *name;
{
	symtab_t *st;
	funclist_t *fl;

	for (st = get_main_st(); st != NULL; st = st->st_next)
	{
	  fil_t *fil;
	  
	  for (fil = st->st_sfiles; fil != NULL; fil = fil->fi_next)
	  {
	    for (fl = fil->fi_funclist_head; fl != NULL; fl = fl->fl_next)
	    {
	      if (strcmp(fl->fl_func->fu_name, name) == 0)
		return fl->fl_func;
	    }
	  }
	}

	return NULL;
}
int
resolve_untyped_name(name, p_v)
const char *name;
var_t **p_v;
{
	global_list_t *glhead;

	glhead = NULL;
	find_matching_globals(get_main_st(), (fil_t *)NULL,
			      (const char *)NULL, 0, name, &glhead);
	
	if (glhead != NULL && glhead->gl_next == NULL)
		*p_v = glhead->gl_var;
	else
		*p_v = NULL;
	
	free_global_list_list(glhead);

	return (*p_v != NULL) ? 0 : -1;
}

int
find_duplicate_file(fil, fil1)
     fil_t *fil;
     fil_t **fil1;
{
  symtab_t *st;

  if (fil)
  {
    for (st = get_main_st(); st; st = st->st_next)
      for (*fil1 = st->st_sfiles; *fil1; *fil1 = (*fil1)->fi_next)
      {
	if (fil != *fil1 && ((*fil1)->fi_flags & FI_DUPLICATE) &&
	    !strcmp(fil->fi_name, (*fil1)->fi_name))
	  return 1;
      }
  }
  return 0;
}

int
find_global_name(name)
const char *name;
{
	global_list_t *glhead;
	int ret = 0;

	glhead = NULL;
	find_matching_globals(get_main_st(), (fil_t *)NULL,
			      (const char *)NULL, 0, name, &glhead);
	
	if (glhead)
	  ret = 1;
	free_global_list_list(glhead);

	return ret;
}

static void
find_matching_globals(st, matchfil, filename, filename_len, pat, p_glhead)
symtab_t *st;
fil_t *matchfil;
const char *filename;
int filename_len;
const char *pat;
global_list_t **p_glhead;
{
	fil_t *fil;
	addrlist_t *al;

	for (fil = st->st_sfiles; fil != NULL; fil = fil->fi_next) {
#ifndef ST_TE
		snlist_t *sn;
#endif
		var_t *v;

		if (!filename_matches(st->st_name, fil, matchfil,
				      (char *)filename, filename_len, CMP_STRCMP))
			continue;

#ifndef ST_TE
		sn = ((stf_t *)fil->fi_stf)->stf_sn;
		for (; sn != NULL; sn = sn->sn_next) {
			if (arg_match(sn->sn_name, pat))
				break;
		}
		if (sn == NULL)
			continue;
#endif

		for (v = FI_VARS(fil); v != NULL; v = v->va_next) {
			if (arg_match(v->va_name, pat))
				add_var_to_global_list(p_glhead, v, fil,
						       st->st_name);
		}
	}

	if (matchfil != NULL || filename != NULL)
		return;
	 
	 for (st = get_main_st(); st != NULL; st = st->st_next) {
		for (al = st->st_addrlist; al != NULL; al = al->al_next) {
			if (!arg_match(al->al_name, pat))
				continue;

			if (al->al_var == NULL) {
				var_t *v;

				v = ci_make_var(st->st_alloc_id,
						al->al_name, CL_EXT,
						ci_code_to_type(TY_INT_ASSUMED),
						al->al_addr);
				v->va_language = LANG_UNKNOWN;
				al->al_var = v;
			}

			add_var_to_global_list(p_glhead, al->al_var,
					       (fil_t *)NULL, st->st_name);
		}
	}
}

static void
add_var_to_global_list(p_glhead, v, fil, st_name)
global_list_t **p_glhead;
var_t *v;
fil_t *fil;
const char *st_name;
{
	global_list_t *gl;

	gl = new_global_list();
	gl->gl_name = v->va_name;
	gl->gl_fil = fil;
	gl->gl_var = v;
	gl->gl_func = NULL;
	gl->gl_addr = v->va_addr;
	gl->gl_next = *p_glhead;
	gl->st_name = st_name;
	*p_glhead = gl;
}

/*  Search for the common block where the variable called name
 *  is defined.  We loop over all the symbol tables, searching
 *  the common block lists of each.  (At present there are no
 *  shared library FORTRAN executables, so the only symbol
 *  table with a common block list will be the main one).
 */
static void
add_common_block_globals(name, f, p_cblock, p_glhead)
const char *name;
func_t *f;
common_block_id_t *p_cblock;
global_list_t **p_glhead;
{
	symtab_t *st;

	for (st = get_main_st(); st != NULL; st = st->st_next) {
		var_t *v;
		fil_t *fil;

		global_and_cblist_to_var(st->st_cblist_id, name, f,
							 p_cblock, &fil, &v);
		if (v != NULL) {
			add_var_to_global_list(p_glhead, v, fil, 0);
			return;
		}
	}
}

static int
glcmp(gl1, gl2)
global_list_t *gl1, *gl2;
{
	int res;

	if ((res = strcmp(gl1->gl_name,  gl2->gl_name)) != 0)
		return res;
	
	if ((gl1->gl_func != NULL) != (gl2->gl_func != NULL))
		return (gl1->gl_func == NULL) ? -1 : 1;
	
	if (gl1->gl_func == NULL)
		return 0;

	return addrcmp(gl1->gl_func, gl2->gl_func);
}

static void
scan_symtabs(matchfil, filename, filename_len, funcpat, want_vars,
	     p_glhead, p_found_file, all_syms)
fil_t *matchfil;
const char *filename;
int filename_len;
const char *funcpat;
bool want_vars;
global_list_t **p_glhead;
bool *p_found_file;
bool all_syms;
{
	symtab_t *st;

	for (st = get_main_st(); st != NULL; st = st->st_next) {
		find_matching_funcs(st->st_name, st->st_funclist,
				    matchfil, filename, filename_len,
			 	    funcpat, p_glhead, p_found_file);
	}

	for (st = get_symtab_cache_list(); st != NULL; st = st->st_next) {
		find_matching_funcs(st->st_name, st->st_funclist,
				    matchfil, filename, filename_len,
			 	    funcpat, p_glhead, p_found_file);
	}

	if (want_vars) {	
	  /* RGA search all symbol tables if all_syms is set */
	  for (st = get_main_st(); st != NULL; st = st->st_next)
	  {
	    find_matching_globals(st, matchfil, filename,
				  filename_len, funcpat, p_glhead);
	    if (!all_syms)
	      break;
	  }
	}
}
					
int
find_func_by_name(name, p_f, p_f1, use_include_path)
char *name;
func_t **p_f;
func_t **p_f1;
int use_include_path;
{
	return find_global_by_name(name, (fil_t *)NULL, (func_t *)NULL, FALSE,
				   p_f, p_f1, (var_t **)NULL, (var_t **)NULL,
				   (common_block_id_t *)NULL, (fil_t **)NULL,
				   (fil_t **)NULL, use_include_path);
				
}

int
find_func_by_exact_name(name, p_f, p_f1, use_include_path)
const char *name;
func_t **p_f;
func_t **p_f1;
int use_include_path;
{
	return find_global_by_name((char *)name, (fil_t *)NULL,
				   (func_t *)NULL, TRUE,
				   p_f, p_f1, (var_t **)NULL, (var_t **)NULL,
				   (common_block_id_t *)NULL, (fil_t **)NULL,
				   (fil_t **)NULL, use_include_path);
				
}

static int
  do_include_path_search(cmd)
     const char *cmd;
{
  FILE *fp;
  char buf[256];
  int ret = 0;
  
  if ((fp = popen(cmd, "r")) == NULL)
  {
    errf("Can't run \"%s\" (%m)", cmd);
    exit(1);
  }
  else
  {
    while (fgets(buf, sizeof(buf), fp) != NULL)
    {
      ret++;
      print_string_to_output_win(buf);
    }
    pclose(fp);
  }
  return(ret);
}

/*  We search all the symbol tables including those in the cache.
 *  This is so we can match names in shared libraries (like printf)
 *  before the target has been started.
 */
int
find_global_by_name(name, fil, f, exact, p_f, p_f1, p_v, p_v1, p_cblock, p_fil,
		    p_fil1, use_include_path)
char *name;
fil_t *fil;
func_t *f;
bool exact;
func_t **p_f, **p_f1;
var_t **p_v, **p_v1;
common_block_id_t *p_cblock;
fil_t **p_fil, **p_fil1;
int use_include_path;
{
	global_list_t *gl, *gl1 = NULL, *glhead, *prev;
	global_list_t *exact_gls[2], *partial_gls[2], *exact_extern_gl;
	int matches, exact_matches, exact_extern_matches, partial_matches;
	int filename_len;
	const char *filename, *funcpat;
	char *funcname, *infile;
	bool found_file;
	char c, c1;
	int ok = 1, ret = -1;
	static const char *file_type = "/*.{h,hh,H}";

	if (p_f1)
	  *p_f1 = NULL;
	if (p_v1)
	  *p_v1 = NULL;
	if (p_fil1)
	  *p_fil1 = NULL;
	
	if ((funcname = strchr(name, ':')) != NULL && funcname[1] != ':') {
		if (funcname[1] == '\0') {
			errf("Missing function name after source file");
			return -1;
		}
		
		filename = name;
		filename_len = funcname - name; 
		++funcname;
	}
	else {
		filename = NULL;
		filename_len = 0;	/* to satisfy gcc */
		funcname = (char *)name;
	}

	if (exact) {
		funcpat = funcname;
	}
	else {
		static const char globchars[] = "*?[]";
		const char *s, *globname;
		char *namecopy;

		for (s = funcname; s != '\0'; ++s) {
			if (strchr(globchars, *s) != NULL)
				break;
		}
		
		if (*s == '\0') {
			namecopy = strf("%s*", funcname);
			globname = namecopy;
		}
		else {
			namecopy = NULL;
			globname = funcname;
		}

		funcpat = arg_do_quoting(&globname, globchars);

		if (namecopy != NULL)
			free(namecopy);

		if (funcpat == NULL)
			return -1;
	}

	glhead = NULL;
	found_file = FALSE;

	if (p_v != NULL)
		*p_cblock = NULL;

	if (fil != NULL && filename == NULL)
		scan_symtabs(fil, (const char *)NULL, 0, funcpat,
					p_v != NULL, &glhead, &found_file, 1);

	if (glhead == NULL) {
		scan_symtabs((fil_t *)NULL, filename, filename_len, funcpat,
					p_v != NULL, &glhead, &found_file, 0);

		if (p_v != NULL)
			add_common_block_globals(name, f, p_cblock, &glhead);

		/* RGA try and find the real instance of a global */
		/* and not its external reference in other files */
		if (fil != NULL && filename == NULL)
		  for (gl = glhead; gl != NULL; gl = gl->gl_next)
		    if (gl->gl_fil)
		    {
		      global_list_t *glhead1;

		      glhead1 = NULL;
		      scan_symtabs(gl->gl_fil, (const char *)NULL, 0, funcpat,
				   p_v != NULL, &glhead1, &found_file, 0);
		      if (glhead1 != NULL)
		      {
			free_global_list_list(glhead);
			glhead = glhead1;
		      }
		      break;
		    }
	}

	if (glhead == NULL && fil != NULL && filename == NULL)
		scan_symtabs(fil, (const char *)NULL, 0, funcpat,
					p_v != NULL, &glhead, &found_file, 1);


	/* RGA fix for source files that contain global data but no */
	/* functions - glhead is non-null, but found_file is 0 */

	if (filename && !found_file) {
		errf("No source file name `%.*s'", filename_len, filename);
		return -1;
	}

	if (glhead == NULL) {
		const char *orvar;

		orvar = (p_v != NULL) ? " or variable" : "";
		
		if (filename != NULL)
		  errf("No function%s matching `%s' in %.*s",
		       orvar, funcname, filename_len, filename);
		else
		{
		  const char *cptr;
		  int count = 0;

		  for (cptr = funcname; *cptr != '\0'; cptr++)
		    if (!(isalnum(*cptr) || (*cptr) == '_'
			|| (*cptr) == '.' || (*cptr) == '*'))
				/* allow widcards */
		    {
		      errf("No function%s `%s'", orvar, funcname);
		      return -1;
		    }
	  
		  if (use_include_path && get_current_srcwin_region())
		  {
		    char buff[80], str[256], *paths, *func_copy = NULL;
		    int old_state;
		    pathlist_t *pl, *pl_root;

		    pl_root = get_source_path_root();
		    if (pl_root)
		    {
		      strcpy(buff, funcname);
		      func_copy = malloc(strlen(funcname) + 1);
		      strcpy(func_copy, funcname);
		      errf("No function%s `%s', searching `use' paths in init file...",
			   orvar, func_copy);
		      old_state = updating(OBJ_UPDATING_QUERY);
		      if (old_state == OBJ_UPDATING_OFF)
			updating(OBJ_UPDATING_ON);
		      open_output_win();
		      updating(old_state);
		      display_win_search(1, 0);	/* set */
		    }
		    for (pl = pl_root; pl != NULL;)
		    {
		      if (pl->pl_next && pl->pl_next->pl_next)
		      {
			paths = (char *)malloc
			  (strlen(pl->pl_path) +
			   strlen(pl->pl_next->pl_path) +
			   strlen(pl->pl_next->pl_next->pl_path) +
			   3*strlen(file_type) + 3);
			sprintf(paths, "%s%s %s%s %s%s",
				pl->pl_path, file_type,
				pl->pl_next->pl_path, file_type,
				pl->pl_next->pl_next->pl_path, file_type);
			pl = pl->pl_next->pl_next->pl_next;
		      }
		      else
			if (pl->pl_next)
			{
			  paths = (char *)malloc
			    (strlen(pl->pl_path) +
			     strlen(pl->pl_next->pl_path) +
			     2 * strlen(file_type) + 2);
			  sprintf(paths, "%s%s %s%s",
				  pl->pl_path, file_type,
				  pl->pl_next->pl_path, file_type);
			  pl = pl->pl_next->pl_next;
			}
			else
			{
			  paths = (char *)malloc
			    (strlen(pl->pl_path) +
			     strlen(file_type) + 1);
			  sprintf(paths, "%s%s",
				  pl->pl_path, file_type);
			  pl = pl->pl_next;
			}
		      /* look for the `defined' symbol surrounded */
		      /* by white space */
		      sprintf
			(str,
			 "csh -c \"egrep '#[\t ]*define[\t ][\t ]*%s[\t (]' %s /dev/null\"", 
			 buff, paths);
		      free(paths);
		      count = do_include_path_search(str);
		      if (count)
			break;	/* bail out if text found in a sub path */
		    }
		    if (func_copy)
		      free(func_copy);
		    display_win_search(0, 1);	/* reset */
		    if (count == 1)
		      errf("\b1 match found in `use' paths");
		    else
		      errf("\b%d matches found in `use' paths", count);
		  }
		  else
		    errf("No function%s `%s'", orvar, funcname);
		}
		return -1;
	      }
	
	matches = exact_matches = exact_extern_matches = partial_matches = 0;
	exact_extern_gl = NULL;		/* to satisfy gcc */

	for (gl = glhead; gl != NULL; gl = gl->gl_next)
		++matches;
	glhead = sort_global_list(glhead, matches, glcmp);

	prev = NULL;

	for (gl = glhead; gl != NULL; gl = gl->gl_next) {

		/*  We can get duplicate symbol table entries for a
		 *  function or variable - ignore these.
		 */
		if (prev != NULL && gl->gl_addr == prev->gl_addr &&
		    strcmp(prev->gl_name, gl->gl_name) == 0 &&
		    (prev->gl_func != NULL) == (gl->gl_func != NULL))
			continue;
		
		if (strcmp(gl->gl_name, funcname) == 0) {
			if ((gl->gl_func != NULL) ?
				      (gl->gl_func->fu_flags & FU_STATIC) == 0 :
				      gl->gl_var->va_class == CL_EXT) {
				++exact_extern_matches;
				exact_extern_gl = gl;
			}

			if (exact_matches < 2)
				exact_gls[exact_matches] = gl;
			++exact_matches;
		}
		else {
			if (partial_matches < 2)
				partial_gls[partial_matches] = gl;
			++partial_matches;
		}

		prev = gl;
	}
	if (exact_matches == 1)
		gl = exact_gls[0];
	else if (partial_matches == 1 && exact_matches == 0)
		gl = partial_gls[0];
	else if (exact_extern_matches == 1)
		gl = exact_extern_gl;
	else
		gl = NULL;

	/* RGA fix for multiple "main" functions, where only second */
	/* has file info */
/*	if (gl && !gl->gl_fil && exact_matches > 1 &&
	    exact_gls[0]->gl_func != NULL &&
	    exact_gls[1]->gl_func != NULL)
	  if (!exact_gls[0]->gl_fil && exact_gls[1]->gl_fil)
	    gl->gl_fil = exact_gls[1]->gl_fil;
	  else
	    if (exact_gls[0]->gl_fil && !exact_gls[1]->gl_fil)
	      gl->gl_fil = exact_gls[0]->gl_fil;
	
	if (gl && !gl->gl_fil && partial_matches > 1 &&
	    partial_gls[0]->gl_func != NULL &&
	    partial_gls[1]->gl_func != NULL)
	  if (!partial_gls[0]->gl_fil && partial_gls[1]->gl_fil)
	    gl->gl_fil = partial_gls[1]->gl_fil;
	  else
	    if (partial_gls[0]->gl_fil && !partial_gls[1]->gl_fil)
	      gl->gl_fil = partial_gls[0]->gl_fil;*/
	
	/* RGA when duplicate function names exist, and the files are */
	/* the same, just use the first one found. This can happen when */
	/* a function is linked statically ahead of a dynamic library */
	/* where the same function is also defined. Then the link order */
	/* takes precedence */

	if (exact_matches > 1 &&
	    exact_gls[0]->gl_func != NULL &&
	    exact_gls[1]->gl_func != NULL &&
	    exact_gls[0]->gl_fil && exact_gls[1]->gl_fil &&
	    !strcmp(exact_gls[0]->gl_fil->fi_name,
		    exact_gls[1]->gl_fil->fi_name) &&
	    !strcmp(exact_gls[0]->gl_fil->fi_path_hint,
		    exact_gls[1]->gl_fil->fi_path_hint))
	{
	    gl = exact_gls[0];
	    gl1 = exact_gls[1];
	}
	
	if (partial_matches > 1 &&
	    partial_gls[0]->gl_func != NULL &&
	    partial_gls[1]->gl_func != NULL &&
	    !strcmp(partial_gls[0]->gl_name, partial_gls[1]->gl_name) &&
	    partial_gls[0]->gl_fil && partial_gls[1]->gl_fil &&
	    !strcmp(partial_gls[0]->gl_fil->fi_name,
		    partial_gls[1]->gl_fil->fi_name) &&
	    !strcmp(partial_gls[0]->gl_fil->fi_path_hint,
		    partial_gls[1]->gl_fil->fi_path_hint))
	{
	    gl = partial_gls[0];
	    gl1 = partial_gls[1];
	}

	if (exact_matches > 1)
	{
	  if (exact)
	    dupf_mesg(funcname, exact_gls[0], exact_gls[1]);
	  else
	    errf("\bUsing first one of multiple/overloaded symbols...");
	  for (gl = glhead; gl != NULL; gl = gl->gl_next) 
	    if (gl->gl_func &&
		!((gl->gl_func->fu_flags & FU_HEADER) ||
		  (gl->gl_func->fu_flags & FU_STATIC)))
	      break;
	  gl = gl ? gl : exact_gls[0];
	  gl1 = exact_gls[1];
	}

	if (gl != NULL) {
		*p_f = gl->gl_func;
		if (p_v != NULL) {
			*p_v = gl->gl_var;
			*p_fil = gl->gl_fil;
		}
		if (gl1 != NULL)
		{
		  *p_f1 = gl1->gl_func;
		  if (p_v1 != NULL)
		  {
		    *p_v1 = gl1->gl_var;
		    *p_fil1 = gl1->gl_fil;
		  }
		}
		free_global_list_list(glhead);
		return 0;
	}

	if (partial_matches <= 1)
		panic("pmatch botch in ffbn");

	if (strcmp(partial_gls[0]->gl_name, partial_gls[1]->gl_name) == 0) {
		dupf_mesg(partial_gls[0]->gl_name,
				partial_gls[0], partial_gls[1]);
		free_global_list_list(glhead);
		return -1;
	}

	infile = (filename != NULL && *filename != '\0')
				? strf(" in %.*s", filename_len, filename)
				: strsave("");

	while(ok)
	{
	  c = glhead->gl_name[strlen(funcname)];
	  if (!c)
	    ok = 0;
	  for (gl = glhead; c && ok && gl != NULL; gl = gl->gl_next)
	  {
	    c1 = gl->gl_name[strlen(funcname)];
	    if (!c1 || c1 != c)
	      ok = 0;
	  }
	  if (ok)
	  {
	    ret = -2;
	    funcname[strlen(funcname) + 1] = 0;
	    funcname[strlen(funcname)] = c;
	  }
	}
	if (partial_matches == 2) {
		errf("`%s' matches both `%s' and `%s'%s",
				funcname, partial_gls[0]->gl_name,
				partial_gls[1]->gl_name, infile);
	}
	else {
		errf("`%s' matches `%s', `%s' (and %d more)%s",
				funcname, partial_gls[0]->gl_name,
				partial_gls[1]->gl_name, 
				partial_matches - 2, infile);
	}
	if (display_win_search(0, 0)) /* dump matches to output win */
	{
	  char buff[256];

	  for (gl = glhead; gl != NULL; gl = gl->gl_next)
	  {
	    if (gl->gl_fil)
	      if (gl->gl_fil->fi_source_path
		  [strlen(gl->gl_fil->fi_source_path)-1] == '/')
		sprintf(buff, "%s%s:\t%s\n",
			gl->gl_fil->fi_source_path,
			gl->gl_fil->fi_name,
			gl->gl_name);
	      else
		sprintf(buff, "%s/%s:\t%s\n",
			gl->gl_fil->fi_source_path,
			gl->gl_fil->fi_name,
			gl->gl_name);
	    else
	      if (gl->st_name)
		sprintf(buff, "%s:\t%s\n",gl->st_name, gl->gl_name);
	      else
		sprintf(buff, "%s\n",gl->gl_name);
	    print_string_to_output_win(buff);
	  }
	}
	free_global_list_list(glhead);

	free(infile);
	return (ret);
}

static void
dupf_mesg(funcname, gl1, gl2)
const char *funcname;
global_list_t *gl1, *gl2;
{
	const char *what;
	fil_t *fil1, *fil2;

	if ((gl1->gl_func != NULL) != (gl2->gl_func != NULL))
		what = "Name";
	else
		what = (gl1->gl_func != NULL) ? "Function" : "Variable";

	fil1 = gl1->gl_fil;
	fil2 = gl2->gl_fil;

	if (fil1 != NULL && fil2 != NULL)
		errf("%s `%s' appears in both %s and %s",
				what, funcname, fil1->fi_name, fil2->fi_name);
	else if (fil1 != NULL && fil2 == NULL)
	  if (gl2->st_name)
	    errf("%s `%s' appears in both %s and in library %s",
		 what, funcname, fil1->fi_name, gl2->st_name);
	  else
	    errf("%s `%s' appears in both %s and elswhere",
		 what, funcname, fil1->fi_name);
	
	else if (fil2 != NULL && fil1 == NULL)
	  if (gl1->st_name)
		errf("%s `%s' appears in both %s and in library %s",
				what, funcname, fil2->fi_name, gl1->st_name);
	  else
	    errf("%s `%s' appears in both %s and elsewhere",
		 what, funcname, fil2->fi_name);
	else
	  if (gl1->st_name && gl2->st_name)
		errf("%s `%s' appears in libraries %s and %s",
		     what, funcname, gl1->st_name, gl2->st_name);
	else
		errf("%s `%s' appears more than once!", what, funcname);
}

/*  Return a pointer to the func_t block of the function wherein
 *  lies address addr, or NULL if addr does not lie within any
 *  known function.  We search all the symbol tables.
 */
func_t *
addr_to_func(addr)
taddr_t addr;
{
	symtab_t *st;
	func_t *f;

	for (st = get_main_st(); st != NULL; st = st->st_next) {
		f = addr_and_functab_to_func(st->st_functab_id, addr);
		if (f != NULL)
			return f;
	}

	return NULL;
}

/*  Call (*func)() for each source file with the source file as an argument.
 *
 *  Do the source files in the cache as well.
 */
void
iterate_over_source_files(func)
void (*func)PROTO((fil_t *fil));
{
	symtab_t *st;
	fil_t *fil;

	for (st = get_main_st(); st != NULL; st = st->st_next)
		for (fil = st->st_sfiles; fil != NULL; fil = fil->fi_next)
			if (((((stf_t *)fil->fi_stf)->stf_flags & STF_HIDE) == 0))
			  (*func)(fil);
	for (st = get_symtab_cache_list(); st != NULL; st = st->st_next)
		for (fil = st->st_sfiles; fil != NULL; fil = fil->fi_next)
			if (((((stf_t *)fil->fi_stf)->stf_flags & STF_HIDE) == 0))
			  (*func)(fil);
}

/*  Call (*func)(cblock) for each common block.
 */
void
iterate_over_common_blocks(func)
void (*func)PROTO((common_block_id_t cblock));
{
	symtab_t *st;

	for (st = get_main_st(); st != NULL; st = st->st_next)
		iterate_over_cblist(st->st_cblist_id, func);
}

static void
find_matching_funcs_only(st_name, funclist, matchfil, filename,
			 flen, funcpat, funchead, cmp_style)
const char *st_name;
func_t *funclist;
fil_t *matchfil;
const char *filename;
int flen;
const char *funcpat;
func_bp_list_t **funchead;
int cmp_style;
{
	func_t *f;

	for (f = funclist; f != NULL; f = f->fu_next) {

	  if (!filename_matches(st_name, f->fu_fil, matchfil,
				(char *)filename, flen, cmp_style))
	    continue;

	  if (arg_match(f->fu_name, funcpat)) {
			func_bp_list_t *fl;

			fl = new_func_bp_list();
			fl->fl_func = f;
			fl->fl_bp = 0;
			fl->fl_next = *funchead;
			*funchead = fl;
		}
	}
}

static int
flcmp(fl1, fl2)
func_bp_list_t *fl1, *fl2;
{
	int res;

	if ((res = strcmp(fl1->fl_func->fu_name,  fl2->fl_func->fu_name)) != 0)
		return res;
	
	if ((fl1->fl_func != NULL) != (fl2->fl_func != NULL))
		return (fl1->fl_func == NULL) ? -1 : 1;
	
	if (fl1->fl_func == NULL)
		return 0;

	return addrcmp(fl1->fl_func, fl2->fl_func);
/*	return strcmp(fl1->fl_func->fu_name, fl2->fl_func->fu_name);*/
}

void
get_matching_func_bp_list(name, funchead)
const char *name;
func_bp_list_t **funchead;
{
  symtab_t *st;
  static const char globchars[] = "*?[]";
  const char *s, *globname, *filename, *funcpat;
  char *funcname, *namecopy;
  int matches, filename_len, cmp_style = CMP_STRCMP;
  func_bp_list_t *fl;

  if ((funcname = strchr(name, ':')) != NULL && funcname[1] != ':') {
    if (funcname[1] == '\0') {
      errf("Missing function name after source file");
      return;
    }
		
    filename = name;
    filename_len = funcname - name;
    if (filename[0]  == '*')
      cmp_style = CMP_STRSTR;
    else
      if (filename[filename_len - 1]  == '*')
	cmp_style = CMP_STRNCMP;
    ++funcname;
  }
  else {
    filename = NULL;
    filename_len = 0;		/* to satisfy gcc */
    funcname = (char *)name;
  }
  
  for (s = funcname; s != '\0'; ++s) {
    if (strchr(globchars, *s) != NULL)
      break;
  }
  
  if (*s == '\0') {
    namecopy = strf("%s*", funcname);
    globname = namecopy;
  }
  else {
    namecopy = NULL;
    globname = funcname;
  }
  
  funcpat = arg_do_quoting(&globname, globchars);
  
  if (namecopy != NULL)
    free(namecopy);
  
  if (funcpat == NULL)
    return ;
    
  for (st = get_main_st(); st != NULL; st = st->st_next) 
    find_matching_funcs_only(st->st_name, st->st_funclist,
			     (fil_t *)NULL, filename, filename_len,
			     funcpat, funchead, cmp_style);
  
  for (st = get_symtab_cache_list(); st != NULL; st = st->st_next)
    find_matching_funcs_only(st->st_name, st->st_funclist,
			     (fil_t *)NULL, filename, filename_len,
			     funcpat, funchead, cmp_style);

				/* sort list so duplicates are together */
  matches = 0;
  for (fl = *funchead; fl != NULL; fl = fl->fl_next)
    ++matches;
  *funchead = sort_func_bp_list(*funchead, matches, flcmp);

}

void
do_free_func_bp_list_list(funchead)
  func_bp_list_t *funchead;
{
  free_func_bp_list_list(funchead);
}

const char *
get_target_basename(textname)
const char *textname;
{
	symtab_t *st;

	st = get_main_st();
        if (st->st_sfiles != NULL && st->st_sfiles->fi_next == NULL)
                return st->st_sfiles->fi_name;
        else
                return textname;
}

