/* ui.h - declarations of public user interface routines */

/*  Copyright 1991 Mark Russell, University of Kent at Canterbury.
 *
 *  You can do what you like with this source code as long as
 *  you don't try to make money out of it and you include an
 *  unaltered copy of this message (including the copyright).
 */

/* @(#)ui.h	1.18 13/8/92 (UKC) */

char *get_typing_line_string PROTO((void));

#ifdef REG_H_INCLUDED
region_id_t divide_window_into_regions PROTO((int wn));
#endif

#ifdef EDIT_H_INCLUDED
Edit_display *make_text_display PROTO((int wn, int fg, int bg,
				       const char *text));

char *get_buffer_contents PROTO((Edit_buffer *buffer));

bool search_buffer PROTO((Edit_buffer *buffer, const char *pattern,
			  size_t start_point, bool forwards,
			  size_t *p_start, size_t *p_lim));
#endif

void display_message PROTO((const char *mesg));
int dump_obj PROTO((char *arg, int level, char *line));
void update_target_menu_state PROTO((void));
void add_outwin PROTO((void));

typedef void (*abort_func_t)PROTO((void));
abort_func_t set_user_abort_func PROTO((abort_func_t func));
bool user_wants_stop PROTO((void));

#ifdef SYMTAB_H_INCLUDED
void get_updated_source PROTO((fil_t *fil, int lnum,
			       int old_nlines, int new_nlines));
void show_source PROTO((fil_t *fil, int lnum, bool centre, bool check_stat));
int highlight_source PROTO((fil_t *fil, int lnum));
fil_t *get_displayed_fil PROTO((void));
#endif

#ifdef OBJ_H_INCLUDED
void dynamic_menu_func PROTO((char *data, int md, int rv));
int ups_get_object_type PROTO((objid_t obj));
void add_to_new_selection PROTO((objid_t obj));
int ensure_visible PROTO((objid_t obj));
bool change_field PROTO((objid_t obj, int fnum, const char *new));
void n_draw PROTO((struct drawst *dets));
bool can_select PROTO((objid_t obj));
void gen_select PROTO((int wn, objid_t obj, int x, int y,
					int width, int height, int flags));
#endif /* OBJ_H_INCLUDED */

int prompt_for_choice PROTO((const char *what, const char *prompt,
			     const char *default_choice, const char *options,
			     int *p_choice));
int prompt_for_string PROTO((const char *what, const char *prompt,
			     const char *default_string, char **p_string));
int prompt_for_output_file PROTO((const char *prompt,
				  char **p_path, const char **p_mode));

#ifdef FED_H_INCLUDED
int suppress_ta_cursor_then_edit_field PROTO((edesc_t *edesc,
					      const char *what,
					      int expand_on_tab,
					      int preserve_ed_copy));
int message_on_null_name PROTO((edesc_t *edesc, const char *field_type));
int force_quit PROTO((int n_tries));
void num_filter PROTO((edesc_t *edesc));
#endif

#ifdef CI_H_INCLUDED
typedef struct {
	parse_id_t cr_parse_id;
	code_id_t cr_code_id;
	bool cr_code_has_func_calls;
} compile_res_t;

compile_res_t *compile_code PROTO((const char **lines, int nlines,
				   block_t *block, char *grp_arg,
				   lexinfo_t *error_lx,
				   const char *firstline, const char *lastline,
				   const char *repl_text, const char *repl_val));

code_id_t recompile_code PROTO((parse_id_t parse_id, code_id_t code_id,
								char *grp_arg));
void free_parse_and_code_ids PROTO((parse_id_t parse_id, code_id_t code_id));
#endif

/*  The following stuff would be in ui_priv.h were it not for the fact
 *  that tdr.c needs access to it.
 */
void target_menu_func PROTO((char *data, int md, int rv));
void permanent_menu_func PROTO((char *data, int md, int rv));
void do_debug_command PROTO((const char *s));

/*  State for set_dynamic_menu_updating_state().
 */
typedef enum { DMU_OFF, DMU_ON } dmu_state_t;

void set_dynamic_menu_updating_state PROTO((dmu_state_t new_state));
void set_bphead_menu_state PROTO((bool enable));
void set_dynamic_bphead_menu_state PROTO((int md, bool enable));
void set_dynamic_cmd_menu_state PROTO((int md, bool enable));
void set_dynamic_cmd_menu_demangling_state PROTO((int md, bool enable));
void print_string_to_output_win PROTO((const char *buf));
void open_output_win PROTO((void));
int change_to_hourglass_cursor PROTO((void));
