/* ui_layout.c - division of the window into regions, and window (re)drawing */

/*  Copyright 1991 Mark Russell, University of Kent at Canterbury.
 *
 *  You can do what you like with this source code as long as
 *  you don't try to make money out of it and you include an
 *  unaltered copy of this message (including the copyright).
 */

char ups_ui_layout_c_sccsid[] = "@(#)ui_layout.c	1.25 29/6/92 (UKC)";

#include <mtrprog/ifdefs.h>

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>

#include <local/wn.h>
#include <local/menu3.h>

#include <local/ukcprog.h>
#include <local/obj/obj.h>
#include <local/obj/o_priv.h>
#include <local/obj/mhole.h>
#include <local/obj/fed.h>
#include <local/obj/newtb.h>
#include <local/edit/edit.h>
#include <local/edit/win.h>
#include <local/edit/display.h>
#include <local/wn.h>

#include "cursors.h"
#include "reg.h"
#include "ui.h"
#include "tdr.h"
#include "textwin.h"
#include "ups.h"
#include "symtab.h"
#include "src.h"
#include "output.h"
#include "objtypes.h"
#include "ui_priv.h"
#include "va.h"
#include "state.h"
#include "menudata.h"
#include "debug.h"
#include "sccsdata.h"
#include "obj_env.h"
#include "obj_bpt.h"
#include "obj_misc.h"
#include "obj_signal.h"
#include "obj_bpt.h"
#include "obj_stack.h"
#include "sccsdata.h"

typedef struct {
	int ma_md;
	const char *ma_menu_name;
	bool ma_md_changed;
	void (*ma_menu_func)PROTO((char *data, int md, int rv));
	char *ma_data;
} menu_arg_t;

#define FIXED	0.0
#define REST	1.0

        sel_t *sel;
static void show_version PROTO((Edit_display *display));
static void dragbox_input PROTO((region_id_t region_id, event_t *ev));
static void dragbox_draw PROTO((region_id_t region_id, int wn,
				    int width, int height,
				    re_draw_reason_t draw_reason));

static void line_input PROTO((region_id_t region_id, event_t *ev));
static void line_draw PROTO((region_id_t region_id, int wn,
				    int width, int height,
				    re_draw_reason_t draw_reason));

static void tbar_input PROTO((region_id_t region_id, event_t *ev));
static void tbar_draw PROTO((region_id_t region_id, int wn,
			     int width, int height,
			     re_draw_reason_t draw_reason));

static void outwin_draw PROTO((region_id_t region_id, int wn,
			       int width, int height,
			       re_draw_reason_t draw_reason));

static void outwin_input PROTO((region_id_t region_id, event_t *ev));
static void namewin_input PROTO((region_id_t region_id, event_t *ev));
static void srcwin_input PROTO((region_id_t region_id, event_t *ev));
static void srcwin_draw PROTO((region_id_t region_id, int wn,
			       int width, int height,
			       re_draw_reason_t draw_reason));

static void typing_line_input PROTO((region_id_t region_id, event_t *ev));
static void typing_line_draw PROTO((region_id_t region_id, int wn,
				    int width, int height,
				    re_draw_reason_t draw_reason));

static void display_area_input PROTO((region_id_t region_id, event_t *ev));
static void display_area_draw PROTO((region_id_t region_id, int wn,
			       int width, int height,
			       re_draw_reason_t draw_reason));

static void menu_input PROTO((region_id_t region_id, event_t *ev));
static void menu_draw PROTO((region_id_t region_id, int wn,
			     int width, int height,
			     re_draw_reason_t draw_reason));


static void message_area_draw PROTO((region_id_t region_id, int wn,
				     int width, int height,
				     re_draw_reason_t draw_reason));

static void mousehole_draw PROTO((region_id_t region_id, int wn,
				  int width, int height,
				  re_draw_reason_t draw_reason));

void initialise_menu PROTO((region_id_t region_id, int md,
			    const char *menu_name,
			    void (*menu_func)(char *data, int mfmd, int rv),
			    char *menu_func_data));
static void add_srcwin PROTO((region_id_t region, double srcwin_fraction));
static tbar_id_t make_tbar PROTO((region_id_t region_id));

static int tbar_srcwin_callback PROTO((char *data, tb_action_t action, int val));
static int tbar_outwin_callback PROTO((char *data, tb_action_t action, int val));
static int tbar_obj_callback PROTO((char *data, tb_action_t action, int val));

static void srcwin_tbar_callback PROTO((char *data, tw_change_t change,
						int lnum_val, int pixel_val));
static void display_area_tbar_callback PROTO((char *data,
					      obj_change_t change, int val));
static void initialise_typing_line PROTO((region_id_t region_id));
static int get_default PROTO((const char *name, int default_value, int min_value));
static void output_window_menu_func PROTO((char *data, int md, int rv));
static void init_line_region PROTO((region_id_t line));
static bool typing_line_quit PROTO((region_id_t region, char *data, bool force));
static bool typing_line_accept_focus PROTO((region_id_t region, char *data));

static void
show_version(display)
Edit_display *display;
{
	errf("%s", _ups_sccsdata[0]);
}	

static int
get_default(name, default_value, min_value)
const char *name;
int default_value, min_value;
{
	const char *valstr;
	int value;

	/*  Some calls to us calculate default_value as 100 - percentage, so
	 *  set minumum constrain the default to ve above the minimum.
	 */
	if ((valstr = wn_get_default(name)) == NULL)
		return (default_value > min_value) ? default_value : min_value;

	if ((value = atoi(valstr)) >= min_value)
		return value;

	errf("Value %d for %s too small - using min value %d",
							value, name, min_value);
	return min_value;
}

/*  Do the initial division of the window into regions.
 */
region_id_t
divide_window_into_regions(wn)
int wn;
{
	enum {
		TOP, DISP_AREA_AND_TB, TOPLEFT,
		TYPING_LINE_AND_PMENU, TARGET_MENU_AND_DRAGBOX,
		TYPING_LINE, DMENU, MESSAGE_AREA, DRAGBOX,
		PMENU, MOUSEHOLE, DISPLAY_AREA_TB, DISPLAY_AREA,
		TARGET_MENU,
		START_TARGET_MENU, NEXT_TARGET_MENU, STEP_TARGET_MENU,
		CONT_TARGET_MENU, STOP_TARGET_MENU, UP_TARGET_MENU,
		DOWN_TARGET_MENU, EXIT_TARGET_MENU,
		OUTPUT, MAX_REGNUM
	};
	region_id_t root, line, dragbox_line, regs[(int)MAX_REGNUM];
	target_menu_info_t tmbuf;
	font_t *sysfont, *menufont;

	/*  Dimensions and percentages */
	int mousehole_width, menu_height;
	int pmenu_width, tbar_width, srcwin_name_width;
	int display_area_percent, srcwin_percent;
	float display_area_fraction, srcwin_fraction;

	int i, md;
	tbar_id_t obj_tbar_id;
	extern MENU cmd_men;
	extern MENU tgt_start_men, tgt_next_men, tgt_step_men;
	extern MENU tgt_cont_men, tgt_stop_men, tgt_up_men,
	tgt_down_men, tgt_exit_men;
	static MENU *tgtmenus[(int)TM_NTAGS] = {
		&tgt_start_men, &tgt_next_men, &tgt_step_men,
		&tgt_cont_men, &tgt_stop_men, &tgt_up_men,
		&tgt_down_men, &tgt_exit_men
	};

	/*  Set up fixed dimensions.  Some of these vary with font heights.
	 */
	sysfont = wn_get_sysfont();
	menufont = Mstdfont();
	menu_height = menufont->ft_height + 8;

	/*  BUG: get rid of these explicit pixel sizes.
	 */
	mousehole_width = get_default("MouseholeWidth", 160, 5);
	pmenu_width = get_default("PermanentMenuWidth", 160, 5);
	tbar_width = get_default("ScrollbarWidth", 20, 5);
	srcwin_name_width = get_default("SrcwinNameWidth", 150, 20);

	if (wn_get_default("SrcwinPercent") != NULL) {
		srcwin_percent = get_default("SrcwinPercent", 50, 10);
		display_area_percent = get_default("DisplayAreaPercent",
							  100 - srcwin_percent, 1);
	}
	else {
		display_area_percent = get_default("DisplayAreaPercent", 50, 1);
		srcwin_percent = get_default("SrcwinPercent", 
						100 - display_area_percent, 1);
	}

	display_area_fraction = (float)display_area_percent / 100;
	srcwin_fraction = (float)srcwin_percent / 100;

	for (i = 0; i < (int)MAX_REGNUM; ++i)
		regs[i] = NULL;
	line = NULL;

	root = re_make_region(wn);

	re_divide(root, RE_HORIZONTAL,
		  &regs[(int)TOP],	        0,		0.0,
		  &line,			1,		FIXED,
		  RE_END);
	init_line_region(line);
	re_divide(root, RE_HORIZONTAL,
		  &regs[(int)DISP_AREA_AND_TB],	0,	       display_area_fraction,
		  &line,			1,		FIXED,
		  &regs[(int)TARGET_MENU_AND_DRAGBOX],menu_height,FIXED,
		  RE_END);
	dragbox_line = line;
	add_srcwin(root, srcwin_fraction);

	re_divide(regs[(int)TOP], RE_VERTICAL,
		  &regs[(int)TOPLEFT],		0,			1.0,
		  &line,			1,			FIXED,
		  &regs[(int)MOUSEHOLE],	mousehole_width,	FIXED,
		  RE_END);
	
	re_divide(regs[(int)TOPLEFT], RE_HORIZONTAL,
		  &regs[(int)TYPING_LINE_AND_PMENU],	menu_height,	FIXED,
		  &line,				1,		FIXED,
		  &regs[(int)DMENU],			menu_height,	FIXED,
		  &line,				1,		FIXED,
		  &regs[(int)MESSAGE_AREA],		menu_height,	FIXED,
		  RE_END);
	
	re_divide(regs[(int)TYPING_LINE_AND_PMENU], RE_VERTICAL,
		  &regs[(int)TYPING_LINE],		0,		REST,
		  &line,				1,		FIXED,
		  &regs[(int)PMENU],			pmenu_width,	FIXED,
		  RE_END);
	
	re_divide(regs[(int)TARGET_MENU_AND_DRAGBOX], RE_VERTICAL,
		  &regs[(int)TARGET_MENU],		0,		REST,
		  &line,				1,		FIXED,
		  &regs[(int)DRAGBOX],			menu_height,	FIXED,
		  RE_END);
	
	re_divide(regs[(int)TARGET_MENU], RE_VERTICAL,
		  &regs[(int)START_TARGET_MENU],	0,		1.0,
		  &line,				1,		FIXED,
		  &regs[(int)NEXT_TARGET_MENU],		0,		1.0,
		  &line,				1,		FIXED,
		  &regs[(int)STEP_TARGET_MENU],		0,		1.0,
		  &line,				1,		FIXED,
		  &regs[(int)CONT_TARGET_MENU],		0,		1.0,
		  &line,				1,		FIXED,
		  &regs[(int)STOP_TARGET_MENU],		0,		1.0,
		  &line,				1,		FIXED,
		  &regs[(int)UP_TARGET_MENU],		0,		1.0,
		  &line,				1,		FIXED,
		  &regs[(int)DOWN_TARGET_MENU],		0,		1.0,
		  &line,				1,		FIXED,
		  &regs[(int)EXIT_TARGET_MENU],		0,		1.0,
		  RE_END);
	
	re_divide(regs[(int)DISP_AREA_AND_TB], RE_VERTICAL,
		  &regs[(int)DISPLAY_AREA_TB],		tbar_width,	FIXED,
		  &line,				1,		FIXED,
		  &regs[(int)DISPLAY_AREA],		0,		1.0,
		  RE_END);

	md = Minsert(&cmd_men);

	initialise_menu(regs[(int)PMENU], md, "pmenu",
					permanent_menu_func, (char *)NULL);
	set_dynamic_cmd_menu_state(md, 0);	/* RGA reset */

	initialise_menu(regs[(int)DMENU], -1, NULL,
					dynamic_menu_func, (char *)NULL);
	set_dynamic_menu_region(regs[(int)DMENU]);

	re_set_callbacks(regs[(int)MESSAGE_AREA], message_area_draw,
				(re_input_proc_t)NULL, (re_destroy_proc_t)NULL);
	set_message_wn(re_get_wn(regs[(int)MESSAGE_AREA]));

	re_set_callbacks(regs[(int)MOUSEHOLE], mousehole_draw,
				(re_input_proc_t)NULL, (re_destroy_proc_t)NULL);

	re_set_callbacks(regs[(int)DRAGBOX], dragbox_draw,
					dragbox_input, (re_destroy_proc_t)NULL);
	re_set_margins(regs[(int)DRAGBOX], 1, 1, 1, 1);
	re_set_data(regs[(int)DRAGBOX], (char *)dragbox_line);
	re_set_mhcaps(regs[(int)DRAGBOX], "drag", "up", "down");
	re_set_cursor(regs[(int)DRAGBOX], CU_DRAGBOX);

	obj_tbar_id = make_tbar(regs[(int)DISPLAY_AREA_TB]);
	tb_set_callback_and_data(obj_tbar_id, tbar_obj_callback, (char *)NULL);
	obj_set_callback_and_data(display_area_tbar_callback, (char *)obj_tbar_id);

	re_set_callbacks(regs[(int)DISPLAY_AREA], display_area_draw,
				display_area_input, (re_destroy_proc_t)NULL);
	set_display_area_wn(re_get_wn(regs[(int)DISPLAY_AREA]));
	re_set_cursor(regs[(int)DISPLAY_AREA], CU_DISPLAY);
	re_set_mhcaps(regs[(int)DISPLAY_AREA], "select", "edit", "toggle");
	
	for (i = 0; i < (int)TM_NTAGS; ++i) {
		tmbuf.tm_mdtab[i].md = Minsert(tgtmenus[i]);
		tmbuf.tm_mdtab[i].wn = re_get_wn(regs[(int)START_TARGET_MENU + i]);
		initialise_menu(regs[(int)START_TARGET_MENU + i],
				tmbuf.tm_mdtab[i].md, "tmenu", target_menu_func,
				(char *)NULL);
	}
	tmbuf.tm_current_md = -1;
	set_target_menu_info(&tmbuf);

	re_set_minsize(regs[(int)DISPLAY_AREA], 1, 2 * sysfont->ft_height + 3);

	initialise_typing_line(regs[(int)TYPING_LINE]);
	re_set_keyboard_focus(regs[TYPING_LINE], (char *)NULL);
	re_set_default_keyboard_focus(regs[TYPING_LINE], (char *)NULL);
	set_typing_line_region(regs[TYPING_LINE]);

	re_set_rootwin(root);
	re_redraw_root(EV_WINDOW_RESIZED, TRUE);

	return root;
}

static void
line_input(region_id, ev)
region_id_t region_id;
event_t *ev;
{
	int buttons, width, height;
	bool changed_position, is_vertical_shift;

	buttons = ev->ev_buttons;

	wn_get_window_size(re_get_wn(region_id), &width, &height);
	is_vertical_shift = width > height;
	changed_position = FALSE;

	if (buttons & B_LEFT) {
		int wn, x, y, last_coord;

		/*  As we are shifting windows around, get a stable window
		 *  to read events from, and translate the current event
		 *  coordinates to that window.
		 */
		wn = re_get_wn(re_get_parent(region_id));
		wn_trans_coords(ev->ev_wn, ev->ev_x, ev->ev_y, wn, &x, &y);

		last_coord = is_vertical_shift ? y : x;
		do {
			int coord;

			wn_next_event(wn, EV_MOUSE_MOVED | EV_BUTTON_UP, ev);
			coord = is_vertical_shift ? ev->ev_y : ev->ev_x;
			if (coord != last_coord) {
				re_change_position(region_id, coord - last_coord);
				last_coord = coord;
				changed_position = TRUE;
			}
		} while (ev->ev_type != EV_BUTTON_UP);
	}
	else if (buttons & (B_MIDDLE | B_RIGHT)) {
		int delta;

		delta = (buttons & B_MIDDLE) ? -1 : 1;
		re_change_position(region_id, delta);
		wn_wait_for_release_of(ev->ev_wn, B_MIDDLE | B_RIGHT);

		/*  The shift we've just done has moved the window, and we
		 *  want the cursor in the same window relative position so
		 *  that it remains over the line, so we warp it.
		 */
		wn_warp_mouse(ev->ev_wn, ev->ev_x, ev->ev_y);

		changed_position = TRUE;
	}

	if (changed_position)
		re_expose(re_get_parent(region_id), TRUE);
}

static void
dragbox_input(region_id, ev)
region_id_t region_id;
event_t *ev;
{
	line_input((region_id_t)re_get_data(region_id), ev);
}

static void
dragbox_draw(region_id, wn, win_width, win_height, draw_reason)
region_id_t region_id;
int wn;
int win_width, win_height;
re_draw_reason_t draw_reason;
{
	/*  The arrow_* #defines and data were generated by the X10 bitmap program.
	 *  They define the double headed arrow used in the source window drag box.
	 */
#define arrow_bitmap_width 23
#define arrow_bitmap_height 23
	static unsigned short arrow_bitmap_bits[] = {
	   0x0000, 0x0000, 0x0000, 0x0000,
	   0x0000, 0x0000, 0x0800, 0x0000,
	   0x1c00, 0x0000, 0x3e00, 0x0000,
	   0x7f00, 0x0000, 0xff80, 0x0000,
	   0xffc0, 0x0001, 0xffe0, 0x0003,
	   0x0000, 0x0000, 0x0000, 0x0000,
	   0x0000, 0x0000, 0xffe0, 0x0003,
	   0xffc0, 0x0001, 0xff80, 0x0000,
	   0x7f00, 0x0000, 0x3e00, 0x0000,
	   0x1c00, 0x0000, 0x0800, 0x0000,
	   0x0000, 0x0000, 0x0000, 0x0000,
	   0x0000, 0x0000
	};

	static bitmap_t arrow_bm = wn_static_bm(arrow_bitmap_width,
						arrow_bitmap_height, 1,
						0, 0, BM_BIT0_LEFT,
						(unsigned short *)arrow_bitmap_bits);
	int bm_x, bm_y, win_x, win_y;
	int width, height;

	if (draw_reason != RE_EXPOSED && draw_reason != RE_RESIZED)
		return;

	if (arrow_bm.bm_width < win_width) {
		bm_x = 0;
		width = arrow_bm.bm_width;
		win_x = (win_width - arrow_bm.bm_width) / 2;
	}
	else {
		bm_x = (arrow_bm.bm_width - win_width) / 2;
		width = win_width;
		win_x = 0;
	}

	if (arrow_bm.bm_height < win_height) {
		bm_y = 0;
		height = arrow_bm.bm_height;
		win_y = (win_height - arrow_bm.bm_height) / 2;
	}
	else {
		bm_y = (arrow_bm.bm_height - win_height) / 2;
		height = win_height;
		win_y = 0;
	}

	wn_rop_from_mem(&arrow_bm, bm_x, bm_y, width, height,
			wn, win_x, win_y, R_RPL);
}

static void
line_draw(region_id, wn, width, height, draw_reason)
region_id_t region_id;
int wn;
int width, height;
re_draw_reason_t draw_reason;
{
	switch (draw_reason) {
	case RE_EXPOSED:
	case RE_RESIZED:
		wn_set_area(wn, 0, 0, width, height, WN_FG);
		break;
	case RE_RESIZING:
	case RE_UNDRAW:
		wn_invert_area(wn, 0, 0, width, height);
		break;
	default:
		panic("bad dr in ld");
	}
}

static void
initialise_typing_line(region_id)
region_id_t region_id;
{
	edesc_t *edesc;

	re_set_callbacks(region_id, typing_line_draw,
			 typing_line_input, (re_destroy_proc_t)NULL);
	
	re_set_keyboard_focus_callbacks(region_id,
					typing_line_accept_focus,
					typing_line_quit);
	
	re_set_margins(region_id, 5, 5, 4, 4);
	re_set_cursor(region_id, CU_DISPLAY);
	re_set_mhcaps(region_id, "", "edit", "(menu)");
	re_set_newdata(region_id, (char *)NULL);

	edesc = (edesc_t *)e_malloc(sizeof(edesc_t));
	re_set_data(region_id, (char *)edesc);
	set_typing_line_edesc(edesc);
}

static void
message_area_draw(region_id, wn, width, height, draw_reason)
region_id_t region_id;
int wn;
int width, height;
re_draw_reason_t draw_reason;
{
	if (draw_reason == RE_EXPOSED || draw_reason == RE_RESIZED)
		wn_set_area(wn, 0, 0, width, height, WN_BG);
}

/*  Set up the typing area
 */
static void
typing_line_draw(region_id, wn, width, height, draw_reason)
region_id_t region_id;
int wn;
int width, height;
re_draw_reason_t draw_reason;
{
	Edit_display *display;

	if (draw_reason == RE_EXPOSED || draw_reason == RE_RESIZED)
	  ta_init((edesc_t *)re_get_data(region_id), wn, wn_get_sysfont());

	display = (Edit_display *)re_get_newdata(region_id);

	if (display == NULL) {
		Edit_keymap *keymap;
		
		display = make_text_display(wn, WN_FG, WN_BG, "");
		re_set_newdata(region_id, (char *)display);

		keymap = edit_create_keymap(edit_get_keymap(display));
		edit_add_keymap_control_entry(keymap, 'v', show_version);
		edit_set_keymap(display, keymap);
	}
	else {
		switch (draw_reason) {
		case RE_EXPOSED:
			edit_redraw_display(display);
			break;
		case RE_RESIZED:
			edit_update_wn_window_size(display, wn);
			edit_redraw_display(display);
			break;
		case RE_RESIZING:
		case RE_UNDRAW:
			break;
		}
	}
}

static void
typing_line_input(region, ev)
     region_id_t region;
     event_t *ev;
{
  Edit_display *display;
  int expand_on_tab = 0;
  char text_buff[200];

  display = (Edit_display *)re_get_newdata(region);

  if (ev->ev_type == EV_KEY) {
    edit_handle_wn_key_event(display, ev);

    expand_on_tab = ev->ev_char == '\t';
    if (edit_quit_requested(display) || expand_on_tab)
    {
      char *text, *c;

      if (expand_on_tab)
      {
	open_output_win();
	display_win_search(1, 0);	/* set */
      }
      text = get_typing_line_string();
      strcpy(text_buff, text);
      process_typing_line_text
	(text_buff, ev->ev_buttons & B_SHIFT_KEY);
      if (strlen(text_buff) > strlen(text))
	for(c = text_buff+strlen(text); c && *c; c++)
	{
	  ev->ev_char = *c;
	  edit_handle_wn_key_event(display, ev);
	}
      if (expand_on_tab)
	display_win_search(0, 1);	/* reset */
      free(text);
    }
  }
  else if (ev->ev_type == EV_BUTTON_DOWN) {
    if (ev->ev_buttons == B_LEFT) {
      do_selection(display, ev);
    }
    else if (ev->ev_buttons == B_RIGHT) {
      static const char *bptcaps_strs[] = {
	"UPS_F1_STR",
	"UPS_F2_STR",
	"UPS_F3_STR",
	"UPS_F4_STR",
	"UPS_F5_STR",
	"UPS_F6_STR",
	"UPS_F7_STR",
	"UPS_F8_STR",
	"UPS_F9_STR",
	"UPS_F10_STR",
	"UPS_F11_STR",
	"UPS_F12_STR",
	NULL
      };
      static const char *bptcaps[13];
      static i = -1;
      char *env, *c;
      static popup_t bptpop = { -1, FALSE, 0, bptcaps };
      int res, j;

      if (i == -1)
      {
	for (i = 0, j = 0;j < 12 ; j++)
	{
	  env = getenv(bptcaps_strs[j]);
	  if (env)
	  {
	    char *copy;
	    int controls = 0;

	    /*
	     **   Count the number of control characters in env
	     */
	    for ( copy = env; *copy; copy++ )
	      if ( *copy < 0x20) controls++;

	    /*
	     **   Now allocate space for the string
	     **   and copy, mapping Control-X to ^X.
	     */
	    bptcaps[i] = copy = malloc(strlen(env)+1 + controls);
	    for ( ; *env; env++ )
	    {
	      if ( *env < 0x20 )
	      {
		*copy++ = '^';
		*copy++ = 0x40 | *env;
	      }
	      else
		*copy++ = *env;
	    }
	    *copy='\0';
	    i++;
	  }
	}
	bptcaps[i] = NULL;
      }
      if (i > 0)
      {
	/*  
	**   Allow control chars in paste string.
	*/
	int escape = FALSE;
	int control = FALSE;
	res = select_from_popup(ev->ev_wn, B_RIGHT, &bptpop,
				ev->ev_x, ev->ev_y);
	switch(res) {
	case 0:
	case 1: 
	case 2: 
	case 3: 
	case 4: 
	case 5: 
	case 6: 
	case 7: 
	case 8: 
	case 9: 
	case 10:
	case 11:
	  for (c = (char *)*(bptcaps+res); c && *c; c++) {
	    ev->ev_char = *c;
	    if ( !escape )
	    {
		if ( control )
		{
		    ev->ev_char &= 0x1F;
		    control = FALSE;
		}
		else if (ev->ev_char == '\\' )
		{
		    escape = TRUE;
		    continue;
		}
		else if (ev->ev_char == '^' )
		{
		    control = TRUE;
		    continue;
		}
	    } else
	    {
		switch ( ev->ev_char )
		{
		case '^':
		case '\\':
		    /*
		    **   Pass \^ or \\ as ^ or \ respectively
		    */
		    break;

		case 'n':
		    /*
		    ** We paste in ESC rather than RETURN
		    ** here because of the check for RETURN
		    ** in the default case in edit_field
		    */
		case 'e':
		    /*
		    **   If at the end of the buffer, map to ESC to terminate the edit
		    */
		    if ( c[1] == '\0' )
		    {
			ev->ev_char = ESC;
			break;
		    }
		    /* else fall into default case */
		default:
		    /*
		    **    Back up and pass the '\' through
		    */
		    ev->ev_char = '\\';
		    c--;
		}
		escape = FALSE;	/* New */
	    }
	    edit_handle_wn_key_event(display, ev);
	  }
	  if (edit_quit_requested(display)) {
	    char *text, *c;
			
	    text = get_typing_line_string();
	    strcpy(text_buff, text);
	    process_typing_line_text
	       (text_buff, ev->ev_buttons & B_SHIFT_KEY);
	    if (strlen(text_buff) > strlen(text))
	      for(c = text_buff+strlen(text); c && *c; c++)
	      {
		ev->ev_char = *c;
		edit_handle_wn_key_event(display, ev);
	      }
	    free(text);
	  }
	  break;
	default:
	  break;
	}
      }
    }
    else if (ev->ev_buttons == B_MIDDLE) {
      size_t point;
			
      if (!re_set_keyboard_focus(region, (char *)NULL)) {
	wn_wait_for_release_of(ev->ev_wn, B_ANY);
	return;
      }
      if (edit_pixel_to_point(display, ev->ev_x, ev->ev_y,
			      &point))
	edit_move_point(display, point);

      edit_set_want_cursor(display, TRUE);
    }
    else if (ev->ev_buttons == B_MIDDLE | B_SHIFT_KEY)
    {
      /* make it into a ^y for a paste */
      ev->ev_char = 'y';
      ev->ev_buttons |= B_CONTROL_KEY;
      edit_handle_wn_key_event(display, ev);
    }
  }
}


static void
init_line_region(line)
region_id_t line;
{
	if ((Debug_flags & DBFLAG_MVLINES) == 0) {
		re_set_callbacks(line, line_draw, (re_input_proc_t)NULL,
							(re_destroy_proc_t)NULL);
	}
	else {
		re_set_callbacks(line, line_draw, line_input,
							(re_destroy_proc_t)NULL);
		re_set_mhcaps(line, "drag", "up", "down");
		re_set_cursor(line, CU_LINE);
	}
}

void
add_outwin()
{
	extern MENU out_men;
	static int md = -1;
	region_id_t srcwin_region, root, line, dragbox_line;
	region_id_t text_and_tb_region;
	region_id_t outwin_region, tbar_region;
	region_id_t menu_and_dragbox_region, dragbox_region, menu_region;
	double outwin_fraction;
	int header_height, outwin_percent, tbar_width;
	outwin_id_t outwin_id;
	tbar_id_t tbar_id;

	srcwin_region = get_current_srcwin_region();

	root = srcwin_region;
	for(;;) {
		region_id_t parent;

		parent = re_get_parent(root);
		if (parent == NULL)
			break;
		root = parent;
	}
				/* RGA make output window default smaller */
	outwin_percent = get_default("OutwinPercent", 10, 1);
	header_height = wn_get_sysfont()->ft_height + 6;
	tbar_width = get_default("ScrollbarWidth", 20, 5);

	outwin_fraction = (double)outwin_percent / 100;

	line = NULL;
	re_divide(root, RE_HORIZONTAL,
		  &line,			1,		FIXED,
		  RE_END);
	dragbox_line = line;
	init_line_region(line);

	menu_and_dragbox_region = text_and_tb_region = NULL;
	re_divide(root, RE_HORIZONTAL,
		  &menu_and_dragbox_region,	header_height,	FIXED,
		  &line,			1,		FIXED,
		  &text_and_tb_region,		0,		outwin_fraction,
		  RE_END);

	tbar_region = outwin_region = NULL;
	re_divide(text_and_tb_region, RE_VERTICAL,
		  &tbar_region,			tbar_width,	FIXED,
		  &line,			1,		FIXED,
		  &outwin_region,		0,		REST,
		  RE_END);
	
	menu_region = dragbox_region = NULL;
	re_divide(menu_and_dragbox_region, RE_VERTICAL,
		  &menu_region,			0,		REST,
		  &line,			1,		FIXED,
		  &dragbox_region,		header_height,	FIXED,
		  RE_END);

	outwin_id = ow_make_outwin(re_get_wn(outwin_region), WN_FG, WN_BG,
								wn_get_sysfont());
	re_set_data(outwin_region, (char *)outwin_id);
	re_set_callbacks(outwin_region, outwin_draw,
				outwin_input, (re_destroy_proc_t)NULL);
/*	re_set_callbacks(outwin_region, outwin_draw,
				(re_input_proc_t)NULL, (re_destroy_proc_t)NULL);*/

	re_set_callbacks(dragbox_region, dragbox_draw,
					dragbox_input, (re_destroy_proc_t)NULL);
	re_set_margins(dragbox_region, 1, 1, 1, 1);
	re_set_data(dragbox_region, (char *)dragbox_line);
	re_set_mhcaps(dragbox_region, "drag", "up", "down");
	re_set_cursor(dragbox_region, CU_DRAGBOX);

	tbar_id = make_tbar(tbar_region);
	tb_set_callback_and_data(tbar_id, tbar_outwin_callback, (char *)outwin_id);
	ow_set_textwin_callback_and_data(outwin_id, srcwin_tbar_callback,
								(char *)tbar_id);

	if (md == -1)
		md = Minsert(&out_men);
	else
		md = Mdup(md);

	initialise_menu(menu_region, md, "outwin",
			output_window_menu_func, (char *)outwin_id);

	re_redraw_root(EV_WINDOW_RESIZED, TRUE);

	set_current_outwin(outwin_id);
}
	
static void
add_srcwin(region, srcwin_fraction)
region_id_t region;
double srcwin_fraction;
{
	static int md = -1;
	extern MENU src_men;
	enum {
		LINE, HEADER, TEXT_AND_TB,
		TEXT, TB, NAMEWIN, SRC_MENU,
		NREGS
	};
	tbar_id_t tbar_id;
	srcwin_id_t srcwin_id;
	region_id_t regs[(int)NREGS];
	int i, tbar_width;

	tbar_width = get_default("ScrollbarWidth", 20, 5);

	for (i = 0; i < (int)NREGS; ++i)
		regs[i] = NULL;

	re_divide(region, RE_HORIZONTAL,
		  &regs[(int)LINE],		1,		FIXED,
		  RE_END);

	init_line_region(regs[(int)LINE]);

	re_divide(region, RE_HORIZONTAL,
		  &regs[(int)HEADER],		1,		0.0,
		  &regs[(int)LINE],		1,		FIXED,
		  &regs[(int)TEXT_AND_TB],	0,		srcwin_fraction,
		  RE_END);

	re_divide(regs[(int)HEADER], RE_VERTICAL,
		  &regs[(int)NAMEWIN],		1,		FIXED,
		  &regs[(int)LINE],		1,		FIXED,
		  &regs[(int)SRC_MENU],		0,		REST,
		  RE_END);

	re_divide(regs[(int)TEXT_AND_TB], RE_VERTICAL,
		  &regs[(int)TB],		tbar_width,	FIXED,
		  &regs[(int)LINE],		1,		FIXED,
		  &regs[(int)TEXT],		0,		REST,
		  RE_END);

	re_set_minsize(regs[(int)NAMEWIN], wn_get_sysfont()->ft_width * 30,
					   wn_get_sysfont()->ft_height + 6);
	re_set_minsize(regs[(int)SRC_MENU], 0, Mstdfont()->ft_height + 6);

	tbar_id = make_tbar(regs[(int)TB]);
	srcwin_id = src_make_srcwin(re_get_wn(regs[(int)TEXT]));

	/*  Tie all the bits together with callbacks.
	 */
	re_set_data(regs[(int)TEXT], (char *)srcwin_id);
	re_set_data(regs[(int)NAMEWIN], (char *)srcwin_id);

	tb_set_callback_and_data(tbar_id, tbar_srcwin_callback, (char *)srcwin_id);
	src_set_textwin_callback_and_data(srcwin_id, srcwin_tbar_callback,
								(char *)tbar_id);
	src_set_data(srcwin_id, (char *)re_get_wn(regs[(int)NAMEWIN]));

	re_set_callbacks(regs[(int)TEXT], srcwin_draw, srcwin_input,
							(re_destroy_proc_t)NULL);
	re_set_cursor(regs[(int)TEXT], CU_DISPLAY);
	re_set_mhcaps(regs[(int)TEXT], "select", "edit", "menu");
	
	re_set_callbacks(regs[(int)NAMEWIN], (re_draw_proc_t)NULL,
			 namewin_input, (re_destroy_proc_t)NULL);
	re_set_cursor(regs[(int)NAMEWIN], CU_DISPLAY);
	re_set_mhcaps(regs[(int)NAMEWIN], "", "", "menu");

	if (md == -1)
		md = Minsert(&src_men);
	else
		md = Mdup(md);

	initialise_menu(regs[(int)SRC_MENU], md, "srcwin",
			source_window_menu_func, (char *)srcwin_id);
	
	set_current_srcwin_region(regs[(int)TEXT]);
}

static int
tbar_obj_callback(unused_data, action, val)
char *unused_data;
tb_action_t action;
int val;
{
	int retval;

	switch (action) {
	case TB_GOTO:
		display_from(0, val);
		retval = val;
		break;
	case TB_SCROLL:
		retval = v_scroll(val);
		break;
	case TB_SHOW_POS:
		retval = 0;
		break;
	default:
		panic("bad action in tsc");
		retval = 0;	/* to satisfy gcc */
	}
	return retval;
}

static int
tbar_outwin_callback(data, action, val)
char *data;
tb_action_t action;
int val;
{
	outwin_id_t outwin_id;
	int retval;

	outwin_id = (outwin_id_t)data;

	switch (action) {
	case TB_GOTO:
		ow_goto_ypos(outwin_id, val);
		retval = val;
		break;
	case TB_SCROLL:
		retval = ow_scroll(outwin_id, val);
		break;
	case TB_SHOW_POS:
		retval = 0;
		break;
	default:
		panic("bad action in toc");
		retval = 0;	/* to satisfy gcc */
	}
	return retval;
}

static int
tbar_srcwin_callback(data, action, val)
char *data;
tb_action_t action;
int val;
{
	srcwin_id_t srcwin_id;
	int retval;
	const char *name;
	srcinfo_t sibuf;

	srcwin_id = (srcwin_id_t)data;

	switch (action) {
	case TB_GOTO:
		src_goto_ypos(srcwin_id, val);
		retval = val;
		break;
	case TB_SCROLL:
		retval = src_scroll(srcwin_id, val);
		break;
	case TB_SHOW_POS:
		src_get_info(srcwin_id, -1, val, &sibuf);
		name = (sibuf.si_fil != NULL) ? sibuf.si_fil->fi_name : NULL;
		echo_src_name_and_lnum(srcwin_id, name, sibuf.si_fil_lnum);
		retval = 0;
		break;
	default:
		panic("bad action in tsc");
		retval = 0;	/* to satisfy gcc */
	}
	return retval;
}

static void
display_area_tbar_callback(data, change, val)
char *data;
obj_change_t change;
int val;
{
	tbar_id_t tbar_id;

	tbar_id = (tbar_id_t)data;
	switch (change) {
	case OBJ_NEW_WINDOW_HEIGHT:
		tb_set_visible_units(tbar_id, val);
		break;
	case OBJ_NEW_HEIGHT:
		tb_set_total_units(tbar_id, val);
		break;
	case OBJ_NEW_YPOS:
		tb_goto(tbar_id, val, FALSE);
		break;
	default:
		panic("bad change in datc");
	}
}

static void
srcwin_tbar_callback(data, change, unused_lnum_val, pixel_val)
char *data;
tw_change_t change;
int unused_lnum_val, pixel_val;
{
	tbar_id_t tbar_id;

	tbar_id = (tbar_id_t)data;
	switch (change) {
	case TW_NEW_VISIBLE_UNITS:
		tb_set_visible_units(tbar_id, pixel_val);
		break;
	case TW_NEW_TOTAL_UNITS:
		tb_set_total_units(tbar_id, pixel_val);
		break;
	case TW_NEW_POS:
		tb_goto(tbar_id, pixel_val, FALSE);
		break;
	default:
		panic("bad change in stc");
	}
}

static tbar_id_t
make_tbar(region_id)
region_id_t region_id;
{
	int wn;
	tbar_id_t tbar_id;

	wn = re_get_wn(region_id);
	tbar_id = tb_create(wn);
	re_set_callbacks(region_id, tbar_draw, tbar_input, (re_destroy_proc_t)NULL);
	re_set_data(region_id, (char *)tbar_id);
	re_set_cursor(region_id, CU_SBAR);
	re_set_mhcaps(region_id, "down", "goto", "up");
	return tbar_id;
}

/*  Input function for the display area.  Just push back the event
 *  and hand control over to the object library.
 *
 *  BUG: for the moment, we ignore the region_id, and assume that there
 *       is only one display area.
 */
static void
display_area_input(region_id, ev)
region_id_t region_id;
event_t *ev;
{
#define min(x,y) ((x) > (y) ? (y) : (x))

	int buttons, ret = 0;

	buttons = ev->ev_buttons & B_ANY;

	if (buttons != 0) {
	  struct objst *obj;
	  sel_t *sel;
	  
	  if (ev->ev_buttons & B_LEFT && ev->ev_type == EV_BUTTON_DOWN &&
	      (ev->ev_buttons & B_CONTROL_KEY))
	  {
	    int last_x, last_y, x, y, buttons = B_LEFT;
	    event_t evbuf;
	    
	    last_x = ev->ev_x;
	    last_y = ev->ev_y;
	    
	    wn_get_release_event(ev->ev_wn, buttons, &evbuf);
	    set_event_win(&evbuf);
	    if (ev->ev_wn == evbuf.ev_wn)
	    {
	      x = evbuf.ev_x;
	      y = evbuf.ev_y;
	      wn_text_capture(1, 0, &last_x, &last_y, &x, &y, 0); /* set */
	      set_obj_wn(ev->ev_wn);
	      wn_text_capture(0, 1, &last_x, &last_y, &x, &y, 0); /* reset */
	      return;
	    }
	  }

	  if (buttons != B_MIDDLE)
			set_dynamic_menu_updating_state(DMU_OFF);
		wn_pushback_event(ev);
	 	obj = select_list();
		if (buttons != B_MIDDLE)
			set_dynamic_menu_updating_state(DMU_ON);
	  /* RGA: accelerators for (shift) left click on some objects... */ 
	  if (buttons == B_LEFT && obj)
	  {
	    td_set_obj_updating(OBJ_UPDATING_OFF);
	    for (sel = get_selection(); sel != NULL;)
	    {
	      switch (obj->ob_type)
	      {
	      case OT_VAR:
		if (ev->ev_buttons & B_SHIFT_KEY)
		{
		  struct objst *ob;
		  for (ob = obj->ob_child; ob; ob = ob->ob_next)
		    if (!ob->ob_child)
		      break;
		  do_var_or_expr((objid_t)sel->se_code,
				 (!obj->ob_child || !ob) ?
				 MR_VAR_EXPAND : MR_VAR_COLLAPSE,
				 (char *)NULL);
		}
		break;
	      case OT_EXPR:
		if (ev->ev_buttons & B_SHIFT_KEY)
		{
		  struct objst *ob;
		  for (ob = obj->ob_child; ob; ob = ob->ob_next)
		    if (!ob->ob_child)
		      break;
		  do_var_or_expr((objid_t)sel->se_code,
				 (!obj->ob_child || !ob) ?
				 MR_VAR_EXPAND : MR_VAR_COLLAPSE,
				 (char *)NULL);
		}
		break;
	      case OT_BPHEAD:
		set_bphead_menu_state(global_bp_enabled(0, 0));
		if (ev->ev_buttons & B_SHIFT_KEY)
		  do_bps((objid_t)sel->se_code, MR_ADD_BREAKPOINT, (char *)NULL);
		break;
	      case OT_SRCHEAD:
		if (ev->ev_buttons & B_SHIFT_KEY)
		  do_srchead((objid_t)sel->se_code, !obj->ob_child ?
			     MR_SHOW_SOURCE_FILES : MR_HIDE_SOURCE_FILES,
			     (char *)NULL);
		break;
	      case OT_ENVHEAD:
		if (ev->ev_buttons & B_SHIFT_KEY)
		  do_envhead
		    ((objid_t)sel->se_code, !obj->ob_child ?
		     MR_SHOW_ALL_ENV_ENTRIES : MR_HIDE_ALL_ENV_ENTRIES,
		     (char *)NULL);
		break;
	      case OT_SGHEAD:
		if (ev->ev_buttons & B_SHIFT_KEY)
		  do_sgh((objid_t)sel->se_code, !obj->ob_child ?
			 MR_SHOW_ALL_SIGNALS : MR_HIDE_ALL_SIGNALS,
			 (char *)NULL);
		break;
	      case OT_BPT:
		if (ev->ev_buttons & B_SHIFT_KEY)
		{
		  ret = do_bpt2((objid_t)sel->se_code, MR_BPTMEN_REMOVE,
			       (char *)NULL);
		  if (ret)
		  {
		    sel = get_selection();
		    if (sel == NULL)
		      break;
		  }
		}
		else
		  do_bpt((objid_t)sel->se_code, MR_BPTMEN_SOURCE, (char *)NULL);
		break;
	      case OT_SFILE:
	      case OT_SFILE_EX:
		if (ev->ev_buttons & B_SHIFT_KEY)
		  do_file((objid_t)sel->se_code, !obj->ob_child ?
			  MR_EXPAND_LIKE_BEFORE : MR_COLLAPSE,
			  (char *)NULL);
		else
		  do_file((objid_t)sel->se_code, MR_DISPLAY_SOURCE,
			  (char *)NULL);
		break;
	      case OT_FUNC:
		if (ev->ev_buttons & B_SHIFT_KEY)
		  do_func((objid_t)sel->se_code, !obj->ob_child ?  
			  MR_EXPAND_LIKE_BEFORE : MR_COLLAPSE,
			  (char *)NULL);
		else
		  do_func((objid_t)sel->se_code, MR_DISPLAY_SOURCE,
			  (char *)NULL);
		break;
	      case OT_BLOCK:
		if (ev->ev_buttons & B_SHIFT_KEY)
		  do_block((objid_t)sel->se_code, !obj->ob_child ?  
			  MR_EXPAND_LIKE_BEFORE : MR_COLLAPSE,
			   (char *)NULL);
		else
		  do_block((objid_t)sel->se_code, MR_DISPLAY_SOURCE,
			   (char *)NULL);
		break;
	      }
	      if (!ret)
		sel = sel->se_next;
	    }
	    td_set_obj_updating(OBJ_UPDATING_ON);
	  }
	}
}

static void
display_area_draw(region_id, wn, width, height, draw_reason)
region_id_t region_id;
int wn;
int width, height;
re_draw_reason_t draw_reason;
{
	if (draw_reason == RE_EXPOSED || draw_reason == RE_RESIZED)
		set_obj_wn(wn);
}

static void
output_window_menu_func(arg, md, rv)
char *arg;
int md, rv;
{
	outwin_id_t outwin_id;

	outwin_id = (outwin_id_t)arg;

	switch (rv) {
	case MR_OUTWIN_PAGE_UP:
		ow_page(outwin_id, OW_PAGE_UP);
		break;
	case MR_OUTWIN_PAGE_DOWN:
		ow_page(outwin_id, OW_PAGE_DOWN);
		break;
	case MR_OUTWIN_CLEAR:
		ow_clear(outwin_id);
		break;
	case MR_OUTWIN_SEARCH_FORWARDS:
	case MR_OUTWIN_SEARCH_BACKWARDS:
		ow_search(outwin_id, get_typing_line_string(),
						rv == MR_OUTWIN_SEARCH_FORWARDS);
		break;
	default:
		panic("bad rv in owmf");
	}
}

char *
get_typing_line_string()
{
	Edit_display *display;
	
	display = (Edit_display *)re_get_newdata(get_typing_line_region());
	return get_buffer_contents(edit_get_buffer(display));
}
		
static bool
typing_line_accept_focus(region, data)
region_id_t region;
char *data;
{
	Edit_display *display;

	display = (Edit_display *)re_get_newdata(region);

	if (display != NULL)
		edit_set_want_cursor(display, TRUE);
	
	return TRUE;
}

static bool
typing_line_quit(region, data, force)
region_id_t region;
char *data;
bool force;
{
	edit_set_want_cursor((Edit_display *)re_get_newdata(region), FALSE);
	return TRUE;
}

static void
outwin_draw(region_id, wn, width, height, draw_reason)
region_id_t region_id;
int wn;
int width, height;
re_draw_reason_t draw_reason;
{
	if (draw_reason == RE_EXPOSED || draw_reason == RE_RESIZED)
		ow_redraw((outwin_id_t)re_get_data(region_id));
}

static void
namewin_input(region_id, ev)
region_id_t region_id;
event_t *ev;
{
	name_window_event_handler((srcwin_id_t)re_get_data(region_id), ev);
}

static void
srcwin_draw(region_id, wn, width, height, draw_reason)
region_id_t region_id;
int wn;
int width, height;
re_draw_reason_t draw_reason;
{
	if (draw_reason == RE_EXPOSED || draw_reason == RE_RESIZED)
		src_redraw((srcwin_id_t)re_get_data(region_id));
}

static void
srcwin_input(region_id, ev)
region_id_t region_id;
event_t *ev;
{
	source_window_event_handler((srcwin_id_t)re_get_data(region_id), ev);
}

static void
tbar_draw(region_id, wn, width, height, draw_reason)
region_id_t region_id;
int wn;
int width, height;
re_draw_reason_t draw_reason;
{
	tbar_id_t tbar_id;

	if (draw_reason != RE_EXPOSED && draw_reason != RE_RESIZED)
		return;

	tbar_id = (tbar_id_t)re_get_data(region_id);
	if (draw_reason == RE_RESIZED)
		tb_set_box_pos_and_size(tbar_id, TRUE, 0, width / 3);
	tb_expose(tbar_id);
}

static void
tbar_input(region_id, ev)
region_id_t region_id;
event_t *ev;
{
	tbar_event_handler((tbar_id_t)re_get_data(region_id), ev);
}

void
initialise_menu(region_id, md, menu_name, menu_func, menu_func_data)
region_id_t region_id;
int md;
const char *menu_name;
void (*menu_func)PROTO((char *mf_data, int mf_md, int mf_rv));
char *menu_func_data;
{
	menu_arg_t *ma;

	ma = (menu_arg_t *)e_malloc(sizeof(menu_arg_t));

	if (md != -1) {
		ma->ma_md = md;
		Mnobox(ma->ma_md);
	}
	else
		ma->ma_md = -1;

	ma->ma_menu_name = menu_name;
	ma->ma_menu_func = menu_func;
	ma->ma_data = menu_func_data;
	ma->ma_md_changed = FALSE;

	re_set_data(region_id, (char *)ma);
	re_set_callbacks(region_id, menu_draw, menu_input, (re_destroy_proc_t)NULL);
	re_set_cursor(region_id, CU_MENU);
	re_set_mhcaps(region_id, "select", "", "");
}

static void
menu_input(region_id, ev)
region_id_t region_id;
event_t *ev;
{
	extern MENU tgt_step_men;
	menu_arg_t *ma;
	int wn, mask, rv, step_menu = 0;

	ma = (menu_arg_t *)re_get_data(region_id);

	if (ma->ma_md == -1 || (ev->ev_buttons & B_LEFT) == 0)
		return;

	if (!strcmp(ma->ma_menu_name, "tmenu"))
	  if (allow_step_submenu())
	  {
	    step_menu = 1;
	    Mactivate_submenu(&tgt_step_men);
	  }
	  else
	    Mdeactivate_submenu(&tgt_step_men);
	wn = ev->ev_wn;
	mask = ~(1 << ma->ma_md);
	if (!Mselect(ev->ev_x, ev->ev_y, wn, MS_PRESSED, mask))
		return;

	for (;;) {
		wn_next_event(wn, EVENT_MASK, ev);
		if ((ev->ev_buttons & B_LEFT) == 0)
			break;
		Mselect(ev->ev_x, ev->ev_y, wn, MS_CONTINUE, mask);
	}
	rv = MS_rv(Mselect(ev->ev_x, ev->ev_y, wn, MS_RELEASED, mask));

	if (rv != 0) {

	  /* Refresh before moving stepping into possible reverse video */
	  /* file. Also refresh outwin when listing help verbiage  RGA */

	  if (step_menu || !strcmp(ma->ma_menu_name, "pmenu"))	
	    re_expose(region_id, FALSE);
	  td_record_menu_command(ma->ma_menu_name, rv);
	  (*ma->ma_menu_func)(ma->ma_data, ma->ma_md, rv);
	}

	Mclear(ma->ma_md);
}

/*  Show the target menu.
 */
static void
menu_draw(region_id, wn, width, height, draw_reason)
region_id_t region_id;
int wn;
int width, height;
re_draw_reason_t draw_reason;
{
	menu_arg_t *ma;

	if (draw_reason != RE_EXPOSED && draw_reason != RE_RESIZED)
		return;

	ma = (menu_arg_t *)re_get_data(region_id);

	if (ma->ma_md != -1) {
		if (draw_reason == RE_RESIZED || ma->ma_md_changed) {
			Mplace(ma->ma_md, 0, 0);
			Msize(ma->ma_md, width, height);
			ma->ma_md_changed = FALSE;
		}
		Mdisplay(ma->ma_md, wn, FALSE);
	}
	else
		wn_shade_area(wn, 0, 0, width, height, WN_GREY4, R_RPL);
}

/*  Set the dynamic menu to md.
 */
void
set_dynamic_bphead_menu_state(md, enable)
int md;
bool enable;
{
	menu_arg_t *ma;
	region_id_t dmenu_region;
	char rvs[2];

	dmenu_region = get_dynamic_menu_region();

	ma = (menu_arg_t *)re_get_data(dmenu_region);
	rvs[0] = (enable) ?
	  (char)MR_ENABLE_ALL_BREAKPOINTS : (char)MR_DISABLE_ALL_BREAKPOINTS;
	rvs[1] = 0;
	Mnonsel(md, rvs);
	re_expose(dmenu_region, FALSE);
}


/*  Set the dynamic menu to md.
 */
void
set_dynamic_cmd_menu_state(md, enable)
int md;
bool enable;
{
	char rvs[2];

	rvs[0] = (enable) ?
	  (char)MR_LOGGING_ON : (char)MR_LOGGING_OFF;
	rvs[1] = 0;
	Mnonsel(md, rvs);
}

/*  Set the dynamic menu to md.
 */
void
set_dynamic_menu(md, menu_name)
int md;
const char *menu_name;
{
	menu_arg_t *ma;
	region_id_t dmenu_region;

	dmenu_region = get_dynamic_menu_region();

	ma = (menu_arg_t *)re_get_data(dmenu_region);

	if (md != ma->ma_md && ma->ma_md != -1)
		Mremove(ma->ma_md);

	ma->ma_menu_name = menu_name;
	ma->ma_md = md;
	ma->ma_md_changed = TRUE;

	re_expose(dmenu_region, FALSE);
}

static void
mousehole_draw(region_id, wn, width, height, draw_reason)
region_id_t region_id;
int wn;
int width, height;
re_draw_reason_t draw_reason;
{
	if (draw_reason == RE_EXPOSED || draw_reason == RE_RESIZED) {
		wn_set_area(wn, 0, 0, width, height, WN_BG);
		draw_mousehole(wn, 0, 0, width, height);
	}
}
static void
outwin_input(region_id, ev)
region_id_t region_id;
event_t *ev;
{
 	outwin_window_event_handler((outwin_id_t)re_get_data(region_id), ev);
}

