/* ui_src.c - input handling in the source window */

/*  Copyright 1991 Mark Russell, University of Kent at Canterbury.
 *
 *  You can do what you like with this source code as long as
 *  you don't try to make money out of it and you include an
 *  unaltered copy of this message (including the copyright).
 */

char ups_ui_src_c_sccsid[] = "@(#)ui_src.c	1.30 13/9/92 (UKC)";

#include <mtrprog/ifdefs.h>

#include <sys/types.h>
#include <sys/param.h>		/* for MAXPATHLEN */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <local/wn.h>
#include <local/ukcprog.h>
#include <local/obj/obj.h>
#include <local/menu3.h>

#include <local/obj/fed.h>
#include "ups.h"
#include "symtab.h"
#include "src.h"
#include "tdr.h"
#include "ui.h"
#include "ui_priv.h"
#include "cursors.h"
#include "menudata.h"

#include "expr.h"
#include "obj_bpt.h"
#include "exec.h"
#include "state.h"
#include "objtypes.h"

#define MAXVARLEN	128	/* RGA increase for C++ */

typedef enum { QR_CONFIRM, QR_NEWLINE, QR_DEL, QR_OTHER } quit_reason_t;

static void confirm_var PROTO((event_t *ev, srcwin_id_t srcwin_id,
			       fil_t *fil, int lnum, int fil_lnum,
			       const char *text,
			       const char *start, const char *end));
static void edit_editblock PROTO((event_t *ev, srcwin_id_t srcwin_id, srcinfo_t *si));
static int editblock_quitfunc PROTO((edesc_t *edesc, int n_tries));
static void editblock_keyfunc PROTO((edesc_t *edesc));
static int get_edit_tabspaces PROTO((void));
static int update_lines PROTO((srcinfo_t *si, editblockinfo_t *ei, int lnum,
			       const char *line, int nlines, lexinfo_t *lx));
static void null_ofunc PROTO((const char *s));
static language_t get_user_lang PROTO(());

static void
null_ofunc(s)
const char *s;
{
}

static language_t 
get_user_lang()
{
  static language_t lang = LANG_UNKNOWN;
  static int checked = 0;
  char *c, *c_str = NULL;
  
  if (!checked)
  {
    c_str = (char *)getenv("UPS_LANGUAGE");
    if (c_str)
      for (c = c_str; *c; c++)
	*c = tolower(*c);
    if (c_str)
      if (!strcmp(c_str, "c++"))
	lang = LANG_CC;
      else
	if (!strcmp(c_str, "c"))
	  lang = LANG_C;
	else
	  if (!strcmp(c_str, "fortran"))
	    lang = LANG_FORTRAN;
    checked = 1;
  }
  return lang;
}

/*  Give feedback on the selection of a variable in the source
 *  window.  Text points to the string containing the line being
 *  selected; start and end point to the start and end of the
 *  variable name.
 *
 *  We draw a box round the variable name which is lit only when
 *  when the mouse is over the name.  If the button is released
 *  with the box displayed, we call show_var() or show_func() to
 *  try to display the variable.
 */
static void
confirm_var(ev, srcwin_id, fil, lnum, fil_lnum, text, start, end)
event_t *ev;
srcwin_id_t srcwin_id;
fil_t *fil;
int lnum;
int fil_lnum;
const char *text, *start, *end;
{
	char *cptr;
	int have_box, in_word, found = 0, *this_obj_var,
	this_obj_exists = 0, mangled = 0;
	srcinfo_t sibuf;
	char buf[MAXVARLEN + 1];
	errf_ofunc_t oldf = NULL;

	have_box = FALSE;
	in_word = TRUE;
	this_obj_var = 0;
	for(;;) {
		if (in_word != have_box) {
			src_draw_box(srcwin_id, lnum, start - text, end - start);
			have_box = !have_box;
		}

		wn_next_event(ev->ev_wn, EVENT_MASK, ev);
		if (!(ev->ev_buttons & B_LEFT))
			break;

		src_get_info(srcwin_id, ev->ev_x, ev->ev_y, &sibuf);
		in_word = sibuf.si_cnum != -1 && sibuf.si_src_lnum == lnum &&
			  sibuf.si_text + sibuf.si_cnum >= start &&
			  sibuf.si_text + sibuf.si_cnum < end;
	}

	if (have_box) {
		cursor_t old_cursor;

		old_cursor = wn_get_window_cursor(WN_STDWIN);
		set_bm_cursor(WN_STDWIN, CU_WAIT);

		src_draw_box(srcwin_id, lnum, start - text, end - start);
		(void) strncpy(buf, start, end - start);

		buf[end - start] = '\0';

		/*  f77 maps all upper case variable names to
		 *  lower case, so follow suit.
		 */
		if (fil->fi_language == LANG_FORTRAN) {
			for (cptr = buf; *cptr != '\0'; ++cptr)	
				if (isupper(*cptr))
					*cptr = tolower(*cptr);
		}

		td_record_show_var(fil, fil_lnum, buf);
		/* RGA for mangled C++ names */
		if (!(ev->ev_buttons & B_SHIFT_KEY) &&
		    fil->fi_language == LANG_CC || get_user_lang() == LANG_CC)
		{
		  /* suppress error messages for all except last try */
		  oldf = errf_set_ofunc(null_ofunc);
		  mangled = 1;
		}
		if (mangled)
		  found = display_var_in_mangled_form
		    (buf, &this_obj_exists, &this_obj_var);
		if (!found)
		  found = show_var(srcwin_id, fil, fil_lnum, buf,
				   ev->ev_buttons & B_SHIFT_KEY, oldf);

		if (mangled)
		  errf_set_ofunc(oldf);
		if (found != 1)
		{
		  if (mangled)
		    if (!this_obj_exists)
		    {
		      if (this_obj_var)
			delete_variable(this_obj_var); /* get rid of "this"  */
		    }
		    else
		      if (this_obj_var)
			clear_selection();
		  if (found != 2) /* 2 = may be message about virtual or */
				  /* inherited match */
		    errf("No function or variable `%s'", buf);
		}
	
	wn_define_cursor(WN_STDWIN, old_cursor);
	      }
      }

/*  Display the name and lnum as the source file name and line
 *  number.
 *
 *  BUG: this function is a quick hack, full of magic numbers.
 */
void
echo_src_name_and_lnum(srcwin_id, name, lnum)
srcwin_id_t srcwin_id;
const char *name;
int lnum;
{
	static char old[60];
	char nbuf[60];
	int wn, nchars, name_nchars, namelen, width, height;
	font_t *font;

	wn = (int)src_get_data(srcwin_id);
	if (name == NULL || lnum == -1)
		return;

	wn_get_window_size(wn, &width, &height);
	font = wn_get_sysfont();
	nchars = (width - 8) / font->ft_width;
	if (nchars > sizeof(nbuf) - 1)
		nchars = sizeof(nbuf) - 1;

	namelen = strlen(name);
	name_nchars = nchars - 6;
	if (strlen(name) < name_nchars)
		sprintf(nbuf, "%*s:%-5d", name_nchars, name, lnum % 100000);
	else
		sprintf(nbuf, "%s:%-5d", name + namelen - name_nchars, lnum % 100000);

	if (strcmp(nbuf, old) != 0) {
		wn_tputs(wn, nbuf, 4, (height - font->ft_height) / 2);
		(void) strcpy(old, nbuf);
	}
}

/*  Mouse hole caption function for the source window.  No captions
 *  displayed if there is no source in the source window.
 */
int
mfn_source(caps, arg)
int caps;
char *arg;
{
	srcinfo_t sibuf;

	src_get_info((srcwin_id_t )arg, 0, 0, &sibuf);
	return (sibuf.si_fil != NULL) ? caps : 0;
}

static int
get_edit_tabspaces()
{
	font_t *srcfont, *editfont;
	int tabspaces;

	src_get_fonts(get_current_srcwin(), &srcfont, &editfont);

	tabspaces = get_tabwidth();
	if (editfont->ft_width != 0)
		tabspaces = (tabspaces * srcfont->ft_width) / editfont->ft_width;
	
	return tabspaces;
}

/*  Keyboard input function source edit blocks.
 */
static void
editblock_keyfunc(edesc)
edesc_t *edesc;
{
	int i, tabspaces, nspaces;

	if (edesc->ed_act != EDA_KEY)
		return;

	switch (edesc->ed_char) {
	case RETURN:
		edesc->ed_user = (int)QR_NEWLINE;
		break;
	case OOPS:
	case DEL:
		if (edesc->ed_curpos == 0) {
			edesc->ed_user = (int)QR_DEL;
			edesc->ed_meaning = EDM_CONFIRM;
		}
		break;
	case CONTROL('D'):
		/*  Sort of do what ^D does in input mode in vi.
		 *
		 *  Sorry if you use EMACS.  I wrote it, I get to
		 *  to choose what editor it is almost, but not quite
		 *  entirely unlike.
		 */

		if (edesc->ed_curpos == 0)
			break;
		
		for (i = 0; i < edesc->ed_curpos; ++i) {
			if (edesc->ed_copy[i] != ' ')
				break;
		}
		if (i < edesc->ed_curpos)
			break;

		tabspaces = get_edit_tabspaces();
		wn_updating_off(edesc->ed_wn);
		do {
			do_edit(edesc, EDM_DELETE_CHAR);
		} while (edesc->ed_curpos % tabspaces != 0);
		wn_updating_on(edesc->ed_wn);

		edesc->ed_meaning = EDM_CONT;
		break;
	case '\t':
		tabspaces = get_edit_tabspaces();
		nspaces = tabspaces - edesc->ed_curpos % tabspaces;

		edesc->ed_char = ' ';
		wn_updating_off(edesc->ed_wn);
		while (--nspaces >= 0)
			do_edit(edesc, EDM_INSERT_CHAR);
		wn_updating_on(edesc->ed_wn);

		/*  Make the meaning a NOP.
		 */
		edesc->ed_meaning = EDM_CONT;
		break;
	
	default:
		break;
	}
}

static int
editblock_quitfunc(edesc, n_tries)
edesc_t *edesc;
int n_tries;
{
	if (strcmp(edesc->ed_copy, edesc->ed_orig) == 0)
		return EDR_CONFIRM_NO_CHANGE;
	return EDR_CONFIRM_CHANGE;
}

/*  We call this routine to tell the editblock stuff about the line
 *  that has changed as a result of the user's edit to line lnum
 *  of the block.
 *
 *  We call this on confirmation of the edit and also (with lx NULL)
 *  just before doing a line split or join.  The reason for the latter
 *  cases is to refresh the textwin's knowledge of what is displayed
 *  on each line.  If we don't do this it will do it's optimised minimal
 *  redraw based on stale data and get things wrong.
 */
static int
update_lines(si, ei, lnum, line, nlines, lx)
srcinfo_t *si;
editblockinfo_t *ei;
int lnum;
const char *line;
int nlines;
lexinfo_t *lx;
{
	const char **lines;
	int i, res;
	call_callback_t call_callback;

	lines = (const char **)e_malloc(nlines * sizeof(char *));

	for (i = 0; i < nlines; ++i)
		lines[i] = ei->ei_lines[i];
	lines[lnum] = line;

	call_callback = (lx != NULL) ? DO_CALL_CALLBACK : DONT_CALL_CALLBACK;

	res = change_editblock(si->si_editblock_id, lines, nlines,
					call_callback, (callback_arg_t)lx);

	free((char *)lines);

	/*  Refresh caller's passed in editblock info, as the
	 *  change_editblock call above will have made ei_lines stale.
	 */
	get_editblock_info(si->si_editblock_id, ei);

	return res;
}

/*  Handle user edits of an editblock block.  The bulk of the work
 *  is supporting multi-line editing on top of edit_field() single
 *  line edits.
 */
static void
edit_editblock(ev, srcwin_id, si)
event_t *ev;
srcwin_id_t srcwin_id;
srcinfo_t *si;
{
	int line_height, event_nchars, width, lnum;
	lexinfo_t lxbuf;
	bool finished;
	quit_reason_t quitres;

	if (si->si_wn != ev->ev_wn)
		panic("wn botch in ee");

	lnum = si->si_editblock_offset;
	width = wn_get_width(si->si_wn);
	line_height = si->si_font->ft_height;
	event_nchars = -1;

	clear_selection();

	wn_pushback_event(ev);

	finished = FALSE;
	do {
		edesc_t edescbuf;
		editblockinfo_t eibuf;
		int res, edit_res, i, edit_wn, y, nlines;
		const char **lines;

		get_editblock_info(si->si_editblock_id, &eibuf);
		nlines = eibuf.ei_nlines;

		src_show_lnum(srcwin_id, eibuf.ei_src_lnum + lnum, TRUE);
		y = src_lnum_to_ypos(srcwin_id, eibuf.ei_fil,
							eibuf.ei_src_lnum + lnum);

		/*  Make a window that fits the edit block.
		 */
		edit_wn = wn_create_subwin(si->si_wn, 0, y, width, line_height,
								   WN_INPUT_OUTPUT);

		make_edesc(&edescbuf, edit_wn, eibuf.ei_lines[lnum],
						width / si->si_font->ft_width,
						eibuf.ei_fg, eibuf.ei_bg);
		edescbuf.ed_user = (int)QR_CONFIRM;
		edescbuf.ed_quitfunc = (ed_quitfunc_t)editblock_quitfunc;
		edescbuf.ed_keyfunc = (ed_keyfunc_t)editblock_keyfunc;
		edescbuf.ed_font = si->si_font;
		edescbuf.ed_char = ' ';

		if (event_nchars != -1) {
			edescbuf.ed_newpos = event_nchars;
			do_edit(&edescbuf, EDM_SETCURSOR);
		}

		edit_res = suppress_ta_cursor_then_edit_field
		  (&edescbuf, (const char *)NULL, 0, 0);
		if (edit_res == EDR_CONFIRM_CHANGE ||
						edit_res == EDR_CONFIRM_NO_CHANGE)
			quitres = (quit_reason_t)edescbuf.ed_user;
		else
			quitres = QR_OTHER;

		switch (quitres) {
		case QR_OTHER:
			finished = TRUE;
			break;
		case QR_CONFIRM:
			{
				bool in_lines;

				if (edescbuf.ed_act == EDA_PUCK) {
					int ty;

					ty = edescbuf.ed_puck_y;
					in_lines = ty >= -lnum * line_height &&
						   ty < (nlines - lnum) * line_height;
				}
				else
					in_lines = FALSE;
						
				res = update_lines(si, &eibuf,
						   lnum, edescbuf.ed_copy,
						   nlines,
						   in_lines ? NULL : &lxbuf);

				if (in_lines)
					finished = TRUE;
				else if (lxbuf.lx_lnum == -1)
					finished = TRUE;
				else {
					if (edescbuf.ed_flags & EDF_UNGETPUCK)
						wn_next_event(edit_wn,EVENT_MASK,ev);
					if (lxbuf.lx_lnum >= nlines) {
						lnum = nlines - 1;
						event_nchars =
						   strlen(eibuf.ei_lines[lnum]);
					}
					else {
						lnum = lxbuf.lx_lnum;
						event_nchars = lxbuf.lx_cnum;
					}
				}
				if (finished)
				  if (si->si_fil->fi_flags & FI_DUPLICATE)
				  {
				    fil_t *fil1;
				    
				    if (find_duplicate_file(si->si_fil, &fil1))
				    {
				      editblock_id_t ed_t;
				      
				      ed_t = update_tracking_editblock
					(si->si_editblock_id, fil1, 
					 (const char **)eibuf.ei_lines, nlines);
				      if (ed_t)
				      {
					edesc_t edescbuf1;
					editblockinfo_t eibuf1;
					srcinfo_t *si1, si0;

					si1 = &si0;
					memcpy((char *)&si1, (char *)&si,
					       sizeof(srcinfo_t));
					si1->si_fil = fil1;
					si1->si_editblock_id = ed_t;
					memcpy((char *)&eibuf1, (char *)&eibuf,
					       sizeof(editblockinfo_t));
					eibuf1.ei_fil = fil1;
					get_editblock_info(si1->si_editblock_id,
							   &eibuf1);
					make_edesc(&edescbuf1, edit_wn,
						   eibuf1.ei_lines[lnum],
						   width/si1->si_font->ft_width,
						   eibuf1.ei_fg, eibuf1.ei_bg);
					memcpy((char *)&edescbuf1,
					       (char *)&edescbuf,
					       sizeof(edesc_t));
					res = update_lines(si1, &eibuf1,
							   lnum,
							   edescbuf1.ed_copy,
							   nlines, 
							   in_lines ? NULL : &lxbuf);
				      }
				  }
				}
			}
			break;
		case QR_DEL:
			{
				char *joined_line;

				if (lnum == 0) {
					errf("Can't delete past line 1");
					event_nchars = 0;
					break;
				}

				update_lines(si, &eibuf, lnum,
					     edescbuf.ed_copy, nlines,
					     (lexinfo_t *)NULL);

				joined_line = strf("%s%s",
						eibuf.ei_lines[lnum - 1],
						edescbuf.ed_copy);
				event_nchars = strlen(eibuf.ei_lines[lnum - 1]);

				--nlines;
				lines = (const char **)e_malloc(nlines *
								sizeof(char *));
				for (i = 0; i < lnum - 1; ++i)
					lines[i] = eibuf.ei_lines[i];
				lines[i++] = joined_line;
				for (; i < nlines; ++i)
					lines[i] = eibuf.ei_lines[i + 1];

				res = change_editblock(si->si_editblock_id, lines,
							nlines, DONT_CALL_CALLBACK,
							(callback_arg_t)NULL);
				free(joined_line);
				free((char *)lines);
				--lnum;
			}
			break;
		case QR_NEWLINE:
			{
				char *oldpart, *newpart;
				const char *cptr, *oldline;
				int pos;

				update_lines(si, &eibuf, lnum,
					     edescbuf.ed_copy, nlines,
					     (lexinfo_t *)NULL);

				pos = edescbuf.ed_curpos;
				oldline = edescbuf.ed_copy;
				
				oldpart = strf("%.*s", pos, oldline);
				for (cptr = oldline; isspace(*cptr); ++cptr)
					;
				event_nchars = cptr - oldline;
				newpart = strf("%.*s%s",
						       event_nchars, oldline,
						       oldline + pos); 

				++nlines;
				lines = (const char **)e_malloc(nlines *
								   sizeof(char *));
				for (i = 0; i < lnum; ++i)
					lines[i] = eibuf.ei_lines[i];
				lines[i++] = oldpart;
				lines[i++] = newpart;
				for (; i < nlines; ++i)
					lines[i] = eibuf.ei_lines[i - 1];

				res = change_editblock(si->si_editblock_id, lines,
							nlines, DONT_CALL_CALLBACK,
							(callback_arg_t)NULL);
				free(oldpart);
				free(newpart);
				free((char *)lines);
				++lnum;
			}
			break;
		default:
			panic("bad quitres in eel");
		}
		wn_close_window(edit_wn);
	} while (!finished);
}

int
highlight_source(fil, lnum)
fil_t *fil;
int lnum;
{
	srcwin_id_t srcwin_id;

	if (td_set_displayed_source(fil, lnum, "highlight"))
		return 0;

	/* BUG: fil == NULL means turn any highlighting off, but we can't
	 *      call set_highlighted_fil with a NULL fil, so for the
	 *	moment we ignore the problem.
	 */
	if (fil == NULL)
		return 0;

	srcwin_id = get_current_srcwin();

	/*  Convert from 1.. numbering to 0..
	 */
	--lnum;

	set_highlighted_line(fil, lnum);

	if (src_show(srcwin_id, fil, lnum, TRUE, FALSE) != 0)
		return -1;

	echo_src_name_and_lnum(srcwin_id, fil->fi_name, lnum + 1);
	return 0;
}

void
show_source(fil, lnum, centre, check_stat)
fil_t *fil;
int lnum;
bool centre;
bool check_stat;
{
	srcwin_id_t srcwin_id;

	if (td_set_displayed_source(fil, lnum, "showsource"))
		return;

	/*  Convert from 1.. numbering to 0..
	 */
	--lnum;

	srcwin_id = get_current_srcwin();
	src_show(srcwin_id, fil, lnum, centre, check_stat);
	echo_src_name_and_lnum(srcwin_id, fil->fi_name, lnum + 1);
}

void
get_updated_source(fil, lnum, old_nlines, new_nlines)
fil_t *fil;
int lnum;
int old_nlines, new_nlines;
{
	srcwin_id_t srcwin_id;

	if (td_set_displayed_source(fil, lnum, "showsource"))
		return;

	/*  Convert from 1.. numbering to 0..
	 */
	--lnum;

	srcwin_id = get_current_srcwin();
	src_updated(srcwin_id, old_nlines, new_nlines);
}

fil_t *
get_displayed_fil()
{
	srcinfo_t sibuf;

	if (td_get_displayed_fil(&sibuf.si_fil) != 0)
		src_get_info(get_current_srcwin(), 0, 0, &sibuf);
	return sibuf.si_fil;
}

void
source_window_menu_func(arg, md, rv)
char *arg;
int md, rv;
{
	srcwin_id_t srcwin_id;

	srcwin_id = (srcwin_id_t)arg;

	switch (rv) {
	case MR_SRCWIN_BACK:
		src_pop(srcwin_id);
		break;
	case MR_SRCWIN_SEARCH_FORWARDS:
	case MR_SRCWIN_SEARCH_BACKWARDS:
		src_search(srcwin_id, get_typing_line_string(),
						rv == MR_SRCWIN_SEARCH_FORWARDS);
		break;
	case MR_SRCWIN_UP_STACK:
		change_displayed_stack_level(TRUE);
		break;
	case MR_SRCWIN_DOWN_STACK:
		change_displayed_stack_level(FALSE);
		break;
	default:
		panic("bad rv in swmf");
	}
	Mclear(md);
}

/*  Input function for the source display area.  The left button
 *  selects variable or function names, the middle button
 *  produces the forward/back popup menu, and the right button
 *  the add breakpoint/execute to here menu.
 */
/* ARGSUSED */
void
source_window_event_handler(srcwin_id, ev)
srcwin_id_t srcwin_id;
event_t *ev;
{
	srcinfo_t sibuf;
	int buttons;

	src_get_info(srcwin_id, ev->ev_x, ev->ev_y, &sibuf);
	echo_src_name_and_lnum(srcwin_id, 
			       (sibuf.si_fil != NULL) ? sibuf.si_fil->fi_name : NULL,
			       sibuf.si_fil_lnum + 1);

	buttons = ev->ev_buttons & (B_ANY | B_SHIFT_KEY | B_CONTROL_KEY);
	if (ev->ev_buttons & B_LEFT && ev->ev_type == EV_BUTTON_DOWN &&
	    (ev->ev_buttons & B_CONTROL_KEY))
	{
	  event_t evbuf;
	  int last_x, last_y;

	  set_event_win(ev);
	  last_x = ev->ev_x;
	  last_y = ev->ev_y;
	  
	  wn_get_release_event(ev->ev_wn, B_ANY, &evbuf);
	  set_event_win(&evbuf);
	  if (evbuf.ev_wn == ev->ev_wn)
	    source_window_select(srcwin_id, &evbuf, last_x, last_y);
	}

	if ((buttons == B_LEFT || buttons == (B_LEFT | B_SHIFT_KEY)) &&
	    sibuf.si_src_lnum != -1) {
		const char *start, *end;

		if (sibuf.si_editblock_id != NULL) {
			int oldstate;

			oldstate = updating(OBJ_UPDATING_OFF);
			clear_selection();
			select_editblock(sibuf.si_editblock_id, EDL_SELECT,
								DO_CALL_CALLBACK);
			(void) updating(oldstate);
		}
		else if (sibuf.si_cnum != -1 &&
		        get_varname(sibuf.si_fil,
				    sibuf.si_text,
				    sibuf.si_text + sibuf.si_cnum,
				    MAXVARLEN,
				    &start, &end))
/*			confirm_var(ev, srcwin_id, sibuf.si_fil, sibuf.si_src_lnum,
							sibuf.si_text, start, end);*/

/* RGA seems like it needs the file line number here as well
   .. otherwise, if there is lots of breakpoint code, then clicking on a 
   variable at the bottom of a function will fail because si_src_lnum is beyond
   the last file number of the function */
			confirm_var(ev, srcwin_id, sibuf.si_fil, sibuf.si_src_lnum,
				    sibuf.si_fil_lnum, sibuf.si_text, start, end);
		else 
			wn_wait_for_release_of(ev->ev_wn, B_ANY);
	}
	else if (buttons == B_MIDDLE) {
		if (sibuf.si_editblock_id != NULL) {

			/* BUG: this check should probably not be here,
			 *      as this file shouldn't know about the
			 *      meaning of source code.
			 */
			if (!(sibuf.si_fil->fi_language == LANG_C ||
			      sibuf.si_fil->fi_language == LANG_CC)) {
				errf("Sorry, can only add breakpoint code in C functions");
				wn_wait_for_release_of(ev->ev_wn, B_ANY);
			}
			else
				edit_editblock(ev, srcwin_id, &sibuf);
		}
		else {
			errf("Not in an editable section");
			wn_wait_for_release_of(ev->ev_wn, B_ANY);
		}
	}
	else if (buttons == B_RIGHT && sibuf.si_src_lnum != -1) {
		static const char *bptcaps[] = {
			"add breakpoint",
			"execute to here",
                        "edit source",
			NULL
		};
		static popup_t bptpop = { -1, TRUE, 0, bptcaps };
		long sa;
		int res, lnum;
		func_t *f, *f1 = NULL;

		sa = src_draw_arrow(srcwin_id, sibuf.si_src_lnum);
		res = select_from_popup(ev->ev_wn, B_RIGHT, &bptpop,
							ev->ev_x, ev->ev_y);
		wn_restore_area(sa);
		wn_free_area(sa);

		lnum = sibuf.si_fil_lnum + 1;

		switch(res) {
		case 0:
		case 1:
		  f = lnum_to_func(sibuf.si_fil, lnum);
		  if (f != NULL)
		  {
		    td_record_func_and_lnum_cmd(f, lnum,
						(res == 0) ? "breakpoint" : "execto");
		    if (res == 0)
		      add_breakpoint_object(f, f->fu_fil, lnum);
		    if (sibuf.si_fil->fi_flags & FI_DUPLICATE)
		    {
		      fil_t *fil1;
		      
		      if (find_duplicate_file(sibuf.si_fil, &fil1))
		      {
			f1 = lnum_to_func(fil1, lnum);
			
			if (f1 != NULL)
			{
			  td_record_func_and_lnum_cmd
			    (f1, lnum,
			     (res == 0) ? "breakpoint" : "execto");
			  if (res == 0)
			  {
			    add_breakpoint_object(f1, f1->fu_fil, lnum);
			    errf("\bAdding duplicate breakpoint because function has been multiply linked");

			  }
			}
			else
			  errf("Line %d of %s is not within a function",
			       lnum, fil1->fi_name);
		      }
		    }
		    if (res != 0)
		      exec_to_lnum(f, f1, lnum);
		  }
		  else
		    errf("Line %d of %s is not within a function",
			 lnum, sibuf.si_fil->fi_name);
		  break;
                case 2:
                  {
		    char *ed;
		    char buf[MAXPATHLEN + 1];

		    if ((ed = getenv("EDITOR")) == NULL)
		      ed = "vi"; /* default if EDITOR is not set */
		    
		    if (strstr(ed,"emacs"))
		      sprintf
			(buf,
			 "%s +%d %s%s%s&",
			 ed,
			 lnum,
			 sibuf.si_fil->fi_source_path,
			 (sibuf.si_fil->fi_source_path
			  [strlen(sibuf.si_fil->fi_source_path)-1] == '/') ?
			 "":"/",
			 sibuf.si_fil->fi_name);
		    else
		      sprintf	/* vi is default */
			(buf,
			 "xterm -geometry 80x40+200+200 -e %s +%d %s%s%s&",
			 ed,
			 lnum,
			 sibuf.si_fil->fi_source_path,
			 (sibuf.si_fil->fi_source_path
			  [strlen(sibuf.si_fil->fi_source_path)-1] == '/') ?
			 "":"/",
			 sibuf.si_fil->fi_name);
		    system(buf);
		  }
		  break;
		case NO_SELECTION:
		  break;
		default:
		  panic("unknown popup return value");
		}
	      }
      }

int change_to_hourglass_cursor()
{
  cursor_t old_cursor;

  old_cursor = wn_get_window_cursor(WN_STDWIN);
  set_bm_cursor(WN_STDWIN, CU_WAIT);
  return (int)old_cursor;
}

void restore_cursor(old_cursor_int)
int old_cursor_int;
{
  cursor_t old_cursor = (cursor_t)old_cursor_int;

  wn_define_cursor(WN_STDWIN, old_cursor);
}

void
name_window_event_handler(srcwin_id, ev)
srcwin_id_t srcwin_id;
event_t *ev;
{
  srcinfo_t sibuf;
  int buttons;

  src_get_info(srcwin_id, ev->ev_x, ev->ev_y, &sibuf);
  buttons = ev->ev_buttons & (B_ANY | B_SHIFT_KEY | B_CONTROL_KEY);
  if (buttons == B_RIGHT && sibuf.si_src_lnum != -1)
  {
    static const char *bptcaps[] = {
      "edit source",
      "filepath used",
      "filepath assumed",
      "rematch file",
      "reload file",
      "file dates",
      NULL
    };
    static popup_t bptpop = { -1, TRUE, 0, bptcaps };
    int res, lnum;
    func_t *f1 = NULL;
    fil_t *fil;

    fil = sibuf.si_fil;
    res = select_from_popup(ev->ev_wn, B_RIGHT, &bptpop,
			    ev->ev_x, ev->ev_y);

    lnum = sibuf.si_fil_lnum + 1;

    switch(res) {
    case 0:
    {
      char *ed;
      char buf[MAXPATHLEN + 1];

      if ((ed = getenv("EDITOR")) == NULL)
	ed = "vi";		/* default if EDITOR is not set */
		    
      if (strstr(ed,"emacs"))
	sprintf
	  (buf,
	   "%s +%d %s%s%s&",
	   ed,
	   lnum,
	   fil->fi_source_path,
	   (fil->fi_source_path
	    [strlen(fil->fi_source_path)-1] == '/') ?
	   "":"/",
	   fil->fi_name);
      else
	sprintf			/* vi is default */
	  (buf,
	   "xterm -geometry 80x40+200+200 -e %s +%d %s%s%s&",
	   ed,
	   lnum,
	   fil->fi_source_path,
	   (fil->fi_source_path
	    [strlen(fil->fi_source_path)-1] == '/') ?
	   "":"/",
	   fil->fi_name);
      system(buf);
    }
      break;
    case 1:			/* filepath used */
      if (fil && fil->fi_source_path && fil->fi_name)
	if (fil->fi_source_path[strlen(fil->fi_source_path)-1] == '/')
	  errf("\b%s%s",fil->fi_source_path, fil->fi_name);
	else
	  errf("\b%s/%s",fil->fi_source_path, fil->fi_name);
      else
	errf("\b(unknown)");
      break;
    case 2:			/* filepath assumed */
      if (fil && fil->fi_path_hint && fil->fi_name)
	errf("\b%s%s",fil->fi_path_hint, fil->fi_name);
      else
	errf("\b(unknown)");
      break;
    case 3:			/* file rematch */
      if (fil->fi_name)
	open_source_file(fil, TRUE, TRUE, FALSE);
      show_source(fil, 1, FALSE, TRUE);
      break;
    case 4:			/* file reload */
      if (fil->fi_name)
	open_source_file(fil, TRUE, FALSE, TRUE);
      show_source(fil, 1, FALSE, TRUE);
      break;
    case 5:			/* file dates */
      show_file_and_target_dates(fil);
      break;
    case NO_SELECTION:
      break;
    default:
      panic("unknown popup return value");
    }
  }
}

