/* va_type - type handling routines */

/*  Copyright 1991 Mark Russell, University of Kent at Canterbury.
 *
 *  You can do what you like with this source code as long as
 *  you don't try to make money out of it and you include an
 *  unaltered copy of this message (including the copyright).
 */

char ups_va_type_c_sccsid[] = "@(#)va_type.c	1.17 20/5/92 (UKC)";

#include <mtrprog/ifdefs.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <local/wn.h>
#include <local/ukcprog.h>
#include <local/obj/obj.h>
#include <mtrprog/alloc.h>

#include "objtypes.h"
#include "ups.h"
#include "symtab.h"
#include "data.h"
#include "preamble.h"
#include "va.h"
#include "va_priv.h"
#include "ci.h"

#include "obj_stack.h"
#include "obj_misc.h"
#include "obj_bpt.h"
#include "obj_signal.h"
#include "stack.h"
#include "src.h"
#include "expr.h"

static void note_file_collapsed PROTO((objid_t obj));
static void update_file_entry PROTO((objid_t obj));


/************** Begin util.c from version 3.14 *************************/
bool
parse_line(line, p_words, p_nwords)
const char *line;
char ***p_words;
int *p_nwords;
{
	ebuf_t *eb;
	int quotec, nbytes, nwords, i;
	const char *iptr;
	char *obuf, *optr, *word;
	char **words, **argv;
	bool new_word, word_exists;

	/*  No escape sequence makes a line longer, so the length of
	 *  the line is an upper bound on the length of the processes line.
	 */
	optr = word = obuf = e_malloc(strlen(line) + 1);
	eb = ebuf_create(TRUE);

	iptr = line;
	quotec = '\0';
	new_word = TRUE;
	word_exists = FALSE;

	while (isspace(*line))
		++line;
		
	for (iptr = line; *iptr != '\0'; ++iptr) {
		if (*iptr == '\\') {
			++iptr;

			if (*iptr == '\0') {
				errf("Backslash at end of line");
				return FALSE;
			}
			else {
				static const char escapes[] = "t\tn\nb\br\r";
				const char *pos;
				int ch;
				
				pos = strchr(escapes, *iptr);

				if (pos != NULL && (pos - escapes) % 2 == 0)
					ch = pos[1];
				else
					ch = *iptr;

				*optr++ = ch;
			}
		}
		else if (*iptr == quotec) {
			quotec = '\0';
		}
		else if ((*iptr == '\'' || *iptr == '"') && quotec == '\0') {
			quotec = *iptr;
			word_exists = TRUE;
		}
		else if (isspace(*iptr) && quotec == '\0') {
			while (isspace(iptr[1]))
				++iptr;
			
			ebuf_add(eb, &word, sizeof(word));

			*optr++ = '\0';
			word = optr;
			word_exists = FALSE;
		}
		else {
			*optr++ = *iptr;
		}
	}

	if (quotec != '\0') {
		errf("Missing closing %s quote",
		     (quotec == '"') ? "double" : "single");
		return FALSE;
	}

	if (optr > word || word_exists) {
		ebuf_add(eb, &word, sizeof(word));
		*optr++ = '\0';
	}
	
	words = (char **)ebuf_get(eb, &nbytes);
	nwords = nbytes / sizeof(char *);

	/*  Make a copy of the vector and argument strings in a single
	 *  buffer.  We need to allow space for the final NULL in argv.
	 *
	 *  The caller frees this when they have finished with it.
	 */
	argv = (char **)e_malloc((size_t)(nwords + 1) * sizeof(char *) +
				 (optr - obuf));
	optr = (char *)&argv[nwords + 1];

	for (i = 0; i < nwords; ++i) {
		argv[i] = strcpy(optr, words[i]);
		optr += strlen(optr) + 1;
	}
	argv[i] = NULL;

	ebuf_free(eb);
	free(obuf);

	*p_words = argv;
	*p_nwords = nwords;
	return TRUE;
}

void
put_quoted_string(fp, str)
FILE *fp;
const char *str;
{
	const char *s;

	fputc('"', fp);
	
	for (s = str; *s != '\0'; ++s) {
		if (*s == '\\' || *s == '"')
			fputc('\\', fp);
		fputc(*s, fp);
	}

	fputc('"', fp);
}

void
dump_text(text, fp)
const char *text;
FILE *fp;
{
	bool at_line_start;
	
	at_line_start = TRUE;
	
	for (; *text != '\0'; ++text) {
		if (at_line_start)
			putc('\t', fp);
		
		putc(*text, fp);
		
		at_line_start = *text == '\n';
	}

	if (!at_line_start)
		putc('\n', fp);
}

bool
same_string(s1, s2)
const char *s1, *s2;
{
	return s1 == s2 || (s1 != NULL && s2 != NULL && strcmp(s1, s2) == 0);
}

bool
get_debug_output_path(cmdline, defpath, p_cmdline, p_path, p_overwrite)
const char *cmdline, *defpath;
const char **p_cmdline, **p_path;
bool *p_overwrite;
{
	const char *new_cmdline, *path;
	bool overwrite;

	if ((path = strchr(cmdline, '>')) == NULL) {
		path = defpath;
		new_cmdline = cmdline;
		overwrite = TRUE;
	}
	else {
		static char *last;
		const char *pos;

		for (pos = path; pos > cmdline && pos[-1] == ' '; --pos)
			;

		overwrite = path[1] == '!';
		if (overwrite)
			++path;

		for (++path; isspace(*path); ++path)
			;
		if (*path == '\0') {
			errf("Redirection to zero length output file");
			return FALSE;
		}

		if (last != NULL)
			free(last);
		last = strf("%.*s", pos - cmdline, cmdline);
		
		new_cmdline = last;
	}

	*p_cmdline = new_cmdline;
	*p_path = path;
	*p_overwrite = overwrite;

	return TRUE;
}

/************** End util.c from version 3.14 ***************************/


/************** Begin lib/libmtrprog/fttostr.c from version 3.14 *******/
const char *
filetype_to_string(mode)
int mode;
{
	static char buf[100];

	if (S_ISREG(mode))
		return "regular file";
	if (S_ISDIR(mode))
		return "directory";
	if (S_ISCHR(mode))
		return "character special file";
	if (S_ISBLK(mode))
		return "block special file";
	if (S_ISFIFO(mode))
		return "named pipe";
#ifdef S_ISSOCK
	if (S_ISSOCK(mode))
		return "socket";
#endif
#ifdef S_ISLNK
	if (S_ISLNK(mode))
		return "symbolic link";
#endif
	
	sprintf(buf, "file with unknown mode 0%o", mode);
	return buf;
}
/************** End lib/libmtrprog/fttostr.c from version 3.14 *********/


/************** Begin config.c from version 3.14 ***************************/
static void show_statefile_error PROTO((const char *mesg));

static errf_ofunc_t Old_ofunc;
static const char *Errs_filename;
static int Errs_lnum;

static void
show_statefile_error(mesg)
const char *mesg;
{
	if (Errs_lnum > 0) {
		char *copy;
		
		copy = strf("%s,%d: %s", Errs_filename, Errs_lnum, mesg);
		(*Old_ofunc)(copy);
		free(copy);
	}
	else {
		(*Old_ofunc)(mesg);
	}
}

int
save_state(state_dir, state_path)
const char *state_dir, *state_path;
{
	struct stat stbuf;
	FILE *fp;
	int res = 0;
	const char *dot;
	
	if (stat(state_dir, &stbuf) != 0) {
		if (errno != ENOENT) {
			errf("Can't stat", "state directory", state_dir);
			return -1;
		}

		if (mkdir(state_dir, 0777) != 0) {
			errf("Can't create", "state directory", state_dir);
			return -1;
		}
	}
	else if (!S_ISDIR(stbuf.st_mode)) {
		errf("%s should be a directory, but is a %s", state_dir,
		     filetype_to_string((int)stbuf.st_mode));
		return -1;
	}

	if ((fp = fopen(state_path, "w")) == NULL) {
		errf("Can't create", "ups state file", state_path);
		return -1;
	}

	dot = strrchr(state_path, '.');
	fputs("# Do not edit this file - ", fp);
	fputs("it is overwritten automatically by ups.\n", fp);
	fprintf(fp, "# Edit %.*s.config, .upsrc or $HOME/.upsrc instead.\n\n",
		dot - state_path, state_path);

	res = save_all_breakpoints_to_file(state_path, fp); 

	if (res == 0)
		res = write_var_state_to_file(state_path, fp);
	
 	if (res == 0)
 		res = save_signal_state_to_file(fp);
 
	if (ferror(fp) || fflush(fp) == EOF) {
		errf("Error writing to", "ups state file", state_path);
		res = -1;
	}

	fclose(fp);
	
	return res;
}

void
make_config_paths(state_dir, basepath, p_state_path, p_user_path)
const char *state_dir, *basepath, **p_state_path, **p_user_path;
{
	const char *target_name, *suf;
	int len;

	if ((target_name = strrchr(basepath, '/')) != NULL)
		++target_name;
	else
		target_name = basepath;

	if ((suf = strrchr(target_name, '.')) != NULL)
		len = suf - target_name;
	else
		len = strlen(target_name);

	*p_state_path = strf("%s/%.*s.state", state_dir, len, target_name);
	*p_user_path = strf("%s/%.*s.config", state_dir, len, target_name);
}

void
load_config(state_path, user_path, p_want_auto_start)
const char *state_path, *user_path;
bool *p_want_auto_start;
{
	const char *home;
	bool want_auto_start;
	
	want_auto_start = FALSE;

	if ((home = getenv("HOME")) != NULL) {
		char *path;

		path = strf("%s/.upsrc", home);
		read_config_file(path, FALSE, FALSE, FALSE, FALSE,
				 &want_auto_start);
	}
		
	read_config_file(".upsrc", FALSE,  FALSE, FALSE, FALSE,
			 &want_auto_start);
	read_config_file(user_path, FALSE,  FALSE, FALSE, FALSE,
			 &want_auto_start);
	read_config_file(state_path, TRUE,  FALSE, FALSE, TRUE,
			 &want_auto_start);

	set_breakpoint_statefile_path(state_path); 

	*p_want_auto_start = want_auto_start;
}

int
read_config_file(path, from_statefile, must_exist, breakpoints_only,
		 ignore_breakpoints, p_want_auto_start)
const char *path;
bool from_statefile, must_exist, breakpoints_only, ignore_breakpoints;
bool *p_want_auto_start;
{
	FILE *fp;
	char *line;
	int res, old_message_wn;
	
	if ((fp = fopen(path, "r")) == NULL) {
		if (must_exist) {
			errf("Can't open", "state file", path);
			return -1;
		}
		
		if (errno != ENOENT)
			errf("Warning: can't open", "state file", path);

		return 0;
	}

	Old_ofunc = errf_set_ofunc(show_statefile_error);

	/*  As there may be lots of errors, send them to stderr.
	 */
/*	old_message_wn = get_message_wn();*/
/*	set_message_wn(-1);*/

	Errs_filename = path;
	Errs_lnum = 0;

	res = 0;
	
	while ((line = fpgetline(fp)) != NULL) {
		char **fields, *cmd, **args;
		int nargs;

		++Errs_lnum;

		line = config_trim_line(line);
		
		fields = ssplit(line, " \t");
		
		if (fields[0] == NULL) {
			free((char *)fields);
			continue;
		}

		cmd = fields[0];
		args = &fields[1];

		for (nargs = 0; args[nargs] != NULL; ++nargs)
			;

		if (strcmp(cmd, "breakpoint") == 0) {
			if (handle_add_breakpoint_command(cmd, args, nargs,
							  from_statefile,
							  ignore_breakpoints,
							  fp, &Errs_lnum)!= 0) {
				res = -1;
			}
		}
		else if (strcmp(cmd, "code") == 0) {
			if (handle_add_code_command(cmd, args, nargs,
						    from_statefile,
						    ignore_breakpoints,
						    fp, &Errs_lnum)!= 0) {
				res = -1;
			}
		}
		else if (breakpoints_only) {
			/* Skip rest (tmp bodge pending command table) */
		}
		else if (strcmp(cmd, "format") == 0) {
			if (handle_format_command(cmd, args, nargs, FALSE) != 0)
				res = -1;
		}
		else if (strcmp(cmd, "format-hint") == 0) {
			if (handle_format_command(cmd, args, nargs, TRUE) != 0)
				res = -1;
		}
		else if (strcmp(cmd, "file") == 0 ||
			 strcmp(cmd, "common") == 0 ||
			 strcmp(cmd, "globals") == 0) {
			if (handle_file_command(cmd, args, nargs,
							  from_statefile,
							  fp, &Errs_lnum) != 0)
				res = -1;
		}
		else if (strcmp(cmd, "function") == 0) {
			if (handle_function_command(cmd, args, nargs,
							  from_statefile,
							  fp, &Errs_lnum) != 0)
				res = -1;
		}
		else if (strcmp(cmd, "auto-start") == 0) {
			if (nargs != 1 || (strcmp(*args, "yes") != 0 &&
					   strcmp(*args, "no") != 0)) {
				errf("Usage: auto-start yes|no");
			}
			else {
				*p_want_auto_start = strcmp(*args, "yes") == 0;
			}
		}
		else if (strcmp(cmd, "echo") == 0) {
			bool want_newline;
			
			want_newline = strcmp(*args, "-n") != 0;
			if (!want_newline)
				++args;

			while (*args != NULL) {
				fputs(*args++, stdout);
				if (*args != NULL)
					fputc(' ', stdout);
			}

			if (want_newline)
				fputc('\n', stdout);

			fflush(stdout);
		}
 		else if (strcmp(cmd, "signal") == 0) {
 			if (handle_signal_command(cmd, args, nargs, FALSE) != 0)
 				res = -1;
  		}
		else {
			errf("Unrecognised command `%s'", cmd);
		}

		free((char *)fields);
	}

	fclose(fp);

	errf_set_ofunc(Old_ofunc);
/*	set_message_wn(old_message_wn);*/

	return res;
}	
/************** End config.c from version 3.14 ***************************/


/************** Begin va_format.c from version 3.14 ********************/

#define DEFAULT_ILEVEL	(-1)

typedef struct Format {
	fil_t *fil;
	func_t *func;
	bool is_struct_member;
	const char *tag;
	const char *name;
	vformat_t format;
	int ilevel;
	bool is_hint;
	struct Format *next;
} Format;

enum {
	SC_ANY_MATCH,		/* match on null pointer or `*' wildcard */
	SC_EXACT_MATCH		/* non-glob match */
};

static void push_format_entry PROTO((fil_t *fil, func_t *f,
				     bool is_struct_member,
				     const char *tag, const char *name, 
                                     vformat_t format, int ilevel,
				     bool is_hint));
static void dump_format PROTO((FILE *fp, Format *f));
static void get_var_info PROTO((dvar_t *dv, fil_t **p_fil, func_t **p_func, 
                                const char **p_tag, bool *p_is_struct_member));
static bool matchstr PROTO((const char *pat, const char *str, bool exact,
			    int *p_score));
static bool matchptr PROTO((bool is_null, bool same, bool exact, int *p_score));
static Format *reverse_list PROTO((Format *list));
static bool lookup_format PROTO((fil_t *fil, func_t *func, const char *tag, 
                                 bool is_struct_member, const char *name, 
                                 bool exact, Format **p_ft));

static Format *Format_list = NULL;


int
handle_format_command(cmd, args, nargs, is_hint)
const char *cmd;
char **args;
int nargs;
bool is_hint;
{
	const char *format_str;
	char *varstr, *tag, *varname, *filestr, *funcstr, *ilevel_str;
	int ilevel;
	vformat_t format;
	bool is_struct_member;
	fil_t *fil;
	func_t *f, *f1;
	
	if (nargs != 3 && nargs != 4) {
		errf("Usage: %s file[:func] [[tag].]varname format [indir-level]", cmd);
		return -1;
	}
	filestr = args[0];
	varstr = args[1];
	format_str = args[2];
	ilevel_str = args[3];

	funcstr = strchr(filestr, ':');

	if (strcmp(filestr, "*") == 0) {
		funcstr = filestr = NULL;
	}
	else if (funcstr != NULL) {
		*funcstr++ = '\0';
	}
	else if (strchr(filestr, '.') == NULL) {
		funcstr = filestr;
		filestr = NULL;
	}

	if (filestr == NULL) {
		fil = NULL;
	}
	else {
		if (*filestr == '\0') {
			errf("Zero length file name");
			return -1;
		}

		if ((fil = name_to_fil(filestr)) == NULL) {
			errf("Unknown source file name `%s'", filestr);
			return -1;
		}
	}

	if (funcstr == NULL) {
		f = NULL;
	}
	else {
		if (*funcstr == '\0') {
			errf("Zero length function name");
			return -1;
		}

		if (fil != NULL) {
			if ((f = name_and_fil_to_func(funcstr, fil)) == NULL) {
				if (find_func_by_name(funcstr, &f, &f1, FALSE) != 0)
					return -1;;
			}
		}
		else {
			if (find_func_by_name(funcstr, &f, &f1, FALSE) != 0)
				return -1;
		}
	}
		
	if (!string_to_format(format_str, &format))
		return -1;

	if (ilevel_str == NULL) {
		ilevel = DEFAULT_ILEVEL;
	}
	else if (!get_num(&ilevel_str, "indirection level", &ilevel)) {
		return -1;
	}

	if ((varname = strchr(varstr, '.')) == NULL) {
		tag = NULL;
		varname = varstr;
		is_struct_member = FALSE;
	}
	else {
		*varname++ = '\0';
	
		if (*varname == '\0') {
			errf("Zero length variable name");
			return -1;
		}

		tag = (*varstr != '\0') ? strsave(varstr) : NULL;

		is_struct_member = TRUE;
	}
	
	push_format_entry(fil, f, is_struct_member, tag, strsave(varname),
			  format, ilevel, is_hint);
	return 0;
}

static Format *
reverse_list(list)
Format *list;
{
	Format *ft, *next, *newlist;

	newlist = NULL;

	for (ft = list; ft != NULL; ft = next) {
		next = ft->next;
		ft->next = newlist;
		newlist = ft;
	}

	return newlist;
}

int
write_format_state_to_file(path, fp)
const char *path;
FILE *fp;
{
	Format *ft;
	
	if (Format_list != NULL)
		fputc('\n', fp);

	Format_list = reverse_list(Format_list);
	
	for (ft = Format_list; ft != NULL; ft = ft->next) {
		dump_format(fp, ft);
		if (ferror(fp))
			return -1;
	}

	Format_list = reverse_list(Format_list);
	
	return 0;
}

static void
dump_format(fp, ft)
FILE *fp;
Format *ft;
{
	if (!ft->is_hint)
		return;
	
	fputs("format-hint ", fp);

	if (ft->fil != NULL) {
		fputs(ft->fil->fi_name, fp);

		if (ft->func != NULL)
			fprintf(fp, ":%s", ft->func->fu_name);
	}
	else {
		fputs((ft->func != NULL) ? ft->func->fu_name : "*", fp);
	}
	fputc(' ', fp);

	if (ft->is_struct_member) {
		if (ft->tag != NULL)
			fputs(ft->tag, fp);
		fputc('.', fp);
	}

	fputs(ft->name, fp);
	fputc(' ', fp);

	fputs(format_to_string(ft->format), fp);

	if (ft->ilevel != DEFAULT_ILEVEL)
		fprintf(fp, " %d", ft->ilevel);

	fputc('\n', fp);
}

static void
push_format_entry(fil, f, is_struct_member, tag, name, format, ilevel, is_hint)
fil_t *fil;
func_t *f;
bool is_struct_member;
const char *tag, *name;
vformat_t format;
int ilevel;
bool is_hint;
{
	Format *ft;

	ft = (Format *)e_malloc(sizeof(Format));
	ft->fil = fil;
	ft->func = f;
	ft->is_struct_member = is_struct_member;
	ft->tag = tag;
	ft->name = name;
	ft->format = format;
	ft->ilevel = ilevel;
	ft->is_hint = is_hint;
	ft->next = Format_list;
	Format_list = ft;
}

static bool
matchstr(pat, str, exact, p_score)
const char *pat, *str;
bool exact;
int *p_score;
{
	int score;
	
	if (pat == NULL) {
		if (exact && str != NULL)
			return FALSE;
			
		score = SC_ANY_MATCH;
	}
	else if (str != NULL && strcmp(pat, str) == 0) {
		score = SC_EXACT_MATCH;
	}
	else {
		return FALSE;
	}

	*p_score = score;
	return TRUE;
}

static bool
matchptr(is_null, same, exact, p_score)
bool is_null, same, exact;
int *p_score;
{
	int score;

	if (same) {
		score = SC_EXACT_MATCH;
	}
	else if (is_null) {
		if (exact)
		       return FALSE;
		
		score = SC_ANY_MATCH;
	}
	else {
		return FALSE;
	}

	*p_score = score;
	return TRUE;
}

static bool
lookup_format(fil, func, tag, is_struct_member, name, exact, p_ft)
fil_t *fil;
func_t *func;
const char *tag;
bool is_struct_member;
const char *name;
bool exact;
Format **p_ft;
{
	Format *ft;
	int bestscore;

	*p_ft = NULL;
	bestscore = 0;

	for (ft = Format_list; ft != NULL; ft = ft->next) {
		int tagscore, filscore, funcscore, namescore;
		
		if (is_struct_member == ft->is_struct_member &&
		    matchstr(ft->tag, tag, exact, &tagscore) &&
		    matchptr(ft->func == NULL, ft->func == func, exact,
			     &funcscore) &&
		    matchptr(ft->fil == NULL, ft->fil == fil, exact,
			     &filscore) &&
		    matchstr(ft->name, name, exact, &namescore)) {
			int score;
			
			score = tagscore + funcscore + filscore + namescore;

			if (score > bestscore) {
				*p_ft = ft;
				bestscore = score;
			}
		}
	}

	return *p_ft != NULL;
}
	
bool
maybe_restore_format(dv)
dvar_t *dv;
{
	Format *ft;
	fil_t *fil;
	func_t *func;
	const char *tag;
	bool is_struct_member;
	
	get_var_info(dv, &fil, &func, &tag, &is_struct_member);
	
	if (!lookup_format(fil, func, tag, is_struct_member,
			   dv->dv_var->va_name, FALSE, &ft))
		return FALSE;

	if (ft->ilevel != DEFAULT_ILEVEL) {
		int i;
		
		for (i = dv->dv_ilevel; i < ft->ilevel; ++i)
			change_dv_level(dv, CL_UP);

		for (i = ft->ilevel; i < dv->dv_ilevel; ++i)
			change_dv_level(dv, CL_DOWN);
	}

	if (ft->format != DF_NONE) {
		dv->dv_format = ft->format;
	}
	else {
		type_t *type;

		type = get_type_at_level(dv->dv_var, dv->dv_ilevel);
		dv->dv_format = default_format(dv->dv_var->va_type, type);
	}

	redo_decl(dv);

	return TRUE;
}

static void
get_var_info(dv, p_fil, p_func, p_tag, p_is_struct_member)
dvar_t *dv;
fil_t **p_fil;
func_t **p_func;
const char **p_tag;
bool *p_is_struct_member;
{
	bool is_struct_member;
	const char *tag;
	fil_t *fil;
	func_t *func;
	objid_t par;
	int partype;
	taddr_t junk;
			
	is_struct_member = (dv->dv_var->va_class == CL_MOU ||
			    dv->dv_var->va_class == CL_MOS);

	/*  Structure member formats probably want to be applied everywhere.
	 */
	if (is_struct_member) {
		type_t *type;
		
		fil = NULL;
		func = NULL;
		
		par = get_code((objid_t)dv, OBJ_PARENT);
		
		switch (ups_get_object_type(par)) {
		case OT_VAR:
			type = get_basetype(((dvar_t *)par)->dv_var->va_type);
			break;
		case OT_EXPR:
			type = get_basetype(get_expr_type(par));
			break;
		default:
			panic("bad partype in nfc");
			type = NULL; /* to satisfy gcc */
		}

		switch (type->ty_code) {
		case TY_STRUCT:
		case TY_UNION:
		case TY_U_STRUCT:
		case TY_U_UNION:
			break;
		default:
			panic("expected aggr basetype in nfc");
		}

		if (type->ty_typedef != NULL)
			tag = type->ty_typedef->td_name;
		else
			tag = type->ty_aggr_or_enum->ae_tag;
	}
	else {
		par = (objid_t)dv;
		tag = NULL;
	}
	
	do {
		par = get_code(par, OBJ_PARENT);
		partype = ups_get_object_type(par);
	} while (partype == OT_VAR || partype == OT_EXPR);
	
	switch (partype) {
	case OT_FUNC:
	case OT_BLOCK:
		func = get_stack_func(par, &junk, &junk);
		fil = func->fu_fil;
		break;
	case OT_SFILE:
		fil = (fil_t *)par;
		func = NULL;
		break;
	case OT_GLOBALS:
		fil = NULL;
		func = NULL;
		break;
	case OT_CBLOCK:
		func = NULL;
/*		fil = get_cblock_fil((common_block_t *)par);*/ 
		fil = NULL;	/* RGA */
		break;
	default:
		panic("unknown par type in nfc");
		fil = NULL;	/* to satisfy gcc */
		func = NULL;	/* to satisfy gcc */
	}

	*p_fil = fil;
	*p_func = func;
	*p_tag = tag;
	*p_is_struct_member = is_struct_member;
}

void
note_format_change(dv)
dvar_t *dv;
{
	Format *ft;
	bool is_struct_member;
	const char *tag;
	fil_t *fil;
	func_t *func;

	get_var_info(dv, &fil, &func, &tag, &is_struct_member);
	
	/*  Don't record the source file for a local, as the function is
	 *  probably sufficient context.  This also means the user can
	 *  move the function to a different source file without problems.
	 */
	if (func != NULL)
		fil = NULL;

	/*  Don't record the source file or function for a struct member.
	 */
	if (is_struct_member) {
		func = NULL;
		fil = NULL;
	}
	
	if (lookup_format(fil, func, tag, is_struct_member, dv->dv_var->va_name,
			  TRUE, &ft)) {
		if (ft->is_hint) {
			ft->format = dv->dv_format;
			ft->ilevel = dv->dv_ilevel;
		}
		return;
	}

	push_format_entry(fil, func, is_struct_member, tag,
			  dv->dv_var->va_name,
			  dv->dv_format, dv->dv_ilevel, TRUE);
}


bool
string_to_format(str, p_format)
const char *str;
vformat_t *p_format;
{
	static struct {
		const char *str;
		vformat_t format;
	} tab[] = {
		{ "default",		DF_NONE		},
		{ "signed_decimal",	DF_SDEC		},
		{ "unsigned_decimal",	DF_UDEC		},
		{ "signed_octal",	DF_SOCT		},
		{ "unsigned_octal",	DF_UOCT		},
		{ "signed_hex",		DF_SHEX		},
		{ "unsigned_hex",	DF_UHEX		},
		{ "ascii",		DF_ASCII	},
		{ "string",		DF_STRING	},
		{ "unsigned_binary",	DF_UBIN		}
	};
	int i;

	for (i = 0; i < sizeof tab / sizeof *tab; ++i) {
		if (strcmp(tab[i].str, str) == 0) {
			*p_format = tab[i].format;
			return TRUE;
		}
	}

	errf("Unknown format `%s'", str);
	return FALSE;
}

const char *
format_to_string(format)
vformat_t format;
{
	switch (format) {
	case DF_NONE:	return "default";
	case DF_SDEC:	return "signed_decimal";
	case DF_UDEC:	return "unsigned_decimal";
	case DF_SOCT:	return "signed_octal";
	case DF_UOCT:	return "unsigned_octal";
	case DF_SHEX:	return "signed_hex";
	case DF_UHEX:	return "unsigned_hex";
	case DF_ASCII:	return "ascii";
	case DF_STRING:	return "string";
	case DF_UBIN:	return "unsigned_binary";
	}

	panic("bad format in fts");
	return NULL;		/* to satisfy gcc */
}

/************** End va_format.c from version 3.14 **********************/


/************** Begin va_state.c from version 3.14 ***********************/

typedef enum {
	VT_VAR,
	VT_EXPR
} Vetype;

typedef struct {
	const char *name;
	vformat_t format;
	bool no_indent;
	int ilevel;
	int icount;
	int *indices;
} Vsvar;

typedef struct {
	const char *text;
	vformat_t format;
} Vsexpr;

typedef struct velist_s {
	Vetype type;
	int block_start;
	int block_end;
	union {
		Vsvar *var;
		Vsexpr *expr;
	} u;
	struct velist_s *children;
	struct velist_s *next;
} Velist;

typedef struct vstack_s {
	alloc_id_t apool;
	func_t *func;
	Velist *velist;
	bool updated;
	bool been_displayed;
	struct vstack_s *next;
} Vstack;

typedef struct vfile_s {
	alloc_id_t apool;
	objid_t obj;
	int objtype;
	Velist *velist;
	bool been_displayed;
	bool user_changed_vars;
	struct vfile_s *next;
} Vfile;

static void note_file_display_change PROTO((objid_t obj));
static Vstack *func_to_vstack PROTO((func_t *f));
static Vstack *push_new_vstack PROTO((func_t *f, alloc_id_t ap,
					Velist *velist));
static bool save_any_stack_changes PROTO((objid_t obj, func_t *f, char *args));
static Velist *push_ve PROTO((alloc_id_t ap, Velist **p_velist,
				block_t *bl, objid_t par, Vetype vetype));
static void push_var_state PROTO((alloc_id_t ap, Velist **p_velist,
				  objid_t par, dvar_t *dv));
static void push_expr_state PROTO((alloc_id_t ap, Velist **p_velist,
				   objid_t par, objid_t obj));
static void push_children PROTO((alloc_id_t ap, Velist **p_velist,
				 objid_t par));
static void restore_display PROTO((objid_t par, Velist *velist));
static objid_t restore_var PROTO((objid_t par, Vsvar *vv,
			       int block_start, int block_end));
static void make_vv_ilist PROTO((dvar_t *dv, int *indices, int count));
static objid_t restore_expr PROTO((objid_t par, Vsexpr *vc,
				int block_start, int block_end));
static Vfile *obj_to_vfile PROTO((objid_t obj));
static block_t *lnum_range_to_block PROTO((block_t *parbl,
					   int start_lnum, int end_lnum));
static void restore_children PROTO((objid_t par, Velist *children));
static void restore_dv_state PROTO((Vsvar *vv, dvar_t *dv));
static int dump_children PROTO((const char *path, FILE *fp,
				Velist *velist, int level));
static void dump_var PROTO((FILE *fp, Vsvar *vv));
static void dump_expr PROTO((FILE *fp, Vsexpr *vc));
static int read_velist PROTO((alloc_id_t ap, Velist **p_velist,
			      FILE *fp, int line_offset, int *p_lnum));
static int read_var PROTO((alloc_id_t ap, char *line, Vsvar *vv));
static int read_expr PROTO((alloc_id_t ap, char *line, Vsexpr *vc));
static bool get_word PROTO((char **p_line, const char *what, char **p_cmd));
static Vfile *push_new_vfile PROTO((objid_t obj, int objtype, alloc_id_t ap,
				    Velist *velist));
static int load_velist PROTO((FILE *fp, int line_offset, int *p_lnum,
			      bool failed,
			      alloc_id_t *p_ap, Velist **p_velist));
static Velist *reverse_velist PROTO((Velist *velist));
static void update_vf PROTO((Vfile *vf));
static bool update_stack_entry_if_needed PROTO((stack_t *stk, bool force));

static Vstack *Vslist = NULL;
static Vfile *Vflist = NULL;

void
note_var_change(obj)
objid_t obj;
{
	for (;;) {
		obj = get_code(obj, OBJ_PARENT);

		if (obj == NULL)
			break;
		
		switch (ups_get_object_type(obj)) {
		case OT_SFILE:
		case OT_GLOBALS:
		case OT_CBLOCK:
			note_file_display_change(obj);
			break;
		case OT_FUNC:
			((stack_t *)obj)->stk_user_changed_vars = TRUE;
			break;
		}
	}

}

static Vfile *
obj_to_vfile(obj)
objid_t obj;
{
	Vfile *vf;

	for (vf = Vflist; vf != NULL; vf = vf->next) {
		if (vf->obj == obj)
			return vf;
	}

	return NULL;
}

static void
note_file_display_change(obj)
objid_t obj;
{
	Vfile *vf;

	if ((vf = obj_to_vfile(obj)) == NULL) {
		vf = push_new_vfile(obj, ups_get_object_type(obj),
				    alloc_create_pool(), (Velist *)NULL);
	}

	vf->user_changed_vars = TRUE;
}

static Vfile *
push_new_vfile(obj, objtype, ap, velist)
objid_t obj;
int objtype;
alloc_id_t ap;
Velist *velist;
{
	Vfile *vf;
	
	vf = (Vfile *)e_malloc(sizeof(Vfile));
	vf->apool = ap;
	vf->obj = obj;
	vf->objtype = objtype;
	vf->velist = velist;
	vf->next = Vflist;
	vf->user_changed_vars = FALSE;
	vf->been_displayed = FALSE;
	Vflist = vf;

	return vf;
}


static Vstack *
func_to_vstack(f)
func_t *f;
{
	Vstack *vs;

	for (vs = Vslist; vs != NULL; vs = vs->next) {
		if (vs->func == f)
			return vs;
	}

	return NULL;
}

static Vstack *
push_new_vstack(f, ap, velist)
func_t *f;
alloc_id_t ap;
Velist *velist;
{
	Vstack *vs;
	
	vs = (Vstack *)e_malloc(sizeof(Vstack));
	vs->apool = ap;
	vs->func = f;
	vs->velist = velist;
	vs->updated = FALSE;
	vs->been_displayed = FALSE;
	vs->next = Vslist;
	Vslist = vs;

	return vs;
}

static bool
save_any_stack_changes(obj, f, args)
objid_t obj;
func_t *f;
char *args;
{
	return update_stack_entry_if_needed((stack_t *)obj, FALSE);
}

void
update_stack_entry(stk)
stack_t *stk;
{
	update_stack_entry_if_needed(stk, TRUE);
}

static bool
update_stack_entry_if_needed(stk, force)
stack_t *stk;
bool force;
{
	Vstack *vs;

	if (!stk->stk_user_changed_vars)
		return FALSE;

	if ((vs = func_to_vstack(stk->stk_func)) != NULL) {
		
		/*  We ignore all but the innermost stack entry for a
		 *  given function.
		 */
		if (vs->updated && !force)
			return FALSE;
		
		alloc_reset_pool(vs->apool);
		vs->velist = NULL;
	}
	else {
		vs = push_new_vstack(stk->stk_func, alloc_create_pool(),
				     (Velist *)NULL);
	}

	push_children(vs->apool, &vs->velist, (objid_t)stk);
	vs->velist = reverse_velist(vs->velist);
	vs->updated = TRUE;
	vs->been_displayed = TRUE;
	
	stk->stk_user_changed_vars = FALSE;
	return FALSE;
}

static Velist *
push_ve(ap, p_velist, bl, par, vetype)
alloc_id_t ap;
Velist **p_velist;
block_t *bl;
objid_t par;
Vetype vetype;
{
	objid_t obj;
	Velist *ve, *children;

	ve = (Velist *)alloc(ap, sizeof(Velist));
	ve->type = vetype;
	
	if (bl != NULL) {
		ve->block_start = bl->bl_start_lnum;
		ve->block_end = bl->bl_end_lnum;
	}
	else {
		ve->block_start = 0;
		ve->block_end = 0;
	}

	children = NULL;
	
	for (obj = get_code(par, OBJ_CHILD);
	     obj != NULL;
	     obj = get_code(obj, OBJ_NEXT)) {
		if (ups_get_object_type(obj) == OT_VAR) {
			push_var_state(ap, &children,
				       par, (dvar_t *)obj);
		}
	}

	ve->children = reverse_velist(children);
	ve->next = *p_velist;
	*p_velist = ve;

	return ve;
}

static void
push_var_state(ap, p_velist, par, dv)
alloc_id_t ap;
Velist **p_velist;
objid_t par;
dvar_t *dv;
{
	Velist *ve;
	Vsvar *vv;
	ilist_t *il;
	int icount, *indices;
	block_t *bl;

	icount = 0;
	for (il = dv->dv_ilist; il != NULL; il = il->il_next)
		++icount;

	if (icount == 0) {
		indices = NULL;
	}
	else {
		int *ic;
		
		indices = (int *)alloc(ap, icount * sizeof(int));

		if (dv->dv_var->va_language == LANG_FORTRAN) {
			ic = &indices[icount];
			for (il = dv->dv_ilist; il != NULL; il = il->il_next)
				*--ic = il->il_index;
		}
		else {
			ic = indices;
			for (il = dv->dv_ilist; il != NULL; il = il->il_next)
				*ic++ = il->il_index;
		}
	}

	vv = (Vsvar *)alloc(ap, sizeof(Vsvar));
	vv->name = dv->dv_var->va_name;
	vv->format = dv->dv_format;
	vv->ilevel = dv->dv_ilevel;
	vv->icount = icount;
	vv->no_indent = get_no_indent((objid_t)dv);
	vv->indices = indices;

	if (ups_get_object_type(par) == OT_BLOCK)
		bl = obj_to_block(par);
	else
		bl = NULL;

	ve = push_ve(ap, p_velist, bl, (objid_t)dv, VT_VAR);
	ve->u.var = vv;
}

static void
push_expr_state(ap, p_velist, par, obj)
alloc_id_t ap;
Velist **p_velist;
objid_t par, obj;
{
	Vsexpr *vc;
	Velist *ve;
	const char *text;
	vformat_t format;
	block_t *bl;

	get_expr_info(obj, &bl, &text, &format);

	if (ups_get_object_type(par) != OT_BLOCK)
		bl = NULL;

	vc = (Vsexpr *)alloc(ap, sizeof(Vsexpr));
	vc->text = text;
	vc->format = format;

	ve = push_ve(ap, p_velist, bl, obj, VT_EXPR);
	ve->u.expr = vc;
}

static void
push_children(ap, p_velist, par)
alloc_id_t ap;
Velist **p_velist;
objid_t par;
{
	objid_t obj;

	for (obj = get_code(par, OBJ_CHILD);
	     obj != NULL;
	     obj = get_code(obj, OBJ_NEXT)) {
		switch (ups_get_object_type(obj)) {
		case OT_VAR:
			push_var_state(ap, p_velist, par, (dvar_t *)obj);
			break;
		case OT_EXPR:
			push_expr_state(ap, p_velist, par, obj);
			break;
		case OT_BLOCK:
			push_children(ap, p_velist, obj);
			break;
		default:
			panic("unknown obj type in actv"); 
		}
	}
}

void
save_var_display_state()
{
	Vstack *vs;
	Vfile *vf;

	for (vs = Vslist; vs != NULL; vs = vs->next)
		vs->updated = FALSE;
	
	iterate_over_stack_funcs(save_any_stack_changes, (char *)NULL,
				 STK_INNER, STK_OUTER);
	
	for (vs = Vslist; vs != NULL; vs = vs->next)
		vs->updated = TRUE;

	for (vf = Vflist; vf != NULL; vf = vf->next)
		update_vf(vf);
}

static void
update_vf(vf)
Vfile *vf;
{
	if (vf->user_changed_vars) {
		Velist *velist;
		
		alloc_reset_pool(vf->apool);

		velist = NULL;
		push_children(vf->apool, &velist, vf->obj);
		vf->velist = reverse_velist(velist);
		
		vf->user_changed_vars = FALSE;
		vf->been_displayed = TRUE;
	}
}

static
void
update_file_entry(obj)
objid_t obj;
{
	Vfile *vf;

	if ((vf = obj_to_vfile(obj)) != NULL)
		update_vf(vf);
}

void
restore_file_displays()
{
	Vfile *vf;
	
	for (vf = Vflist; vf != NULL; vf = vf->next) {
		if (!vf->been_displayed)
			continue;

		switch (vf->objtype) {
		case OT_CBLOCK:
			add_common_block_object_if_necessary((common_block_id_t)vf->obj);
			break;
		case OT_SFILE:
			add_source_file_object_if_necessary((fil_t *)vf->obj);
			break;
		case OT_GLOBALS:
			break;
		default:
			panic("unknown objtype in rfd");
			break;
		}
	
		restore_display(vf->obj, vf->velist);
		vf->user_changed_vars = FALSE;
	}
}

bool
restore_file_display(obj, force)
objid_t obj;
bool force;
{
	Vfile *vf;
	
	vf = obj_to_vfile(obj);

	if (vf != NULL && (force || vf->been_displayed)) {
		restore_display(vf->obj, vf->velist);
		vf->been_displayed = TRUE;
	}

	return vf != NULL;
}

static 
void
note_file_collapsed(obj)
objid_t obj;
{
	Vfile *vf;
	
	if ((vf = obj_to_vfile(obj)) != NULL) {
		vf->been_displayed = FALSE;
		vf->user_changed_vars = FALSE;
	}
}

void
note_stack_entry_collapsed(stk)
stack_t *stk;
{
	Vstack *vs;
	
	if ((vs = func_to_vstack(stk->stk_func)) != NULL)
		vs->been_displayed = FALSE;

	stk->stk_user_changed_vars = FALSE;
}

bool
restore_stack_entry_display(stk, force)
stack_t *stk;
bool force;
{
	Vstack *vs;
	
	vs = func_to_vstack(stk->stk_func);
	
	if (vs != NULL && (force || vs->been_displayed)) {
		restore_display((objid_t)stk, vs->velist);
		vs->been_displayed = TRUE;
	}

	stk->stk_user_changed_vars = FALSE;

	return vs != NULL;
}

bool
maybe_restore_var_state(par, dv)
objid_t par;
dvar_t *dv;
{
	Vstack *vs;
	Vfile *vf;
	Velist *velist, *ve;

	switch (ups_get_object_type(par)) {
	case OT_FUNC:
		if ((vs = func_to_vstack(((stack_t *)par)->stk_func)) == NULL)
			velist = NULL;
		else
			velist = vs->velist;
		break;
	case OT_SFILE:
	case OT_CBLOCK:
	case OT_GLOBALS:
		if ((vf = obj_to_vfile(par)) == NULL)
			velist = NULL;
		else
			velist = vf->velist;
		break;
	default:
		/*  TODO: OT_BLOCK
		 */
		velist = NULL;
		break;
	}

	for (ve = velist; ve != NULL; ve = ve->next) {
		if (ve->type == VT_VAR &&
		    strcmp(ve->u.var->name, dv->dv_var->va_name) == 0)
			break;
	}

	if (ve == NULL)
		return maybe_restore_format(dv);

	restore_dv_state(ve->u.var, dv);
	restore_children((objid_t)dv, ve->children);

	return TRUE;
}

static void
restore_display(par, velist)
objid_t par;
Velist *velist;
{
	Velist *ve;
	
	for (ve = velist; ve != NULL; ve = ve->next) {
		objid_t obj;
		
		switch(ve->type) {
		case VT_VAR:
			obj = restore_var(par, ve->u.var,
					  ve->block_start, ve->block_end);
			break;
		case VT_EXPR:
			obj = restore_expr(par, ve->u.expr,
					   ve->block_start, ve->block_end);
			break;
		default:
			panic("bad vtype in rdv");
			obj = NULL; /* to satisfy gcc */
		}

		if (obj != NULL)
			restore_children(obj, ve->children);
	}
}

static void
restore_children(par, children)
objid_t par;
Velist *children;
{
	Velist *ve;
	type_t *btype;

	for (ve = children; ve != NULL; ve = ve->next) {
		Vsvar *vv;
		taddr_t junk;
		var_t *v;
		
		if (ve->type != VT_VAR)
			panic("type botch in rc");
		vv = ve->u.var;
				/* RGA function return value reversed in 3.14 */
		if (get_member_of_aggr(par, vv->name, &v, &junk, &btype) != 0) {
			objid_t obj;
		
			obj = add_var_object(par, v, OBJ_LAST_CHILD, TRUE);

			restore_dv_state(vv, (dvar_t *)obj);
			restore_children(obj, ve->children);
		}
	}
}

static objid_t
restore_var(par, vv, block_start, block_end)
objid_t par;
Vsvar *vv;
int block_start, block_end;
{
	objid_t obj;
	
	switch (ups_get_object_type(par)) {
	case OT_FUNC:
		display_local(par, ((stack_t *)par)->stk_func, block_start,
			      vv->name, TRUE, &obj);
		break;
		
	case OT_SFILE:
		show_global((srcwin_id_t)NULL, (fil_t *)par, (func_t *)NULL,
			    (char *)vv->name, FALSE, FALSE, FALSE, TRUE, &obj);
		break;
		
	case OT_GLOBALS:
		show_global((srcwin_id_t)NULL, (fil_t *)NULL, (func_t *)NULL,
			    (char *)vv->name, FALSE, FALSE, FALSE, TRUE, &obj);
		
		break;
		
	case OT_CBLOCK:
		show_global((srcwin_id_t)NULL, (fil_t *)NULL, (func_t *)NULL,
			    (char *)vv->name, FALSE, FALSE, FALSE, TRUE, &obj);
		
		break;
		
	default:
		panic("bad objtype in rv");
	}

	if (obj != NULL)
		restore_dv_state(vv, (dvar_t *)obj);
	
	return obj;
}

static void
restore_dv_state(vv, dv)
Vsvar *vv;
dvar_t *dv;
{
	make_vv_ilist(dv, vv->indices, vv->icount);
	
	dv->dv_format = vv->format;

	if (vv->no_indent)
		set_no_indent((objid_t)dv, TRUE);

	redo_decl(dv);
}

static void
make_vv_ilist(dv, indices, count)
dvar_t *dv;
int *indices, count;
{
	int i;
	ilist_t *il;
	
	for (i = dv->dv_ilevel; i < count; ++i)
		change_dv_level(dv, CL_UP);

	/*  It is not a fatal error for there not to be enough indices,
	 *  as the saved state may not match the variable (it can come
	 *  from an out of date state file, for example).
	 */
	if (dv->dv_var->va_language == LANG_FORTRAN) {
		i = count;
		for (il = dv->dv_ilist; il != NULL; il = il->il_next) {
			if (i > 0)
				il->il_index = indices[--i];
		}
	}
	else {
		i = 0;
		for (il = dv->dv_ilist; il != NULL; il = il->il_next) {
			if (i < count)
				il->il_index = indices[i++];
		}
	}
}

static block_t *
lnum_range_to_block(parbl, start_lnum, end_lnum)
block_t *parbl;
int start_lnum, end_lnum;
{
	block_t *bl;
	
	if (start_lnum == parbl->bl_start_lnum &&
	    end_lnum == parbl->bl_end_lnum)
		return parbl;

	for (bl = parbl->bl_blocks; bl != NULL; bl = bl->bl_next) {
		block_t *cbl;

		cbl = lnum_range_to_block(bl, start_lnum, end_lnum); 
		if (cbl != NULL)
			return cbl;
	}

	return NULL;
}

objid_t
add_blocks_as_needed(par, rootbl, bl)
objid_t par;
block_t *rootbl, *bl;
{
	objid_t child;
	
	if (bl == rootbl)
		return par;
	
	if (bl->bl_parent != rootbl)
		par = add_blocks_as_needed(par, rootbl, bl->bl_parent);
	
	if ((child = find_block(par, bl)) == 0)
		child = add_block_object(par, bl);

	return child;
}

static objid_t
restore_expr(par, vc, block_start, block_end)
objid_t par;
Vsexpr *vc;
int block_start, block_end;
{
	block_t *rootbl, *bl;
	
	switch (ups_get_object_type(par)) {
	case OT_FUNC:
		rootbl = FU_BLOCKS(((stack_t *)par)->stk_func);
		break;
	case OT_SFILE:
		rootbl = ((fil_t *)par)->fi_block;
		break;
	default:
		panic("bad objtype in rv");
		rootbl = NULL;	/* to satisfy gcc */
	}

	if (block_start == 0 && block_end == 0)
		bl = rootbl;
	else
		bl = lnum_range_to_block(rootbl, block_start, block_end);
	
	if (bl == NULL) {
		errf("Can't find block to restore expression `%s'",
		     vc->text);
		return NULL;
	}

	par = add_blocks_as_needed(par, rootbl, bl);
	
	return (objid_t)restore_displayed_expr(par, bl, vc->text, vc->format);
}

int
write_var_state_to_file(path, fp)
const char *path;
FILE *fp;
{
	Vstack *vs;
	Vfile *vf;

	for (vs = Vslist; vs != NULL; vs = vs->next) {
		func_t *f;
		lno_t *ln;

		if (vs->velist == NULL)
			continue;

		f = vs->func;
		ln = FU_LNOS(f);
		
		fprintf(fp, "\nfunction %s", f->fu_name);
		
		if (f->fu_fil != NULL) {
			fprintf(fp, " %s %d",
				f->fu_fil->fi_name,
				(ln != NULL) ? ln->ln_num : 0);
		}

		if (dump_children(path, fp, vs->velist, 1) != 0)
			return -1;
	}

	for (vf = Vflist; vf != NULL; vf = vf->next) {
		if (vf->velist == NULL)
			continue;

		fputc('\n', fp);

		switch (vf->objtype) {
			common_block_id_t cb;
			const char *funcname;
			
		case OT_GLOBALS:
			fputs("globals", fp);
			break;
			
		case OT_CBLOCK:
			cb = (common_block_id_t)vf->obj;
			fprintf(fp, "common %s", get_cblock_name(cb));
			
			funcname = get_cblock_funcname(cb);
			if (funcname != NULL)
				fprintf(fp, " %s", funcname);
			
			break;

		case OT_SFILE:
			fprintf(fp, "file %s", ((fil_t *)vf->obj)->fi_name);
			break;

		default:
			panic("unknown objtype in wvstf");
			break;
		}

		if (dump_children(path, fp, vf->velist, 1) != 0)
			return -1;
	}

	return write_format_state_to_file(path, fp);
}

static Velist *
reverse_velist(velist)
Velist *velist;
{
	Velist *ve, *next, *newlist;

	newlist = NULL;

	for (ve = velist; ve != NULL; ve = next) {
		next = ve->next;
		ve->next = newlist;
		newlist = ve;
	}

	return newlist;
}

static int
dump_children(path, fp, velist, level)
const char *path;
FILE *fp;
Velist *velist;
int level;
{
	Velist *ve;
	bool done_one;

	done_one = FALSE;

	for (ve = velist; ve != NULL; ve = ve->next) {
		int i;

		if (!done_one) {
			fputs(" {\n", fp);
			done_one = TRUE;
		}
		
		for (i = 0; i < level; ++i)
			fputc('\t', fp);
		
		if (ve->block_start != 0 || ve->block_end != 0)
			fprintf(fp, "%d..%d ", ve->block_start, ve->block_end);
		
		switch (ve->type) {
		case VT_VAR:
			dump_var(fp, ve->u.var);
			break;
		case VT_EXPR:
			dump_expr(fp, ve->u.expr);
			break;
		default:
			panic("bad vetype in dc");
			break;
		}

		if (ferror(fp)) {
			errf("Error writing to", "file", path);
			break;
		}

		if (dump_children(path, fp, ve->children, level + 1) != 0)
			break;
	}

	if (ve != NULL)
		return -1;
	
	if (done_one) {
		int i;
		
		for (i = 0; i < level - 1; ++i)
			fputc('\t', fp);
		fputc('}', fp);
	}

	fputc('\n', fp);

	return 0;
}

static void
dump_var(fp, vv)
FILE *fp;
Vsvar *vv;
{
	int i;
	
	fprintf(fp, "var %-15s %-16s %s %d [",
		vv->name,
		format_to_string(vv->format),
		vv->no_indent ? "no-indent" : "indent",
		vv->ilevel);

	for (i = 0; i < vv->icount; ++i) {
		fprintf(fp, "%d%s",
			vv->indices[i],
			(i < vv->icount - 1) ? " " : "");
	}

	fputc(']', fp);
}

static void
dump_expr(fp, vc)
FILE *fp;
Vsexpr *vc;
{
	fprintf(fp, "expr %s ", format_to_string(vc->format));
	put_quoted_string(fp, vc->text);
}

int
handle_function_command(cmd, args, nargs, from_statefile, fp, p_lnum)
const char *cmd;
char **args;
int nargs;
bool from_statefile;
FILE *fp;
int *p_lnum;
{
	func_t *f, *f1;
	char *funcname;
	const char *filename, *lnumstr;
	int func_lnum, line_offset;
	alloc_id_t ap;
	Velist *velist;
	lno_t *ln;
	Vstack *vs;
		
	if (nargs < 1 || strcmp(args[nargs - 1], "{") != 0) {
		errf("Missing `{'");
		return -1;
	}
	--nargs;

	if (nargs < 1 || nargs > 3) {
		errf("Usage: function funcname [filename [lnum]] {");
		return -1;
	}

	
	funcname = args[0];
	filename = (nargs > 1) ? args[1] : NULL;
	lnumstr = (nargs > 2) ? args[2] : NULL;

	if (filename == NULL) {
		if (find_func_by_name(funcname, &f, &f1, FALSE) != 0)
			f = NULL;
	}
	else {
		fil_t *fil;
		
		if ((fil = name_to_fil(filename)) == NULL ||
		    (f = name_and_fil_to_func(funcname, fil)) == NULL) {
			/*  Function may have moved to another file.
			 */
			if (find_func_by_name(funcname, &f, &f1, FALSE) != 0)
				f = NULL;
		}
	}

	if (lnumstr == NULL) {
		func_lnum = -1;
	}
	else {
		char *ends;

		func_lnum = strtol(lnumstr, &ends, 10);
		if (func_lnum < 0 || ends == lnumstr || *ends != '\0') {
			errf("Illegal line number `%s' ignored", lnumstr);
			func_lnum = -1;
		}
	}

	if (f != NULL && func_lnum != -1 && (ln = FU_LNOS(f)) != NULL)
		line_offset = ln->ln_num - func_lnum;
	else
		line_offset = 0;

	if (load_velist(fp, line_offset, p_lnum, f == NULL, &ap, &velist) != 0)
		return -1;
	
	if ((vs = func_to_vstack(f)) != NULL) {
		alloc_free_pool(vs->apool);
		vs->apool = ap;
		vs->velist = velist;
	}
	else {
		push_new_vstack(f, ap, velist);
	}

	return 0;
}

static int
load_velist(fp, line_offset, p_lnum, failed, p_ap, p_velist)
FILE *fp;
int line_offset;
int *p_lnum;
bool failed;
alloc_id_t *p_ap;
Velist **p_velist;
{
	alloc_id_t ap;
	Velist *velist;

	ap = alloc_create_pool();
	velist = NULL;

	if (read_velist(ap, &velist, fp, line_offset, p_lnum) != 0)
		failed = TRUE;

	if (failed) {
		alloc_free_pool(ap);
		return -1;
	}

	*p_ap = ap;
	*p_velist = velist;
	return 0;
}

static int
read_velist(ap, p_velist, fp, line_offset, p_lnum)
alloc_id_t ap;
Velist **p_velist;
FILE *fp;
int line_offset;
int *p_lnum;
{
	int save_lnum, res;
	char *line;
	Velist first, *last;

	save_lnum = *p_lnum;

	last = &first;
	res = 0;
	
	while ((line = fpgetline(fp)) != NULL) {
		char *cmd;
		char *endline;
		bool have_children;
		Velist *ve;
		int block_start, block_end;
		
		++*p_lnum;

		line = config_trim_line(line);
		if (*line == '\0')
			continue;

		if (strcmp(line, "}") == 0)
			break;

		endline = line + strlen(line) - 1;
		have_children = *endline == '{';
		
		if (have_children)
			*endline = '\0';

		if (isdigit(*line)) {
			char *rangestr, junkc;
			
			if (!get_word(&line, "line range", &rangestr)) {
				res = -1;
				continue;
			}
			
			if (sscanf(rangestr, "%d..%d%c",
				   &block_start, &block_end, &junkc) != 2) {
				errf("Expected num..num, got `%s' - ignored",
				     rangestr);
				block_start = block_end = 0;
			}
			else {
				block_start += line_offset;
				block_end += line_offset;
			}
		}
		else {
			block_start = block_end = 0;
		}
		
		ve = (Velist *)alloc(ap, sizeof(Velist));
		ve->block_start = block_start;
		ve->block_end = block_end;
		ve->children = NULL;

		if (!get_word(&line, "`var' or `expr'", &cmd)) {
			res = -1;
		}
		else if (strcmp(cmd, "var") == 0) {
			ve->type = VT_VAR;
			ve->u.var = (Vsvar *)alloc(ap, sizeof(Vsvar));
			
			if (read_var(ap, line, ve->u.var) != 0)
				res = -1;
		}
		else if (strcmp(cmd, "expr") == 0) {
			ve->type = VT_EXPR;
			ve->u.expr = (Vsexpr *)alloc(ap, sizeof(Vsexpr));
			
			if (read_expr(ap, line, ve->u.expr) != 0)
				res = -1;
		}
		else {
			errf("Expected `var' or `expr', got `%s'", cmd);
			res = -1;
		}
		
		if (have_children &&
		    read_velist(ap, &ve->children, fp,
				line_offset, p_lnum) != 0) {
			res = -1;
		}

		last->next = ve;
		last = ve;
	}
	
	if (line == NULL) {
		*p_lnum = save_lnum;
		errf("Missing closing `}'");
		res = -1;
	}

	last->next = NULL;
	*p_velist = first.next;
	return res;
}

static bool
get_word(p_line, what, p_cmd)
char **p_line;
const char *what;
char **p_cmd;
{
	char *line;

	line = *p_line;

	while (isspace(*line))
		++line;

	if (*line == '\0') {
		errf("Expected %s but got end of line", what);
		return FALSE;
	}

	*p_cmd = line;
	
	while (!isspace(*line) && *line != '\0')
		++line;

	while (isspace(*line))
		*line++ = '\0';

	*p_line = line;
	return TRUE;
}

bool
get_num(p_line, what, p_val)
char **p_line;
const char *what;
int *p_val;
{
	char *str, *endstr;
	
	if (!get_word(p_line, what, &str))
		return FALSE;

	*p_val = strtol(str, &endstr, 10);

	if (*endstr != '\0') {
		errf("%c%s `%s' is not a decimal integer",
		     toupper(*what), what + 1, str);
		return FALSE;
	}

	return TRUE;
}

static int
read_var(ap, line, vv)
alloc_id_t ap;
char *line;
Vsvar *vv;
{
	static ebuf_t *eb = NULL;
	int int_nbytes, ilevel;
	size_t nbytes;
	char *name, *formatstr, *indentstr, *ibuf, *closebr;
	vformat_t format;
	bool indent;
	
	if (!get_word(&line, "variable name", &name) ||
	    !get_word(&line, "value format", &formatstr) ||
	    !get_word(&line, "`indent' or `no-indent'", &indentstr) ||
	    !get_num(&line, "indirection level", &ilevel))
		return -1;

	if (!string_to_format(formatstr, &format))
		return -1;
	
	indent = strcmp(indentstr, "indent") == 0;
	if (!indent && strcmp(indentstr, "no-indent") != 0) {
		errf("Expected `indent' or `no-indent', got `%s'", indentstr);
		return -1;
	}

	if (*line != '[') {
		if (*line == '\0')
			errf("Missing `['");
		else
			errf("Expected '[', got `%c'", *line);
		return -1;
	}

	for (++line; isspace(*line); ++line)
		;
	
	if ((closebr = strchr(line, ']')) == NULL) {
		errf("Missing ']'");
		return -1;
	}
	*closebr = '\0';
	
	eb = ebuf_start(eb, TRUE);

	while (*line != '\0') {
		int n;

		if (!get_num(&line, "array index", &n))
			return -1;

		ebuf_add(eb, (char *)&n, sizeof(n));
	}

	for (line = closebr + 1; isspace(*line); ++line)
		;
	
	if (*line != '\0') {
		errf("Unexpected text at end of line: `%s'", line);
		return -1;
	}
	
	ibuf = ebuf_get(eb, &int_nbytes);
	nbytes = (size_t)int_nbytes; /* due to wrong ebuf_get interface */

	vv->name = alloc_strdup(ap, name);
	vv->format = format;
	vv->no_indent = !indent;
	vv->ilevel = ilevel;
	vv->icount = nbytes / sizeof(int);
	vv->indices = (int *)memcpy(alloc(ap, nbytes), ibuf, nbytes);
		
	return 0;
}

static int
read_expr(ap, line, vc)
alloc_id_t ap;
char *line;
Vsexpr *vc;
{
	vformat_t format;
	char **words;
	int nwords;

	if (!parse_line(line, &words, &nwords))
		return FALSE;

	if (nwords != 2) {
		errf("Expected format exprstring");
		free((char *)words);
		return -1;
	}
	
	if (!string_to_format(words[0], &format)) {
		free((char *)words);
		return -1;
	}
	
	vc->text = alloc_strdup(ap, words[1]);
	vc->format = format;

	free((char *)words);
	
	return 0;
}

int
handle_file_command(cmd, args, nargs, from_statefile, fp, p_lnum)
const char *cmd;
char **args;
int nargs;
bool from_statefile;
FILE *fp;
int *p_lnum;
{
	objid_t obj;
	int res, objtype;
	alloc_id_t ap;
	Velist *velist;
	Vfile *vf;
	
	res = 0;

	if (strcmp(cmd, "globals") == 0) {
		if (nargs != 1 || strcmp(args[0], "{") != 0) {
			errf("Usage: globals {");
			return -1;
		}
		
		obj = GLOBALS_OBJCODE;
		objtype = OT_GLOBALS;
	}
	else if (strcmp(cmd, "file") == 0) {
		if (nargs != 2 || strcmp(args[1], "{") != 0) {
			errf("Usage: file filename {");
			return -1;
		}
		
		obj = (objid_t)name_to_fil(args[0]);

		if (obj == NULL)
			errf("Unknown filename `%s'", args[0]);

		objtype = OT_SFILE;
	}
	else if (strcmp(cmd, "common") == 0) {
		func_t *f, *f1;
		
		if ((nargs != 2 && nargs != 3) ||
		    strcmp(args[nargs - 1], "{") != 0) {
			errf("Usage: common cblockname [func] {");
			return -1;
		}
		
		if (nargs == 2 || find_func_by_name(args[1], &f, &f1, FALSE) != 0)
			f = NULL;
		
/*		obj = (objid_t)name_to_cblock(args[0], f);*/
		obj = NULL;	/* RGA */

		if (obj == NULL)
			errf("Unknown common block `%s'", args[0]);

		objtype = OT_CBLOCK;
	}
	else {
		panic("bad cmd in hfc");
		obj = NULL;		/* to satisfy gcc */
		objtype = OT_MAXTYPE;	/* to satisfy gcc */
	}

	if (load_velist(fp, 0, p_lnum, obj == NULL, &ap, &velist) != 0)
		return -1;

	if ((vf = obj_to_vfile(obj)) != NULL) {
		alloc_free_pool(vf->apool);
		vf->apool = ap;
		vf->velist = velist;
	}
	else {
		vf = push_new_vfile(obj, objtype, ap, velist);
	}

	return res;
}

/************** End va_state.c from version 3.14 ***********************/


int
typesize(type)
type_t *type;
{
	return dynamic_type_size(type, (ilist_t *)NULL);
}

/*  Return the size in bytes of type.
 *
 *  This routine just calls ci_typesize, except for arrays, where it
 *  checks for variable size arrays.
 */
int
dynamic_type_size(type, il)
type_t *type;
ilist_t *il;
{
	int size;

	if (type->ty_size != -1)
		size = type->ty_size;
	else if (type->ty_code == DT_ARRAY_OF) {
		dim_t *dim;
		int basesize, high, low, size_is_variable;

		dim = type->ty_dim;
		size_is_variable = dim->di_ldynamic || dim->di_hdynamic;
		if (size_is_variable) {
			if (il == NULL || !il->il_low_known || !il->il_high_known)
				return UNKNOWN_SIZE;
			low = il->il_low;
			high = il->il_high;
		}
		else {
			low = dim->di_low;
			high = dim->di_high;
		}
		basesize = dynamic_type_size(type->ty_base,
				    (il != NULL) ? il->il_next
						 : (ilist_t *)NULL);

		/*  The VAX Ultrix C compiler has a bug in it's symbol
		 *  table code (surprise!) - it emits arrays with a size
		 *  of zero if the first thing it sees is a pointer to
		 *  the array.
		 */
		if (high <= low) {
			errf("%s array size - assumed one",
					(low < high) ? "Negative" : "Zero");
			high = low + 1;
		}

		if (basesize == UNKNOWN_SIZE)
			size = UNKNOWN_SIZE;
		else
			size = (high - low) * basesize;
		if (!size_is_variable)
			type->ty_size = size;
	}
	else
		size = ci_typesize((lexinfo_t *)NULL, type);

	if (size <= 0 && size != UNKNOWN_SIZE)
		panic("non positive type size in typesize");

	return size;
}

int
fix_if_fortran_dynamic_char(type, addr, ilist)
type_t *type;
taddr_t addr;
ilist_t *ilist;
{
	type_t *prev, *dtype;
	int size;

	prev = NULL;
	for (dtype = type; ISDERIV(dtype->ty_code); dtype = dtype->ty_base)
		prev = dtype;
	if (dtype->ty_code == TY_CHARACTER && prev != NULL &&
					prev->ty_code == DT_ARRAY_OF &&
					prev->ty_dim->di_hdynamic) {

		for (; type != prev; type = type->ty_base)
			ilist = ilist->il_next;

		if (dread(addr + prev->ty_dim->di_high - 1,
						(char *)&size, sizeof(int)) == 0) {
			ilist->il_high_known = TRUE;
			ilist->il_high = size + 1;
		}
		else
			ilist->il_high_known = FALSE;
		return TRUE;
	}
	else
		return FALSE;
}

type_t *
get_basetype(type)
type_t *type;
{
	while (ISDERIV(type->ty_code))
		type = type->ty_base;
	return type;
}

type_t *
get_type_at_level(v, level)
var_t *v;
int level;
{
	type_t *type;

	for (type = v->va_type; type != NULL && --level >= 0; type = type->ty_base)
		;
	if (type == NULL)
		panic("level too high in get_type_at_level");
	return type;
}
